/* Generated by Cython 0.24 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "extensions/lib/websocket.h"
        ],
        "include_dirs": [
            "extensions/lib"
        ]
    }
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03020000)
    #error Cython requires Python 2.6+ or Python 3.2+.
#else
#define CYTHON_ABI "0_24"
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
#endif
#if !defined(CYTHON_USE_PYLONG_INTERNALS) && CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x02070000
  #define CYTHON_USE_PYLONG_INTERNALS 1
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : PyInstanceMethod_New(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if PY_VERSION_HEX >= 0x030500B1
#define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
#define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
#elif CYTHON_COMPILING_IN_CPYTHON && PY_MAJOR_VERSION >= 3
typedef struct {
    unaryfunc am_await;
    unaryfunc am_aiter;
    unaryfunc am_anext;
} __Pyx_PyAsyncMethodsStruct;
#define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
#else
#define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)

#ifndef CYTHON_INLINE
  #if defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__lib
#define __PYX_HAVE_API__lib
#include "math.h"
#include "websocket.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#ifdef PYREX_WITHOUT_ASSERTIONS
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER) && defined (_M_X64)
    #define __Pyx_sst_abs(value) _abs64(value)
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyObject_AsSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#if PY_MAJOR_VERSION < 3
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u)
{
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#else
#define __Pyx_Py_UNICODE_strlen Py_UNICODE_strlen
#endif
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
#define __Pyx_PyBool_FromLong(b) ((b) ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False))
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_COMPILING_IN_CPYTHON
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c));
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */

static PyObject *__pyx_m;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "extensions/lib/rparser.pyx",
  "extensions/lib/websocket.pyx",
  "extensions/lib/websocket.pxd",
  "extensions/lib/lib.pxd",
  "extensions/lib/lib.pyx",
};

/*--- Type declarations ---*/
struct __pyx_obj_3lib_RedisParser;
struct __pyx_obj_3lib_Task;
struct __pyx_obj_3lib_ArrayTask;
struct __pyx_obj_3lib_Frame;
struct __pyx_obj_3lib_FrameParser;
struct __pyx_obj_3lib___pyx_scope_struct__pack_pipeline;
struct __pyx_obj_3lib___pyx_scope_struct_1_genexpr;
struct __pyx_obj_3lib___pyx_scope_struct_2__pack_command;
struct __pyx_obj_3lib___pyx_scope_struct_3__pack;
struct __pyx_obj_3lib___pyx_scope_struct_4__lua_dict;
struct __pyx_obj_3lib___pyx_scope_struct_5_multi_encode;

/* "extensions/lib/rparser.pyx":15
 * 
 * 
 * cdef class RedisParser:             # <<<<<<<<<<<<<<
 *     cdef object _protocolError
 *     cdef object _responseError
 */
struct __pyx_obj_3lib_RedisParser {
  PyObject_HEAD
  struct __pyx_vtabstruct_3lib_RedisParser *__pyx_vtab;
  PyObject *_protocolError;
  PyObject *_responseError;
  PyObject *_encoding;
  PyObject *_inbuffer;
  struct __pyx_obj_3lib_Task *_current;
};


/* "extensions/lib/rparser.pyx":3
 * from itertools import starmap
 * 
 * cdef class Task             # <<<<<<<<<<<<<<
 * 
 * cdef bytes CRLF = b"\r\n"
 */
struct __pyx_obj_3lib_Task {
  PyObject_HEAD
  struct __pyx_vtabstruct_3lib_Task *__pyx_vtab;
  long _length;
  struct __pyx_obj_3lib_Task *_next;
};


/* "extensions/lib/rparser.pyx":173
 * 
 * 
 * cdef class ArrayTask(Task):             # <<<<<<<<<<<<<<
 *     cdef list _response
 * 
 */
struct __pyx_obj_3lib_ArrayTask {
  struct __pyx_obj_3lib_Task __pyx_base;
  PyObject *_response;
};


/* "extensions/lib/websocket.pyx":8
 * 
 * 
 * cdef class Frame:             # <<<<<<<<<<<<<<
 *     cdef int _opcode
 *     cdef bint _final
 */
struct __pyx_obj_3lib_Frame {
  PyObject_HEAD
  int _opcode;
  int _final;
  int _payload_length;
  PyObject *_masking_key;
  PyObject *_body;
};


/* "extensions/lib/websocket.pyx":57
 * 
 * 
 * cdef class FrameParser:             # <<<<<<<<<<<<<<
 *     '''Encoder and decoder of WebSocket frames.
 * 
 */
struct __pyx_obj_3lib_FrameParser {
  PyObject_HEAD
  struct __pyx_vtabstruct_3lib_FrameParser *__pyx_vtab;
  struct __pyx_obj_3lib_Frame *frame;
  PyObject *buffer;
  PyObject *ProtocolError;
  int kind;
  int _version;
  int _decode_mask_length;
  int _encode_mask_length;
  PyObject *_opcodes;
  PyObject *_extensions;
  PyObject *_protocols;
  PyObject *_close_codes;
};


/* "extensions/lib/rparser.pyx":51
 *         return b''.join(self._pack_command(args))
 * 
 *     def pack_pipeline(self, commands):             # <<<<<<<<<<<<<<
 *         pack = lambda *args: b''.join(self._pack_command(args))
 *         return b''.join(starmap(pack, (args for args, _ in commands)))
 */
struct __pyx_obj_3lib___pyx_scope_struct__pack_pipeline {
  PyObject_HEAD
  PyObject *__pyx_v_commands;
  struct __pyx_obj_3lib_RedisParser *__pyx_v_self;
};


/* "extensions/lib/rparser.pyx":53
 *     def pack_pipeline(self, commands):
 *         pack = lambda *args: b''.join(self._pack_command(args))
 *         return b''.join(starmap(pack, (args for args, _ in commands)))             # <<<<<<<<<<<<<<
 * 
 *     # SERVER ENCODERS
 */
struct __pyx_obj_3lib___pyx_scope_struct_1_genexpr {
  PyObject_HEAD
  struct __pyx_obj_3lib___pyx_scope_struct__pack_pipeline *__pyx_outer_scope;
  PyObject *__pyx_v__;
  PyObject *__pyx_v_args;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "extensions/lib/rparser.pyx":69
 * 
 *     # INTERNALS
 *     def _pack_command(self, args):             # <<<<<<<<<<<<<<
 *         yield ('*%d\r\n' % len(args)).encode('utf-8')
 *         for value in args:
 */
struct __pyx_obj_3lib___pyx_scope_struct_2__pack_command {
  PyObject_HEAD
  PyObject *__pyx_v_args;
  struct __pyx_obj_3lib_RedisParser *__pyx_v_self;
  PyObject *__pyx_v_value;
  Py_ssize_t __pyx_t_0;
  PyObject *__pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "extensions/lib/rparser.pyx":80
 *             yield CRLF
 * 
 *     def _pack(self, args):             # <<<<<<<<<<<<<<
 *         yield ('*%d\r\n' % len(args)).encode('utf-8')
 *         for value in args:
 */
struct __pyx_obj_3lib___pyx_scope_struct_3__pack {
  PyObject_HEAD
  PyObject *__pyx_v_args;
  struct __pyx_obj_3lib_RedisParser *__pyx_v_self;
  PyObject *__pyx_v_value;
  Py_ssize_t __pyx_t_0;
  PyObject *__pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
  PyObject *__pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
};


/* "extensions/lib/rparser.pyx":106
 *                 yield CRLF
 * 
 *     def _lua_dict(self, d):             # <<<<<<<<<<<<<<
 *         index = 0
 *         while True:
 */
struct __pyx_obj_3lib___pyx_scope_struct_4__lua_dict {
  PyObject_HEAD
  PyObject *__pyx_v_d;
  PyObject *__pyx_v_index;
  struct __pyx_obj_3lib_RedisParser *__pyx_v_self;
  PyObject *__pyx_v_v;
};


/* "extensions/lib/websocket.pyx":156
 *                             rsv1, rsv2, rsv3)
 * 
 *     def multi_encode(self, message, bytes masking_key=None, opcode=None,             # <<<<<<<<<<<<<<
 *                      int rsv1=0, int rsv2=0, int rsv3=0,
 *                      cython.ulonglong max_payload=0):
 */
struct __pyx_obj_3lib___pyx_scope_struct_5_multi_encode {
  PyObject_HEAD
  PyObject *__pyx_v_chunk;
  PyObject *__pyx_v_data;
  int __pyx_v_fin;
  PyObject *__pyx_v_masking_key;
  unsigned PY_LONG_LONG __pyx_v_max_payload;
  PyObject *__pyx_v_message;
  PyObject *__pyx_v_opcode;
  int __pyx_v_rsv1;
  int __pyx_v_rsv2;
  int __pyx_v_rsv3;
  struct __pyx_obj_3lib_FrameParser *__pyx_v_self;
};



/* "extensions/lib/rparser.pyx":148
 * 
 * 
 * cdef class Task:             # <<<<<<<<<<<<<<
 *     cdef long _length
 *     cdef Task _next
 */

struct __pyx_vtabstruct_3lib_Task {
  PyObject *(*decode)(struct __pyx_obj_3lib_Task *, struct __pyx_obj_3lib_RedisParser *, PyObject *);
};
static struct __pyx_vtabstruct_3lib_Task *__pyx_vtabptr_3lib_Task;


/* "extensions/lib/rparser.pyx":15
 * 
 * 
 * cdef class RedisParser:             # <<<<<<<<<<<<<<
 *     cdef object _protocolError
 *     cdef object _responseError
 */

struct __pyx_vtabstruct_3lib_RedisParser {
  PyObject *(*_get)(struct __pyx_obj_3lib_RedisParser *, struct __pyx_obj_3lib_Task *);
  PyObject *(*_resume)(struct __pyx_obj_3lib_RedisParser *, struct __pyx_obj_3lib_Task *, PyObject *);
};
static struct __pyx_vtabstruct_3lib_RedisParser *__pyx_vtabptr_3lib_RedisParser;


/* "extensions/lib/rparser.pyx":173
 * 
 * 
 * cdef class ArrayTask(Task):             # <<<<<<<<<<<<<<
 *     cdef list _response
 * 
 */

struct __pyx_vtabstruct_3lib_ArrayTask {
  struct __pyx_vtabstruct_3lib_Task __pyx_base;
};
static struct __pyx_vtabstruct_3lib_ArrayTask *__pyx_vtabptr_3lib_ArrayTask;


/* "extensions/lib/websocket.pyx":57
 * 
 * 
 * cdef class FrameParser:             # <<<<<<<<<<<<<<
 *     '''Encoder and decoder of WebSocket frames.
 * 
 */

struct __pyx_vtabstruct_3lib_FrameParser {
  PyObject *(*_encode)(struct __pyx_obj_3lib_FrameParser *, PyObject *, int, PyObject *, int, int, int, int);
  PyObject *(*_info)(struct __pyx_obj_3lib_FrameParser *, PyObject *, PyObject *, PyObject *);
  PyObject *(*_chunk)(struct __pyx_obj_3lib_FrameParser *, int);
};
static struct __pyx_vtabstruct_3lib_FrameParser *__pyx_vtabptr_3lib_FrameParser;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectGetAttrStr.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* StringJoin.proto */
#if PY_MAJOR_VERSION < 3
#define __Pyx_PyString_Join __Pyx_PyBytes_Join
#define __Pyx_PyBaseString_Join(s, v) (PyUnicode_CheckExact(s) ? PyUnicode_Join(s, v) : __Pyx_PyBytes_Join(s, v))
#else
#define __Pyx_PyString_Join PyUnicode_Join
#define __Pyx_PyBaseString_Join PyUnicode_Join
#endif
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION < 3
    #define __Pyx_PyBytes_Join _PyString_Join
    #else
    #define __Pyx_PyBytes_Join _PyBytes_Join
    #endif
#else
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values);
#endif

/* KeywordStringCheck.proto */
static CYTHON_INLINE int __Pyx_CheckKeywordStrings(PyObject *kwdict, const char* function_name, int kw_allowed);

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunction.proto */
#define __Pyx_CyFunction_USED 1
#include <structmember.h>
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_NewEx(ml, flags, qualname, self, module, globals, code)\
    __Pyx_CyFunction_New(__pyx_CyFunctionType, ml, flags, qualname, self, module, globals, code)
static PyObject *__Pyx_CyFunction_New(PyTypeObject *, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* GetModuleGlobalName.proto */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name);

/* ArgTypeTest.proto */
static CYTHON_INLINE int __Pyx_ArgTypeTest(PyObject *obj, PyTypeObject *type, int none_allowed,
    const char *name, int exact);

/* PyIntBinop.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* decode_c_bytes.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_bytes(
         const char* cstring, Py_ssize_t length, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* decode_bytes.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_bytes(
         PyObject* string, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    return __Pyx_decode_c_bytes(
        PyBytes_AS_STRING(string), PyBytes_GET_SIZE(string),
        start, stop, encoding, errors, decode_func);
}

/* PyThreadStateGet.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = PyThreadState_GET();
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* ListAppend.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* PyIntBinop.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx_PyInt_AndObjC(PyObject *op1, PyObject *op2, long intval, int inplace);
#else
#define __Pyx_PyInt_AndObjC(op1, op2, intval, inplace)\
    (inplace ? PyNumber_InPlaceAnd(op1, op2) : PyNumber_And(op1, op2))
#endif

/* PyIntBinop.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx_PyInt_RshiftObjC(PyObject *op1, PyObject *op2, long intval, int inplace);
#else
#define __Pyx_PyInt_RshiftObjC(op1, op2, intval, inplace)\
    (inplace ? PyNumber_InPlaceRshift(op1, op2) : PyNumber_Rshift(op1, op2))
#endif

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* PyIntBinop.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, long intval, int inplace);
#else
#define __Pyx_PyInt_EqObjC(op1, op2, intval, inplace)\
    PyObject_RichCompare(op1, op2, Py_EQ)
    #endif

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* append.proto */
static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_PY_LONG_LONG(unsigned PY_LONG_LONG value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_PY_LONG_LONG(PY_LONG_LONG value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned PY_LONG_LONG __Pyx_PyInt_As_unsigned_PY_LONG_LONG(PyObject *);

/* SwapException.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyObject *);
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(PyTypeObject *type, __pyx_coroutine_body_t body,
                                                   PyObject *closure, PyObject *name, PyObject *qualname);
static int __Pyx_Coroutine_clear(PyObject *self);
#if 1 || PY_VERSION_HEX < 0x030300B0
static int __Pyx_PyGen_FetchStopIterationValue(PyObject **pvalue);
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue) PyGen_FetchStopIterationValue(pvalue)
#endif

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, closure, name, qualname)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, closure, name, qualname)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* PyObjectSetAttrStr.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o,n,NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_setattr))
        return tp->tp_setattr(obj, PyString_AS_STRING(attr_name), value);
#endif
    return PyObject_SetAttr(obj, attr_name, value);
}
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* VoidPtrExport.proto */
static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static PyObject *__pyx_f_3lib_11RedisParser__get(struct __pyx_obj_3lib_RedisParser *__pyx_v_self, struct __pyx_obj_3lib_Task *__pyx_v_next); /* proto*/
static PyObject *__pyx_f_3lib_11RedisParser__resume(struct __pyx_obj_3lib_RedisParser *__pyx_v_self, struct __pyx_obj_3lib_Task *__pyx_v_task, PyObject *__pyx_v_result); /* proto*/
static PyObject *__pyx_f_3lib_4Task_decode(struct __pyx_obj_3lib_Task *__pyx_v_self, struct __pyx_obj_3lib_RedisParser *__pyx_v_parser, CYTHON_UNUSED PyObject *__pyx_v_result); /* proto*/
static PyObject *__pyx_f_3lib_9ArrayTask_decode(struct __pyx_obj_3lib_ArrayTask *__pyx_v_self, struct __pyx_obj_3lib_RedisParser *__pyx_v_parser, PyObject *__pyx_v_result); /* proto*/
static PyObject *__pyx_f_3lib_11FrameParser__encode(struct __pyx_obj_3lib_FrameParser *__pyx_v_self, PyObject *__pyx_v_data, int __pyx_v_opcode, PyObject *__pyx_v_masking_key, int __pyx_v_fin, int __pyx_v_rsv1, int __pyx_v_rsv2, int __pyx_v_rsv3); /* proto*/
static PyObject *__pyx_f_3lib_11FrameParser__info(struct __pyx_obj_3lib_FrameParser *__pyx_v_self, PyObject *__pyx_v_message, PyObject *__pyx_v_opcode, PyObject *__pyx_v_masking_key); /* proto*/
static PyObject *__pyx_f_3lib_11FrameParser__chunk(struct __pyx_obj_3lib_FrameParser *__pyx_v_self, int __pyx_v_length); /* proto*/

/* Module declarations from 'cython' */

/* Module declarations from 'websocket' */
static CYTHON_INLINE PyObject *__pyx_f_9websocket_to_bytes(PyObject *, PyObject *); /*proto*/

/* Module declarations from 'lib' */
static PyTypeObject *__pyx_ptype_3lib_Task = 0;
static PyTypeObject *__pyx_ptype_3lib_RedisParser = 0;
static PyTypeObject *__pyx_ptype_3lib_ArrayTask = 0;
static PyTypeObject *__pyx_ptype_3lib_Frame = 0;
static PyTypeObject *__pyx_ptype_3lib_FrameParser = 0;
static PyTypeObject *__pyx_ptype_3lib___pyx_scope_struct__pack_pipeline = 0;
static PyTypeObject *__pyx_ptype_3lib___pyx_scope_struct_1_genexpr = 0;
static PyTypeObject *__pyx_ptype_3lib___pyx_scope_struct_2__pack_command = 0;
static PyTypeObject *__pyx_ptype_3lib___pyx_scope_struct_3__pack = 0;
static PyTypeObject *__pyx_ptype_3lib___pyx_scope_struct_4__lua_dict = 0;
static PyTypeObject *__pyx_ptype_3lib___pyx_scope_struct_5_multi_encode = 0;
static double __pyx_v_3lib_clog2;
static PyObject *__pyx_v_3lib_CRLF = 0;
static PyObject *__pyx_v_3lib_RESPONSE_INTEGER = 0;
static PyObject *__pyx_v_3lib_RESPONSE_STRING = 0;
static PyObject *__pyx_v_3lib_RESPONSE_ARRAY = 0;
static PyObject *__pyx_v_3lib_RESPONSE_STATUS = 0;
static PyObject *__pyx_v_3lib_RESPONSE_ERROR = 0;
static PyObject *__pyx_v_3lib_nil = 0;
static PyObject *__pyx_v_3lib_null_array = 0;
#define __Pyx_MODULE_NAME "lib"
int __pyx_module_is_main_lib = 0;

/* Implementation of 'lib' */
static PyObject *__pyx_builtin_TypeError;
static const char __pyx_k_[] = "";
static const char __pyx_k_1[] = "$-1\r\n";
static const char __pyx_k_H[] = "!H";
static const char __pyx_k_Q[] = "!Q";
static const char __pyx_k_d[] = "$%d\r\n";
static const char __pyx_k_s[] = "*%s\r\n";
static const char __pyx_k_BB[] = "BB";
static const char __pyx_k__3[] = "\r\n";
static const char __pyx_k_os[] = "os";
static const char __pyx_k_1_2[] = "*-1\r\n";
static const char __pyx_k__30[] = ":";
static const char __pyx_k__31[] = "$";
static const char __pyx_k__32[] = "*";
static const char __pyx_k__33[] = "+";
static const char __pyx_k__34[] = "-";
static const char __pyx_k_d_2[] = "*%d\r\n";
static const char __pyx_k_get[] = "get";
static const char __pyx_k_len[] = "__len__";
static const char __pyx_k_lib[] = "lib";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_body[] = "body";
static const char __pyx_k_code[] = "code";
static const char __pyx_k_data[] = "data";
static const char __pyx_k_find[] = "find";
static const char __pyx_k_join[] = "join";
static const char __pyx_k_kind[] = "kind";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_next[] = "next";
static const char __pyx_k_pack[] = "_pack";
static const char __pyx_k_perr[] = "perr";
static const char __pyx_k_rerr[] = "rerr";
static const char __pyx_k_rsv1[] = "rsv1";
static const char __pyx_k_rsv2[] = "rsv2";
static const char __pyx_k_rsv3[] = "rsv3";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_clog2[] = "clog2";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_final[] = "final";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_utf_8[] = "utf-8";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_append[] = "append";
static const char __pyx_k_decode[] = "decode";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_extend[] = "extend";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_length[] = "length";
static const char __pyx_k_opcode[] = "opcode";
static const char __pyx_k_pack_2[] = "pack";
static const char __pyx_k_struct[] = "struct";
static const char __pyx_k_unpack[] = "unpack";
static const char __pyx_k_genexpr[] = "genexpr";
static const char __pyx_k_message[] = "message";
static const char __pyx_k_receive[] = "receive";
static const char __pyx_k_starmap[] = "starmap";
static const char __pyx_k_urandom[] = "urandom";
static const char __pyx_k_version[] = "version";
static const char __pyx_k_encoding[] = "encoding";
static const char __pyx_k_lua_dict[] = "_lua_dict";
static const char __pyx_k_pyx_capi[] = "__pyx_capi__";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_itertools[] = "itertools";
static const char __pyx_k_protocols[] = "protocols";
static const char __pyx_k_extensions[] = "extensions";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_close_codes[] = "close_codes";
static const char __pyx_k_masking_key[] = "masking_key";
static const char __pyx_k_max_payload[] = "max_payload";
static const char __pyx_k_multi_encode[] = "multi_encode";
static const char __pyx_k_pack_command[] = "_pack_command";
static const char __pyx_k_ProtocolError[] = "ProtocolError";
static const char __pyx_k_Protocol_Error[] = "Protocol Error";
static const char __pyx_k_payload_length[] = "payload_length";
static const char __pyx_k_bad_masking_key[] = "bad masking key";
static const char __pyx_k_decode_responses[] = "decode_responses";
static const char __pyx_k_RedisParser__pack[] = "RedisParser._pack";
static const char __pyx_k_FIN_must_be_0_or_1[] = "FIN must be 0 or 1";
static const char __pyx_k_masked_server_frame[] = "masked server frame.";
static const char __pyx_k_RedisParser__lua_dict[] = "RedisParser._lua_dict";
static const char __pyx_k_unmasked_client_frame[] = "unmasked client frame.";
static const char __pyx_k_FrameParser_multi_encode[] = "FrameParser.multi_encode";
static const char __pyx_k_Requires_bytes_or_string[] = "Requires bytes or string";
static const char __pyx_k_RedisParser__pack_command[] = "RedisParser._pack_command";
static const char __pyx_k_WEBSOCKET_frame_too_large[] = "WEBSOCKET frame too large";
static const char __pyx_k_pack_pipeline_locals_lambda[] = "pack_pipeline.<locals>.<lambda>";
static const char __pyx_k_pack_pipeline_locals_genexpr[] = "pack_pipeline.<locals>.genexpr";
static const char __pyx_k_WEBSOCKET_control_frame_fragment[] = "WEBSOCKET control frame fragmented";
static const char __pyx_k_WEBSOCKET_control_frame_too_larg[] = "WEBSOCKET control frame too large";
static const char __pyx_k_WEBSOCKET_opcode_a_reserved_valu[] = "WEBSOCKET opcode a reserved value";
static PyObject *__pyx_kp_b_;
static PyObject *__pyx_kp_s_;
static PyObject *__pyx_kp_b_1;
static PyObject *__pyx_kp_b_1_2;
static PyObject *__pyx_n_s_BB;
static PyObject *__pyx_kp_s_FIN_must_be_0_or_1;
static PyObject *__pyx_n_s_FrameParser_multi_encode;
static PyObject *__pyx_kp_s_H;
static PyObject *__pyx_n_s_ProtocolError;
static PyObject *__pyx_kp_s_Protocol_Error;
static PyObject *__pyx_kp_s_Q;
static PyObject *__pyx_n_s_RedisParser__lua_dict;
static PyObject *__pyx_n_s_RedisParser__pack;
static PyObject *__pyx_n_s_RedisParser__pack_command;
static PyObject *__pyx_kp_s_Requires_bytes_or_string;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_kp_s_WEBSOCKET_control_frame_fragment;
static PyObject *__pyx_kp_s_WEBSOCKET_control_frame_too_larg;
static PyObject *__pyx_kp_s_WEBSOCKET_frame_too_large;
static PyObject *__pyx_kp_s_WEBSOCKET_opcode_a_reserved_valu;
static PyObject *__pyx_kp_b__3;
static PyObject *__pyx_kp_b__30;
static PyObject *__pyx_kp_b__31;
static PyObject *__pyx_kp_b__32;
static PyObject *__pyx_kp_b__33;
static PyObject *__pyx_kp_b__34;
static PyObject *__pyx_n_s_append;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_kp_s_bad_masking_key;
static PyObject *__pyx_n_s_body;
static PyObject *__pyx_n_s_clog2;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_close_codes;
static PyObject *__pyx_n_s_code;
static PyObject *__pyx_kp_s_d;
static PyObject *__pyx_kp_s_d_2;
static PyObject *__pyx_n_s_data;
static PyObject *__pyx_n_s_decode;
static PyObject *__pyx_n_s_decode_responses;
static PyObject *__pyx_n_s_encode;
static PyObject *__pyx_n_s_encoding;
static PyObject *__pyx_n_s_extend;
static PyObject *__pyx_n_s_extensions;
static PyObject *__pyx_n_s_final;
static PyObject *__pyx_n_s_find;
static PyObject *__pyx_n_s_genexpr;
static PyObject *__pyx_n_s_get;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_items;
static PyObject *__pyx_n_s_itertools;
static PyObject *__pyx_n_s_join;
static PyObject *__pyx_n_s_kind;
static PyObject *__pyx_n_s_len;
static PyObject *__pyx_n_s_length;
static PyObject *__pyx_n_s_lib;
static PyObject *__pyx_n_s_lua_dict;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_kp_s_masked_server_frame;
static PyObject *__pyx_n_s_masking_key;
static PyObject *__pyx_n_s_max_payload;
static PyObject *__pyx_n_s_message;
static PyObject *__pyx_n_s_multi_encode;
static PyObject *__pyx_n_s_next;
static PyObject *__pyx_n_s_opcode;
static PyObject *__pyx_n_s_os;
static PyObject *__pyx_n_s_pack;
static PyObject *__pyx_n_s_pack_2;
static PyObject *__pyx_n_s_pack_command;
static PyObject *__pyx_n_s_pack_pipeline_locals_genexpr;
static PyObject *__pyx_n_s_pack_pipeline_locals_lambda;
static PyObject *__pyx_n_s_payload_length;
static PyObject *__pyx_n_s_perr;
static PyObject *__pyx_n_s_protocols;
static PyObject *__pyx_n_s_pyx_capi;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_receive;
static PyObject *__pyx_n_s_rerr;
static PyObject *__pyx_n_s_rsv1;
static PyObject *__pyx_n_s_rsv2;
static PyObject *__pyx_n_s_rsv3;
static PyObject *__pyx_kp_s_s;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_starmap;
static PyObject *__pyx_n_s_struct;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_kp_s_unmasked_client_frame;
static PyObject *__pyx_n_s_unpack;
static PyObject *__pyx_n_s_urandom;
static PyObject *__pyx_kp_s_utf_8;
static PyObject *__pyx_n_s_value;
static PyObject *__pyx_n_s_version;
static int __pyx_pf_3lib_11RedisParser___cinit__(struct __pyx_obj_3lib_RedisParser *__pyx_v_self, PyObject *__pyx_v_perr, PyObject *__pyx_v_rerr); /* proto */
static PyObject *__pyx_pf_3lib_11RedisParser_2on_connect(struct __pyx_obj_3lib_RedisParser *__pyx_v_self, PyObject *__pyx_v_connection); /* proto */
static PyObject *__pyx_pf_3lib_11RedisParser_4on_disconnect(CYTHON_UNUSED struct __pyx_obj_3lib_RedisParser *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_3lib_11RedisParser_6get(struct __pyx_obj_3lib_RedisParser *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_3lib_11RedisParser_8feed(struct __pyx_obj_3lib_RedisParser *__pyx_v_self, PyObject *__pyx_v_stream); /* proto */
static PyObject *__pyx_pf_3lib_11RedisParser_10buffer(struct __pyx_obj_3lib_RedisParser *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_3lib_11RedisParser_12pack_command(struct __pyx_obj_3lib_RedisParser *__pyx_v_self, PyObject *__pyx_v_args); /* proto */
static PyObject *__pyx_lambda_funcdef_lambda(PyObject *__pyx_self, PyObject *__pyx_v_args); /* proto */
static PyObject *__pyx_pf_3lib_11RedisParser_13pack_pipeline_1genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_3lib_11RedisParser_14pack_pipeline(struct __pyx_obj_3lib_RedisParser *__pyx_v_self, PyObject *__pyx_v_commands); /* proto */
static PyObject *__pyx_pf_3lib_11RedisParser_16bulk(CYTHON_UNUSED struct __pyx_obj_3lib_RedisParser *__pyx_v_self, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf_3lib_11RedisParser_18multi_bulk_len(CYTHON_UNUSED struct __pyx_obj_3lib_RedisParser *__pyx_v_self, PyObject *__pyx_v_len); /* proto */
static PyObject *__pyx_pf_3lib_11RedisParser_20multi_bulk(struct __pyx_obj_3lib_RedisParser *__pyx_v_self, PyObject *__pyx_v_args); /* proto */
static PyObject *__pyx_pf_3lib_11RedisParser_22_pack_command(CYTHON_UNUSED struct __pyx_obj_3lib_RedisParser *__pyx_v_self, PyObject *__pyx_v_args); /* proto */
static PyObject *__pyx_pf_3lib_11RedisParser_25_pack(struct __pyx_obj_3lib_RedisParser *__pyx_v_self, PyObject *__pyx_v_args); /* proto */
static PyObject *__pyx_pf_3lib_11RedisParser_28_lua_dict(CYTHON_UNUSED struct __pyx_obj_3lib_RedisParser *__pyx_v_self, PyObject *__pyx_v_d); /* proto */
static int __pyx_pf_3lib_4Task___cinit__(struct __pyx_obj_3lib_Task *__pyx_v_self, long __pyx_v_length, struct __pyx_obj_3lib_Task *__pyx_v_next); /* proto */
static int __pyx_pf_3lib_5Frame___cinit__(struct __pyx_obj_3lib_Frame *__pyx_v_self, int __pyx_v_opcode, int __pyx_v_final, int __pyx_v_payload_length); /* proto */
static PyObject *__pyx_pf_3lib_5Frame_5final___get__(struct __pyx_obj_3lib_Frame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_3lib_5Frame_6opcode___get__(struct __pyx_obj_3lib_Frame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_3lib_5Frame_4body___get__(struct __pyx_obj_3lib_Frame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_3lib_5Frame_11masking_key___get__(struct __pyx_obj_3lib_Frame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_3lib_5Frame_10is_message___get__(struct __pyx_obj_3lib_Frame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_3lib_5Frame_8is_bytes___get__(struct __pyx_obj_3lib_Frame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_3lib_5Frame_8is_close___get__(struct __pyx_obj_3lib_Frame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_3lib_5Frame_7is_ping___get__(struct __pyx_obj_3lib_Frame *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_3lib_5Frame_7is_pong___get__(struct __pyx_obj_3lib_Frame *__pyx_v_self); /* proto */
static int __pyx_pf_3lib_11FrameParser___cinit__(struct __pyx_obj_3lib_FrameParser *__pyx_v_self, int __pyx_v_version, int __pyx_v_kind, PyObject *__pyx_v_ProtocolError, PyObject *__pyx_v_extensions, PyObject *__pyx_v_protocols, PyObject *__pyx_v_close_codes); /* proto */
static PyObject *__pyx_pf_3lib_11FrameParser_7version___get__(struct __pyx_obj_3lib_FrameParser *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_3lib_11FrameParser_11max_payload___get__(CYTHON_UNUSED struct __pyx_obj_3lib_FrameParser *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_3lib_11FrameParser_18decode_mask_length___get__(struct __pyx_obj_3lib_FrameParser *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_3lib_11FrameParser_18encode_mask_length___get__(struct __pyx_obj_3lib_FrameParser *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_3lib_11FrameParser_10extensions___get__(struct __pyx_obj_3lib_FrameParser *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_3lib_11FrameParser_9protocols___get__(struct __pyx_obj_3lib_FrameParser *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_3lib_11FrameParser_2ping(struct __pyx_obj_3lib_FrameParser *__pyx_v_self, PyObject *__pyx_v_body); /* proto */
static PyObject *__pyx_pf_3lib_11FrameParser_4pong(struct __pyx_obj_3lib_FrameParser *__pyx_v_self, PyObject *__pyx_v_body); /* proto */
static PyObject *__pyx_pf_3lib_11FrameParser_6close(struct __pyx_obj_3lib_FrameParser *__pyx_v_self, PyObject *__pyx_v_code); /* proto */
static PyObject *__pyx_pf_3lib_11FrameParser_8continuation(struct __pyx_obj_3lib_FrameParser *__pyx_v_self, PyObject *__pyx_v_body, PyObject *__pyx_v_final); /* proto */
static PyObject *__pyx_pf_3lib_11FrameParser_10encode(struct __pyx_obj_3lib_FrameParser *__pyx_v_self, PyObject *__pyx_v_message, PyObject *__pyx_v_final, PyObject *__pyx_v_masking_key, PyObject *__pyx_v_opcode, int __pyx_v_rsv1, int __pyx_v_rsv2, int __pyx_v_rsv3); /* proto */
static PyObject *__pyx_pf_3lib_11FrameParser_12multi_encode(struct __pyx_obj_3lib_FrameParser *__pyx_v_self, PyObject *__pyx_v_message, PyObject *__pyx_v_masking_key, PyObject *__pyx_v_opcode, int __pyx_v_rsv1, int __pyx_v_rsv2, int __pyx_v_rsv3, unsigned PY_LONG_LONG __pyx_v_max_payload); /* proto */
static PyObject *__pyx_pf_3lib_11FrameParser_15decode(struct __pyx_obj_3lib_FrameParser *__pyx_v_self, PyObject *__pyx_v_data); /* proto */
static PyObject *__pyx_tp_new_3lib_Task(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_3lib_RedisParser(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_3lib_ArrayTask(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_3lib_Frame(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_3lib_FrameParser(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_3lib___pyx_scope_struct__pack_pipeline(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_3lib___pyx_scope_struct_1_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_3lib___pyx_scope_struct_2__pack_command(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_3lib___pyx_scope_struct_3__pack(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_3lib___pyx_scope_struct_4__lua_dict(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_3lib___pyx_scope_struct_5_multi_encode(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_4;
static PyObject *__pyx_int_5;
static PyObject *__pyx_int_6;
static PyObject *__pyx_int_7;
static PyObject *__pyx_int_8;
static PyObject *__pyx_int_9;
static PyObject *__pyx_int_10;
static PyObject *__pyx_int_15;
static PyObject *__pyx_int_127;
static PyObject *__pyx_int_128;
static PyObject *__pyx_int_9223372036854775808;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_slice__19;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__29;

/* "extensions/lib/rparser.pyx":22
 *     cdef Task _current
 * 
 *     def __cinit__(self, object perr, object rerr):             # <<<<<<<<<<<<<<
 *         self._protocolError = perr
 *         self._responseError = rerr
 */

/* Python wrapper */
static int __pyx_pw_3lib_11RedisParser_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_3lib_11RedisParser_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_perr = 0;
  PyObject *__pyx_v_rerr = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_perr,&__pyx_n_s_rerr,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_perr)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_rerr)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, 1); __PYX_ERR(0, 22, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 22, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_perr = values[0];
    __pyx_v_rerr = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 22, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("lib.RedisParser.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_3lib_11RedisParser___cinit__(((struct __pyx_obj_3lib_RedisParser *)__pyx_v_self), __pyx_v_perr, __pyx_v_rerr);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_3lib_11RedisParser___cinit__(struct __pyx_obj_3lib_RedisParser *__pyx_v_self, PyObject *__pyx_v_perr, PyObject *__pyx_v_rerr) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "extensions/lib/rparser.pyx":23
 * 
 *     def __cinit__(self, object perr, object rerr):
 *         self._protocolError = perr             # <<<<<<<<<<<<<<
 *         self._responseError = rerr
 *         self._inbuffer = bytearray()
 */
  __Pyx_INCREF(__pyx_v_perr);
  __Pyx_GIVEREF(__pyx_v_perr);
  __Pyx_GOTREF(__pyx_v_self->_protocolError);
  __Pyx_DECREF(__pyx_v_self->_protocolError);
  __pyx_v_self->_protocolError = __pyx_v_perr;

  /* "extensions/lib/rparser.pyx":24
 *     def __cinit__(self, object perr, object rerr):
 *         self._protocolError = perr
 *         self._responseError = rerr             # <<<<<<<<<<<<<<
 *         self._inbuffer = bytearray()
 * 
 */
  __Pyx_INCREF(__pyx_v_rerr);
  __Pyx_GIVEREF(__pyx_v_rerr);
  __Pyx_GOTREF(__pyx_v_self->_responseError);
  __Pyx_DECREF(__pyx_v_self->_responseError);
  __pyx_v_self->_responseError = __pyx_v_rerr;

  /* "extensions/lib/rparser.pyx":25
 *         self._protocolError = perr
 *         self._responseError = rerr
 *         self._inbuffer = bytearray()             # <<<<<<<<<<<<<<
 * 
 *     def on_connect(self, connection):
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)(&PyByteArray_Type)), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_inbuffer);
  __Pyx_DECREF(__pyx_v_self->_inbuffer);
  __pyx_v_self->_inbuffer = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "extensions/lib/rparser.pyx":22
 *     cdef Task _current
 * 
 *     def __cinit__(self, object perr, object rerr):             # <<<<<<<<<<<<<<
 *         self._protocolError = perr
 *         self._responseError = rerr
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("lib.RedisParser.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/rparser.pyx":27
 *         self._inbuffer = bytearray()
 * 
 *     def on_connect(self, connection):             # <<<<<<<<<<<<<<
 *         if connection.decode_responses:
 *             self._encoding = connection.encoding
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_11RedisParser_3on_connect(PyObject *__pyx_v_self, PyObject *__pyx_v_connection); /*proto*/
static PyObject *__pyx_pw_3lib_11RedisParser_3on_connect(PyObject *__pyx_v_self, PyObject *__pyx_v_connection) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("on_connect (wrapper)", 0);
  __pyx_r = __pyx_pf_3lib_11RedisParser_2on_connect(((struct __pyx_obj_3lib_RedisParser *)__pyx_v_self), ((PyObject *)__pyx_v_connection));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_11RedisParser_2on_connect(struct __pyx_obj_3lib_RedisParser *__pyx_v_self, PyObject *__pyx_v_connection) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("on_connect", 0);

  /* "extensions/lib/rparser.pyx":28
 * 
 *     def on_connect(self, connection):
 *         if connection.decode_responses:             # <<<<<<<<<<<<<<
 *             self._encoding = connection.encoding
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_connection, __pyx_n_s_decode_responses); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "extensions/lib/rparser.pyx":29
 *     def on_connect(self, connection):
 *         if connection.decode_responses:
 *             self._encoding = connection.encoding             # <<<<<<<<<<<<<<
 * 
 *     def on_disconnect(self):
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_connection, __pyx_n_s_encoding); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 29, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->_encoding);
    __Pyx_DECREF(__pyx_v_self->_encoding);
    __pyx_v_self->_encoding = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "extensions/lib/rparser.pyx":28
 * 
 *     def on_connect(self, connection):
 *         if connection.decode_responses:             # <<<<<<<<<<<<<<
 *             self._encoding = connection.encoding
 * 
 */
  }

  /* "extensions/lib/rparser.pyx":27
 *         self._inbuffer = bytearray()
 * 
 *     def on_connect(self, connection):             # <<<<<<<<<<<<<<
 *         if connection.decode_responses:
 *             self._encoding = connection.encoding
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("lib.RedisParser.on_connect", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/rparser.pyx":31
 *             self._encoding = connection.encoding
 * 
 *     def on_disconnect(self):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_11RedisParser_5on_disconnect(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_3lib_11RedisParser_5on_disconnect(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("on_disconnect (wrapper)", 0);
  __pyx_r = __pyx_pf_3lib_11RedisParser_4on_disconnect(((struct __pyx_obj_3lib_RedisParser *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_11RedisParser_4on_disconnect(CYTHON_UNUSED struct __pyx_obj_3lib_RedisParser *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("on_disconnect", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/rparser.pyx":35
 * 
 *     # DECODER
 *     def get(self):             # <<<<<<<<<<<<<<
 *         if self._current:
 *             return self._resume(self._current, False)
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_11RedisParser_7get(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_3lib_11RedisParser_7get(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get (wrapper)", 0);
  __pyx_r = __pyx_pf_3lib_11RedisParser_6get(((struct __pyx_obj_3lib_RedisParser *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_11RedisParser_6get(struct __pyx_obj_3lib_RedisParser *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("get", 0);

  /* "extensions/lib/rparser.pyx":36
 *     # DECODER
 *     def get(self):
 *         if self._current:             # <<<<<<<<<<<<<<
 *             return self._resume(self._current, False)
 *         else:
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_self->_current)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 36, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "extensions/lib/rparser.pyx":37
 *     def get(self):
 *         if self._current:
 *             return self._resume(self._current, False)             # <<<<<<<<<<<<<<
 *         else:
 *             return self._get(None)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = ((PyObject *)__pyx_v_self->_current);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_3 = ((struct __pyx_vtabstruct_3lib_RedisParser *)__pyx_v_self->__pyx_vtab)->_resume(__pyx_v_self, ((struct __pyx_obj_3lib_Task *)__pyx_t_2), Py_False); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 37, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "extensions/lib/rparser.pyx":36
 *     # DECODER
 *     def get(self):
 *         if self._current:             # <<<<<<<<<<<<<<
 *             return self._resume(self._current, False)
 *         else:
 */
  }

  /* "extensions/lib/rparser.pyx":39
 *             return self._resume(self._current, False)
 *         else:
 *             return self._get(None)             # <<<<<<<<<<<<<<
 * 
 *     def feed(self, stream):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = ((struct __pyx_vtabstruct_3lib_RedisParser *)__pyx_v_self->__pyx_vtab)->_get(__pyx_v_self, ((struct __pyx_obj_3lib_Task *)Py_None)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 39, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "extensions/lib/rparser.pyx":35
 * 
 *     # DECODER
 *     def get(self):             # <<<<<<<<<<<<<<
 *         if self._current:
 *             return self._resume(self._current, False)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("lib.RedisParser.get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/rparser.pyx":41
 *             return self._get(None)
 * 
 *     def feed(self, stream):             # <<<<<<<<<<<<<<
 *         self._inbuffer.extend(stream)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_11RedisParser_9feed(PyObject *__pyx_v_self, PyObject *__pyx_v_stream); /*proto*/
static PyObject *__pyx_pw_3lib_11RedisParser_9feed(PyObject *__pyx_v_self, PyObject *__pyx_v_stream) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("feed (wrapper)", 0);
  __pyx_r = __pyx_pf_3lib_11RedisParser_8feed(((struct __pyx_obj_3lib_RedisParser *)__pyx_v_self), ((PyObject *)__pyx_v_stream));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_11RedisParser_8feed(struct __pyx_obj_3lib_RedisParser *__pyx_v_self, PyObject *__pyx_v_stream) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("feed", 0);

  /* "extensions/lib/rparser.pyx":42
 * 
 *     def feed(self, stream):
 *         self._inbuffer.extend(stream)             # <<<<<<<<<<<<<<
 * 
 *     def buffer(self):
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_inbuffer, __pyx_n_s_extend); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (!__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_stream); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 42, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 42, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
    __Pyx_INCREF(__pyx_v_stream);
    __Pyx_GIVEREF(__pyx_v_stream);
    PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_stream);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 42, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "extensions/lib/rparser.pyx":41
 *             return self._get(None)
 * 
 *     def feed(self, stream):             # <<<<<<<<<<<<<<
 *         self._inbuffer.extend(stream)
 * 
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("lib.RedisParser.feed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/rparser.pyx":44
 *         self._inbuffer.extend(stream)
 * 
 *     def buffer(self):             # <<<<<<<<<<<<<<
 *         return self._inbuffer
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_11RedisParser_11buffer(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_3lib_11RedisParser_11buffer(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("buffer (wrapper)", 0);
  __pyx_r = __pyx_pf_3lib_11RedisParser_10buffer(((struct __pyx_obj_3lib_RedisParser *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_11RedisParser_10buffer(struct __pyx_obj_3lib_RedisParser *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("buffer", 0);

  /* "extensions/lib/rparser.pyx":45
 * 
 *     def buffer(self):
 *         return self._inbuffer             # <<<<<<<<<<<<<<
 * 
 *     # CLIENT ENCODERS
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_inbuffer);
  __pyx_r = __pyx_v_self->_inbuffer;
  goto __pyx_L0;

  /* "extensions/lib/rparser.pyx":44
 *         self._inbuffer.extend(stream)
 * 
 *     def buffer(self):             # <<<<<<<<<<<<<<
 *         return self._inbuffer
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/rparser.pyx":48
 * 
 *     # CLIENT ENCODERS
 *     def pack_command(self, args):             # <<<<<<<<<<<<<<
 *         return b''.join(self._pack_command(args))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_11RedisParser_13pack_command(PyObject *__pyx_v_self, PyObject *__pyx_v_args); /*proto*/
static PyObject *__pyx_pw_3lib_11RedisParser_13pack_command(PyObject *__pyx_v_self, PyObject *__pyx_v_args) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pack_command (wrapper)", 0);
  __pyx_r = __pyx_pf_3lib_11RedisParser_12pack_command(((struct __pyx_obj_3lib_RedisParser *)__pyx_v_self), ((PyObject *)__pyx_v_args));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_11RedisParser_12pack_command(struct __pyx_obj_3lib_RedisParser *__pyx_v_self, PyObject *__pyx_v_args) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("pack_command", 0);

  /* "extensions/lib/rparser.pyx":49
 *     # CLIENT ENCODERS
 *     def pack_command(self, args):
 *         return b''.join(self._pack_command(args))             # <<<<<<<<<<<<<<
 * 
 *     def pack_pipeline(self, commands):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pack_command); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (!__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_args); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 49, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 49, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
    __Pyx_INCREF(__pyx_v_args);
    __Pyx_GIVEREF(__pyx_v_args);
    PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_args);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 49, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyBytes_Join(__pyx_kp_b_, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "extensions/lib/rparser.pyx":48
 * 
 *     # CLIENT ENCODERS
 *     def pack_command(self, args):             # <<<<<<<<<<<<<<
 *         return b''.join(self._pack_command(args))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("lib.RedisParser.pack_command", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/rparser.pyx":51
 *         return b''.join(self._pack_command(args))
 * 
 *     def pack_pipeline(self, commands):             # <<<<<<<<<<<<<<
 *         pack = lambda *args: b''.join(self._pack_command(args))
 *         return b''.join(starmap(pack, (args for args, _ in commands)))
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_11RedisParser_15pack_pipeline(PyObject *__pyx_v_self, PyObject *__pyx_v_commands); /*proto*/
static PyObject *__pyx_pw_3lib_11RedisParser_15pack_pipeline(PyObject *__pyx_v_self, PyObject *__pyx_v_commands) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pack_pipeline (wrapper)", 0);
  __pyx_r = __pyx_pf_3lib_11RedisParser_14pack_pipeline(((struct __pyx_obj_3lib_RedisParser *)__pyx_v_self), ((PyObject *)__pyx_v_commands));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/rparser.pyx":52
 * 
 *     def pack_pipeline(self, commands):
 *         pack = lambda *args: b''.join(self._pack_command(args))             # <<<<<<<<<<<<<<
 *         return b''.join(starmap(pack, (args for args, _ in commands)))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_11RedisParser_13pack_pipeline_lambda(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_3lib_11RedisParser_13pack_pipeline_lambda = {"lambda", (PyCFunction)__pyx_pw_3lib_11RedisParser_13pack_pipeline_lambda, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_3lib_11RedisParser_13pack_pipeline_lambda(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_args = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("lambda (wrapper)", 0);
  if (unlikely(__pyx_kwds) && unlikely(PyDict_Size(__pyx_kwds) > 0) && unlikely(!__Pyx_CheckKeywordStrings(__pyx_kwds, "lambda", 0))) return NULL;
  __Pyx_INCREF(__pyx_args);
  __pyx_v_args = __pyx_args;
  __pyx_r = __pyx_lambda_funcdef_lambda(__pyx_self, __pyx_v_args);

  /* function exit code */
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_lambda_funcdef_lambda(PyObject *__pyx_self, PyObject *__pyx_v_args) {
  struct __pyx_obj_3lib___pyx_scope_struct__pack_pipeline *__pyx_cur_scope;
  struct __pyx_obj_3lib___pyx_scope_struct__pack_pipeline *__pyx_outer_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("lambda", 0);
  __pyx_outer_scope = (struct __pyx_obj_3lib___pyx_scope_struct__pack_pipeline *) __Pyx_CyFunction_GetClosure(__pyx_self);
  __pyx_cur_scope = __pyx_outer_scope;
  __Pyx_XDECREF(__pyx_r);
  if (unlikely(!__pyx_cur_scope->__pyx_v_self)) { __Pyx_RaiseClosureNameError("self"); __PYX_ERR(0, 52, __pyx_L1_error) }
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_pack_command); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (!__pyx_t_3) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_args); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 52, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 52, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
    __Pyx_INCREF(__pyx_v_args);
    __Pyx_GIVEREF(__pyx_v_args);
    PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_v_args);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 52, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyBytes_Join(__pyx_kp_b_, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("lib.RedisParser.pack_pipeline.lambda", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_3lib_11RedisParser_13pack_pipeline_3generator4(__pyx_CoroutineObject *__pyx_generator, PyObject *__pyx_sent_value); /* proto */

/* "extensions/lib/rparser.pyx":53
 *     def pack_pipeline(self, commands):
 *         pack = lambda *args: b''.join(self._pack_command(args))
 *         return b''.join(starmap(pack, (args for args, _ in commands)))             # <<<<<<<<<<<<<<
 * 
 *     # SERVER ENCODERS
 */

static PyObject *__pyx_pf_3lib_11RedisParser_13pack_pipeline_1genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_3lib___pyx_scope_struct_1_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_3lib___pyx_scope_struct_1_genexpr *)__pyx_tp_new_3lib___pyx_scope_struct_1_genexpr(__pyx_ptype_3lib___pyx_scope_struct_1_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __Pyx_GOTREF(__pyx_cur_scope);
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_3lib___pyx_scope_struct__pack_pipeline *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_3lib_11RedisParser_13pack_pipeline_3generator4, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_pack_pipeline_locals_genexpr); if (unlikely(!gen)) __PYX_ERR(0, 53, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("lib.RedisParser.pack_pipeline.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_3lib_11RedisParser_13pack_pipeline_3generator4(__pyx_CoroutineObject *__pyx_generator, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_3lib___pyx_scope_struct_1_genexpr *__pyx_cur_scope = ((struct __pyx_obj_3lib___pyx_scope_struct_1_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *(*__pyx_t_8)(PyObject *);
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("None", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L8_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 53, __pyx_L1_error)
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_commands)) { __Pyx_RaiseClosureNameError("commands"); __PYX_ERR(0, 53, __pyx_L1_error) }
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_commands)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_commands)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_commands; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_commands); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 53, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 53, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 53, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 53, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 53, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 53, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 53, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
      PyObject* sequence = __pyx_t_4;
      #if CYTHON_COMPILING_IN_CPYTHON
      Py_ssize_t size = Py_SIZE(sequence);
      #else
      Py_ssize_t size = PySequence_Size(sequence);
      #endif
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 53, __pyx_L1_error)
      }
      #if CYTHON_COMPILING_IN_CPYTHON
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_5 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_5 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 53, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 53, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_7 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 53, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_8 = Py_TYPE(__pyx_t_7)->tp_iternext;
      index = 0; __pyx_t_5 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_5)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 1; __pyx_t_6 = __pyx_t_8(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_8(__pyx_t_7), 2) < 0) __PYX_ERR(0, 53, __pyx_L1_error)
      __pyx_t_8 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L7_unpacking_done;
      __pyx_L6_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_8 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 53, __pyx_L1_error)
      __pyx_L7_unpacking_done:;
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_args);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_args, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v__);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v__, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_args);
    __pyx_r = __pyx_cur_scope->__pyx_v_args;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L8_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 53, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/rparser.pyx":51
 *         return b''.join(self._pack_command(args))
 * 
 *     def pack_pipeline(self, commands):             # <<<<<<<<<<<<<<
 *         pack = lambda *args: b''.join(self._pack_command(args))
 *         return b''.join(starmap(pack, (args for args, _ in commands)))
 */

static PyObject *__pyx_pf_3lib_11RedisParser_14pack_pipeline(struct __pyx_obj_3lib_RedisParser *__pyx_v_self, PyObject *__pyx_v_commands) {
  struct __pyx_obj_3lib___pyx_scope_struct__pack_pipeline *__pyx_cur_scope;
  PyObject *__pyx_v_pack = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("pack_pipeline", 0);
  __pyx_cur_scope = (struct __pyx_obj_3lib___pyx_scope_struct__pack_pipeline *)__pyx_tp_new_3lib___pyx_scope_struct__pack_pipeline(__pyx_ptype_3lib___pyx_scope_struct__pack_pipeline, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __Pyx_GOTREF(__pyx_cur_scope);
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_commands = __pyx_v_commands;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_commands);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_commands);

  /* "extensions/lib/rparser.pyx":52
 * 
 *     def pack_pipeline(self, commands):
 *         pack = lambda *args: b''.join(self._pack_command(args))             # <<<<<<<<<<<<<<
 *         return b''.join(starmap(pack, (args for args, _ in commands)))
 * 
 */
  __pyx_t_1 = __Pyx_CyFunction_NewEx(&__pyx_mdef_3lib_11RedisParser_13pack_pipeline_lambda, 0, __pyx_n_s_pack_pipeline_locals_lambda, ((PyObject*)__pyx_cur_scope), __pyx_n_s_lib, __pyx_d, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_pack = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "extensions/lib/rparser.pyx":53
 *     def pack_pipeline(self, commands):
 *         pack = lambda *args: b''.join(self._pack_command(args))
 *         return b''.join(starmap(pack, (args for args, _ in commands)))             # <<<<<<<<<<<<<<
 * 
 *     # SERVER ENCODERS
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_starmap); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __pyx_pf_3lib_11RedisParser_13pack_pipeline_1genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_5 = 1;
    }
  }
  __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (__pyx_t_4) {
    __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
  }
  __Pyx_INCREF(__pyx_v_pack);
  __Pyx_GIVEREF(__pyx_v_pack);
  PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_v_pack);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_t_3);
  __pyx_t_3 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyBytes_Join(__pyx_kp_b_, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "extensions/lib/rparser.pyx":51
 *         return b''.join(self._pack_command(args))
 * 
 *     def pack_pipeline(self, commands):             # <<<<<<<<<<<<<<
 *         pack = lambda *args: b''.join(self._pack_command(args))
 *         return b''.join(starmap(pack, (args for args, _ in commands)))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("lib.RedisParser.pack_pipeline", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_pack);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/rparser.pyx":56
 * 
 *     # SERVER ENCODERS
 *     def bulk(self, bytes value=None):             # <<<<<<<<<<<<<<
 *         if value is None:
 *             return nil
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_11RedisParser_17bulk(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_3lib_11RedisParser_17bulk(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_value = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("bulk (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_value,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject*)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_value);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "bulk") < 0)) __PYX_ERR(0, 56, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_value = ((PyObject*)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("bulk", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 56, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("lib.RedisParser.bulk", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_value), (&PyBytes_Type), 1, "value", 1))) __PYX_ERR(0, 56, __pyx_L1_error)
  __pyx_r = __pyx_pf_3lib_11RedisParser_16bulk(((struct __pyx_obj_3lib_RedisParser *)__pyx_v_self), __pyx_v_value);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_11RedisParser_16bulk(CYTHON_UNUSED struct __pyx_obj_3lib_RedisParser *__pyx_v_self, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("bulk", 0);

  /* "extensions/lib/rparser.pyx":57
 *     # SERVER ENCODERS
 *     def bulk(self, bytes value=None):
 *         if value is None:             # <<<<<<<<<<<<<<
 *             return nil
 *         else:
 */
  __pyx_t_1 = (__pyx_v_value == ((PyObject*)Py_None));
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "extensions/lib/rparser.pyx":58
 *     def bulk(self, bytes value=None):
 *         if value is None:
 *             return nil             # <<<<<<<<<<<<<<
 *         else:
 *             return ('$%d\r\n' % len(value)).encode('utf-8') + value + b'\r\n'
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_3lib_nil);
    __pyx_r = __pyx_v_3lib_nil;
    goto __pyx_L0;

    /* "extensions/lib/rparser.pyx":57
 *     # SERVER ENCODERS
 *     def bulk(self, bytes value=None):
 *         if value is None:             # <<<<<<<<<<<<<<
 *             return nil
 *         else:
 */
  }

  /* "extensions/lib/rparser.pyx":60
 *             return nil
 *         else:
 *             return ('$%d\r\n' % len(value)).encode('utf-8') + value + b'\r\n'             # <<<<<<<<<<<<<<
 * 
 *     def multi_bulk_len(self, len):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    if (unlikely(__pyx_v_value == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 60, __pyx_L1_error)
    }
    __pyx_t_3 = PyBytes_GET_SIZE(__pyx_v_value); if (unlikely(__pyx_t_3 == -1)) __PYX_ERR(0, 60, __pyx_L1_error)
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyString_Format(__pyx_kp_s_d, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_encode); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyNumber_Add(__pyx_t_5, __pyx_v_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_Add(__pyx_t_4, __pyx_kp_b__3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 60, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;
  }

  /* "extensions/lib/rparser.pyx":56
 * 
 *     # SERVER ENCODERS
 *     def bulk(self, bytes value=None):             # <<<<<<<<<<<<<<
 *         if value is None:
 *             return nil
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("lib.RedisParser.bulk", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/rparser.pyx":62
 *             return ('$%d\r\n' % len(value)).encode('utf-8') + value + b'\r\n'
 * 
 *     def multi_bulk_len(self, len):             # <<<<<<<<<<<<<<
 *         return ('*%s\r\n' % len).encode('utf-8')
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_11RedisParser_19multi_bulk_len(PyObject *__pyx_v_self, PyObject *__pyx_v_len); /*proto*/
static PyObject *__pyx_pw_3lib_11RedisParser_19multi_bulk_len(PyObject *__pyx_v_self, PyObject *__pyx_v_len) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("multi_bulk_len (wrapper)", 0);
  __pyx_r = __pyx_pf_3lib_11RedisParser_18multi_bulk_len(((struct __pyx_obj_3lib_RedisParser *)__pyx_v_self), ((PyObject *)__pyx_v_len));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_11RedisParser_18multi_bulk_len(CYTHON_UNUSED struct __pyx_obj_3lib_RedisParser *__pyx_v_self, PyObject *__pyx_v_len) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("multi_bulk_len", 0);

  /* "extensions/lib/rparser.pyx":63
 * 
 *     def multi_bulk_len(self, len):
 *         return ('*%s\r\n' % len).encode('utf-8')             # <<<<<<<<<<<<<<
 * 
 *     def multi_bulk(self, args):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_s, __pyx_v_len); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "extensions/lib/rparser.pyx":62
 *             return ('$%d\r\n' % len(value)).encode('utf-8') + value + b'\r\n'
 * 
 *     def multi_bulk_len(self, len):             # <<<<<<<<<<<<<<
 *         return ('*%s\r\n' % len).encode('utf-8')
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("lib.RedisParser.multi_bulk_len", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/rparser.pyx":65
 *         return ('*%s\r\n' % len).encode('utf-8')
 * 
 *     def multi_bulk(self, args):             # <<<<<<<<<<<<<<
 *         return null_array if args is None else b''.join(self._pack(args))
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_11RedisParser_21multi_bulk(PyObject *__pyx_v_self, PyObject *__pyx_v_args); /*proto*/
static PyObject *__pyx_pw_3lib_11RedisParser_21multi_bulk(PyObject *__pyx_v_self, PyObject *__pyx_v_args) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("multi_bulk (wrapper)", 0);
  __pyx_r = __pyx_pf_3lib_11RedisParser_20multi_bulk(((struct __pyx_obj_3lib_RedisParser *)__pyx_v_self), ((PyObject *)__pyx_v_args));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_11RedisParser_20multi_bulk(struct __pyx_obj_3lib_RedisParser *__pyx_v_self, PyObject *__pyx_v_args) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("multi_bulk", 0);

  /* "extensions/lib/rparser.pyx":66
 * 
 *     def multi_bulk(self, args):
 *         return null_array if args is None else b''.join(self._pack(args))             # <<<<<<<<<<<<<<
 * 
 *     # INTERNALS
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = (__pyx_v_args == Py_None);
  if ((__pyx_t_2 != 0)) {
    __Pyx_INCREF(__pyx_v_3lib_null_array);
    __pyx_t_1 = __pyx_v_3lib_null_array;
  } else {
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_pack); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_args); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 66, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 66, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
      __Pyx_INCREF(__pyx_v_args);
      __Pyx_GIVEREF(__pyx_v_args);
      PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_args);
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 66, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyBytes_Join(__pyx_kp_b_, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 66, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_1 = __pyx_t_4;
    __pyx_t_4 = 0;
  }
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "extensions/lib/rparser.pyx":65
 *         return ('*%s\r\n' % len).encode('utf-8')
 * 
 *     def multi_bulk(self, args):             # <<<<<<<<<<<<<<
 *         return null_array if args is None else b''.join(self._pack(args))
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("lib.RedisParser.multi_bulk", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_3lib_11RedisParser_24generator(__pyx_CoroutineObject *__pyx_generator, PyObject *__pyx_sent_value); /* proto */

/* "extensions/lib/rparser.pyx":69
 * 
 *     # INTERNALS
 *     def _pack_command(self, args):             # <<<<<<<<<<<<<<
 *         yield ('*%d\r\n' % len(args)).encode('utf-8')
 *         for value in args:
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_11RedisParser_23_pack_command(PyObject *__pyx_v_self, PyObject *__pyx_v_args); /*proto*/
static PyObject *__pyx_pw_3lib_11RedisParser_23_pack_command(PyObject *__pyx_v_self, PyObject *__pyx_v_args) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_pack_command (wrapper)", 0);
  __pyx_r = __pyx_pf_3lib_11RedisParser_22_pack_command(((struct __pyx_obj_3lib_RedisParser *)__pyx_v_self), ((PyObject *)__pyx_v_args));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_11RedisParser_22_pack_command(CYTHON_UNUSED struct __pyx_obj_3lib_RedisParser *__pyx_v_self, PyObject *__pyx_v_args) {
  struct __pyx_obj_3lib___pyx_scope_struct_2__pack_command *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_pack_command", 0);
  __pyx_cur_scope = (struct __pyx_obj_3lib___pyx_scope_struct_2__pack_command *)__pyx_tp_new_3lib___pyx_scope_struct_2__pack_command(__pyx_ptype_3lib___pyx_scope_struct_2__pack_command, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __Pyx_GOTREF(__pyx_cur_scope);
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_args = __pyx_v_args;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_args);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_args);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_3lib_11RedisParser_24generator, (PyObject *) __pyx_cur_scope, __pyx_n_s_pack_command, __pyx_n_s_RedisParser__pack_command); if (unlikely(!gen)) __PYX_ERR(0, 69, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("lib.RedisParser._pack_command", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_3lib_11RedisParser_24generator(__pyx_CoroutineObject *__pyx_generator, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_3lib___pyx_scope_struct_2__pack_command *__pyx_cur_scope = ((struct __pyx_obj_3lib___pyx_scope_struct_2__pack_command *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *(*__pyx_t_4)(PyObject *);
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("None", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L4_resume_from_yield;
    case 2: goto __pyx_L8_resume_from_yield;
    case 3: goto __pyx_L9_resume_from_yield;
    case 4: goto __pyx_L10_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 69, __pyx_L1_error)

  /* "extensions/lib/rparser.pyx":70
 *     # INTERNALS
 *     def _pack_command(self, args):
 *         yield ('*%d\r\n' % len(args)).encode('utf-8')             # <<<<<<<<<<<<<<
 *         for value in args:
 *             if isinstance(value, str):
 */
  __pyx_t_1 = PyObject_Length(__pyx_cur_scope->__pyx_v_args); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 70, __pyx_L1_error)
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_d_2, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  /* return from generator, yielding value */
  __pyx_generator->resume_label = 1;
  return __pyx_r;
  __pyx_L4_resume_from_yield:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 70, __pyx_L1_error)

  /* "extensions/lib/rparser.pyx":71
 *     def _pack_command(self, args):
 *         yield ('*%d\r\n' % len(args)).encode('utf-8')
 *         for value in args:             # <<<<<<<<<<<<<<
 *             if isinstance(value, str):
 *                 value = value.encode('utf-8')
 */
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_v_args)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_v_args)) {
    __pyx_t_3 = __pyx_cur_scope->__pyx_v_args; __Pyx_INCREF(__pyx_t_3); __pyx_t_1 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_1 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_cur_scope->__pyx_v_args); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 71, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 71, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_1 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_1); __Pyx_INCREF(__pyx_t_2); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 71, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 71, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_1 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_1); __Pyx_INCREF(__pyx_t_2); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 71, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 71, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_4(__pyx_t_3);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 71, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_value);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_value, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_t_2 = 0;

    /* "extensions/lib/rparser.pyx":72
 *         yield ('*%d\r\n' % len(args)).encode('utf-8')
 *         for value in args:
 *             if isinstance(value, str):             # <<<<<<<<<<<<<<
 *                 value = value.encode('utf-8')
 *             elif not isinstance(value, bytes):
 */
    __pyx_t_5 = PyString_Check(__pyx_cur_scope->__pyx_v_value); 
    __pyx_t_6 = (__pyx_t_5 != 0);
    if (__pyx_t_6) {

      /* "extensions/lib/rparser.pyx":73
 *         for value in args:
 *             if isinstance(value, str):
 *                 value = value.encode('utf-8')             # <<<<<<<<<<<<<<
 *             elif not isinstance(value, bytes):
 *                 value = str(value).encode('utf-8')
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_value, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 73, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 73, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_value);
      __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_value, __pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_7);
      __pyx_t_7 = 0;

      /* "extensions/lib/rparser.pyx":72
 *         yield ('*%d\r\n' % len(args)).encode('utf-8')
 *         for value in args:
 *             if isinstance(value, str):             # <<<<<<<<<<<<<<
 *                 value = value.encode('utf-8')
 *             elif not isinstance(value, bytes):
 */
      goto __pyx_L7;
    }

    /* "extensions/lib/rparser.pyx":74
 *             if isinstance(value, str):
 *                 value = value.encode('utf-8')
 *             elif not isinstance(value, bytes):             # <<<<<<<<<<<<<<
 *                 value = str(value).encode('utf-8')
 *             yield ('$%d\r\n' % len(value)).encode('utf-8')
 */
    __pyx_t_6 = PyBytes_Check(__pyx_cur_scope->__pyx_v_value); 
    __pyx_t_5 = ((!(__pyx_t_6 != 0)) != 0);
    if (__pyx_t_5) {

      /* "extensions/lib/rparser.pyx":75
 *                 value = value.encode('utf-8')
 *             elif not isinstance(value, bytes):
 *                 value = str(value).encode('utf-8')             # <<<<<<<<<<<<<<
 *             yield ('$%d\r\n' % len(value)).encode('utf-8')
 *             yield value
 */
      __pyx_t_7 = PyTuple_New(1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 75, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_value);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_value);
      PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_cur_scope->__pyx_v_value);
      __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)(&PyString_Type)), __pyx_t_7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 75, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_encode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 75, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 75, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_value);
      __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_value, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      __pyx_t_2 = 0;

      /* "extensions/lib/rparser.pyx":74
 *             if isinstance(value, str):
 *                 value = value.encode('utf-8')
 *             elif not isinstance(value, bytes):             # <<<<<<<<<<<<<<
 *                 value = str(value).encode('utf-8')
 *             yield ('$%d\r\n' % len(value)).encode('utf-8')
 */
    }
    __pyx_L7:;

    /* "extensions/lib/rparser.pyx":76
 *             elif not isinstance(value, bytes):
 *                 value = str(value).encode('utf-8')
 *             yield ('$%d\r\n' % len(value)).encode('utf-8')             # <<<<<<<<<<<<<<
 *             yield value
 *             yield CRLF
 */
    __pyx_t_8 = PyObject_Length(__pyx_cur_scope->__pyx_v_value); if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(0, 76, __pyx_L1_error)
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_7 = __Pyx_PyString_Format(__pyx_kp_s_d, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_7;
    __pyx_t_7 = 0;
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __Pyx_XGIVEREF(__pyx_t_3);
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 2;
    return __pyx_r;
    __pyx_L8_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = 0;
    __Pyx_XGOTREF(__pyx_t_3);
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 76, __pyx_L1_error)

    /* "extensions/lib/rparser.pyx":77
 *                 value = str(value).encode('utf-8')
 *             yield ('$%d\r\n' % len(value)).encode('utf-8')
 *             yield value             # <<<<<<<<<<<<<<
 *             yield CRLF
 * 
 */
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_value);
    __pyx_r = __pyx_cur_scope->__pyx_v_value;
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __Pyx_XGIVEREF(__pyx_t_3);
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 3;
    return __pyx_r;
    __pyx_L9_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = 0;
    __Pyx_XGOTREF(__pyx_t_3);
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 77, __pyx_L1_error)

    /* "extensions/lib/rparser.pyx":78
 *             yield ('$%d\r\n' % len(value)).encode('utf-8')
 *             yield value
 *             yield CRLF             # <<<<<<<<<<<<<<
 * 
 *     def _pack(self, args):
 */
    __Pyx_INCREF(__pyx_v_3lib_CRLF);
    __pyx_r = __pyx_v_3lib_CRLF;
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __Pyx_XGIVEREF(__pyx_t_3);
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 4;
    return __pyx_r;
    __pyx_L10_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = 0;
    __Pyx_XGOTREF(__pyx_t_3);
    __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 78, __pyx_L1_error)

    /* "extensions/lib/rparser.pyx":71
 *     def _pack_command(self, args):
 *         yield ('*%d\r\n' % len(args)).encode('utf-8')
 *         for value in args:             # <<<<<<<<<<<<<<
 *             if isinstance(value, str):
 *                 value = value.encode('utf-8')
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "extensions/lib/rparser.pyx":69
 * 
 *     # INTERNALS
 *     def _pack_command(self, args):             # <<<<<<<<<<<<<<
 *         yield ('*%d\r\n' % len(args)).encode('utf-8')
 *         for value in args:
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("_pack_command", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_3lib_11RedisParser_27generator1(__pyx_CoroutineObject *__pyx_generator, PyObject *__pyx_sent_value); /* proto */

/* "extensions/lib/rparser.pyx":80
 *             yield CRLF
 * 
 *     def _pack(self, args):             # <<<<<<<<<<<<<<
 *         yield ('*%d\r\n' % len(args)).encode('utf-8')
 *         for value in args:
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_11RedisParser_26_pack(PyObject *__pyx_v_self, PyObject *__pyx_v_args); /*proto*/
static PyObject *__pyx_pw_3lib_11RedisParser_26_pack(PyObject *__pyx_v_self, PyObject *__pyx_v_args) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_pack (wrapper)", 0);
  __pyx_r = __pyx_pf_3lib_11RedisParser_25_pack(((struct __pyx_obj_3lib_RedisParser *)__pyx_v_self), ((PyObject *)__pyx_v_args));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_11RedisParser_25_pack(struct __pyx_obj_3lib_RedisParser *__pyx_v_self, PyObject *__pyx_v_args) {
  struct __pyx_obj_3lib___pyx_scope_struct_3__pack *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_pack", 0);
  __pyx_cur_scope = (struct __pyx_obj_3lib___pyx_scope_struct_3__pack *)__pyx_tp_new_3lib___pyx_scope_struct_3__pack(__pyx_ptype_3lib___pyx_scope_struct_3__pack, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __Pyx_GOTREF(__pyx_cur_scope);
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_args = __pyx_v_args;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_args);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_args);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_3lib_11RedisParser_27generator1, (PyObject *) __pyx_cur_scope, __pyx_n_s_pack, __pyx_n_s_RedisParser__pack); if (unlikely(!gen)) __PYX_ERR(0, 80, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("lib.RedisParser._pack", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_3lib_11RedisParser_27generator1(__pyx_CoroutineObject *__pyx_generator, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_3lib___pyx_scope_struct_3__pack *__pyx_cur_scope = ((struct __pyx_obj_3lib___pyx_scope_struct_3__pack *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *(*__pyx_t_4)(PyObject *);
  int __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *(*__pyx_t_13)(PyObject *);
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("None", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L4_resume_from_yield;
    case 2: goto __pyx_L8_resume_from_yield;
    case 3: goto __pyx_L9_resume_from_yield;
    case 4: goto __pyx_L10_resume_from_yield;
    case 5: goto __pyx_L11_resume_from_yield;
    case 6: goto __pyx_L12_resume_from_yield;
    case 7: goto __pyx_L13_resume_from_yield;
    case 8: goto __pyx_L14_resume_from_yield;
    case 9: goto __pyx_L17_resume_from_yield;
    case 10: goto __pyx_L20_resume_from_yield;
    case 11: goto __pyx_L21_resume_from_yield;
    case 12: goto __pyx_L22_resume_from_yield;
    case 13: goto __pyx_L23_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 80, __pyx_L1_error)

  /* "extensions/lib/rparser.pyx":81
 * 
 *     def _pack(self, args):
 *         yield ('*%d\r\n' % len(args)).encode('utf-8')             # <<<<<<<<<<<<<<
 *         for value in args:
 *             if value is None:
 */
  __pyx_t_1 = PyObject_Length(__pyx_cur_scope->__pyx_v_args); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 81, __pyx_L1_error)
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_d_2, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  /* return from generator, yielding value */
  __pyx_generator->resume_label = 1;
  return __pyx_r;
  __pyx_L4_resume_from_yield:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 81, __pyx_L1_error)

  /* "extensions/lib/rparser.pyx":82
 *     def _pack(self, args):
 *         yield ('*%d\r\n' % len(args)).encode('utf-8')
 *         for value in args:             # <<<<<<<<<<<<<<
 *             if value is None:
 *                 yield nil
 */
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_v_args)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_v_args)) {
    __pyx_t_3 = __pyx_cur_scope->__pyx_v_args; __Pyx_INCREF(__pyx_t_3); __pyx_t_1 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_1 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_cur_scope->__pyx_v_args); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 82, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 82, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_1 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_1); __Pyx_INCREF(__pyx_t_2); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 82, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 82, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_1 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_COMPILING_IN_CPYTHON
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_1); __Pyx_INCREF(__pyx_t_2); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(0, 82, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 82, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_4(__pyx_t_3);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 82, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_value);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_value, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __pyx_t_2 = 0;

    /* "extensions/lib/rparser.pyx":83
 *         yield ('*%d\r\n' % len(args)).encode('utf-8')
 *         for value in args:
 *             if value is None:             # <<<<<<<<<<<<<<
 *                 yield nil
 *             elif isinstance(value, bytes):
 */
    __pyx_t_5 = (__pyx_cur_scope->__pyx_v_value == Py_None);
    __pyx_t_6 = (__pyx_t_5 != 0);
    if (__pyx_t_6) {

      /* "extensions/lib/rparser.pyx":84
 *         for value in args:
 *             if value is None:
 *                 yield nil             # <<<<<<<<<<<<<<
 *             elif isinstance(value, bytes):
 *                 yield ('$%d\r\n' % len(value)).encode('utf-8')
 */
      __Pyx_INCREF(__pyx_v_3lib_nil);
      __pyx_r = __pyx_v_3lib_nil;
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
      __Pyx_XGIVEREF(__pyx_t_3);
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
      __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 2;
      return __pyx_r;
      __pyx_L8_resume_from_yield:;
      __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
      __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
      __pyx_cur_scope->__pyx_t_1 = 0;
      __Pyx_XGOTREF(__pyx_t_3);
      __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 84, __pyx_L1_error)

      /* "extensions/lib/rparser.pyx":83
 *         yield ('*%d\r\n' % len(args)).encode('utf-8')
 *         for value in args:
 *             if value is None:             # <<<<<<<<<<<<<<
 *                 yield nil
 *             elif isinstance(value, bytes):
 */
      goto __pyx_L7;
    }

    /* "extensions/lib/rparser.pyx":85
 *             if value is None:
 *                 yield nil
 *             elif isinstance(value, bytes):             # <<<<<<<<<<<<<<
 *                 yield ('$%d\r\n' % len(value)).encode('utf-8')
 *                 yield value
 */
    __pyx_t_6 = PyBytes_Check(__pyx_cur_scope->__pyx_v_value); 
    __pyx_t_5 = (__pyx_t_6 != 0);
    if (__pyx_t_5) {

      /* "extensions/lib/rparser.pyx":86
 *                 yield nil
 *             elif isinstance(value, bytes):
 *                 yield ('$%d\r\n' % len(value)).encode('utf-8')             # <<<<<<<<<<<<<<
 *                 yield value
 *                 yield CRLF
 */
      __pyx_t_7 = PyObject_Length(__pyx_cur_scope->__pyx_v_value); if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 86, __pyx_L1_error)
      __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 86, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_8 = __Pyx_PyString_Format(__pyx_kp_s_d, __pyx_t_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 86, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 86, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 86, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_8;
      __pyx_t_8 = 0;
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
      __Pyx_XGIVEREF(__pyx_t_3);
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
      __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 3;
      return __pyx_r;
      __pyx_L9_resume_from_yield:;
      __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
      __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
      __pyx_cur_scope->__pyx_t_1 = 0;
      __Pyx_XGOTREF(__pyx_t_3);
      __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 86, __pyx_L1_error)

      /* "extensions/lib/rparser.pyx":87
 *             elif isinstance(value, bytes):
 *                 yield ('$%d\r\n' % len(value)).encode('utf-8')
 *                 yield value             # <<<<<<<<<<<<<<
 *                 yield CRLF
 *             elif isinstance(value, str):
 */
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_value);
      __pyx_r = __pyx_cur_scope->__pyx_v_value;
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
      __Pyx_XGIVEREF(__pyx_t_3);
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
      __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 4;
      return __pyx_r;
      __pyx_L10_resume_from_yield:;
      __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
      __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
      __pyx_cur_scope->__pyx_t_1 = 0;
      __Pyx_XGOTREF(__pyx_t_3);
      __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 87, __pyx_L1_error)

      /* "extensions/lib/rparser.pyx":88
 *                 yield ('$%d\r\n' % len(value)).encode('utf-8')
 *                 yield value
 *                 yield CRLF             # <<<<<<<<<<<<<<
 *             elif isinstance(value, str):
 *                 value = value.encode('utf-8')
 */
      __Pyx_INCREF(__pyx_v_3lib_CRLF);
      __pyx_r = __pyx_v_3lib_CRLF;
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
      __Pyx_XGIVEREF(__pyx_t_3);
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
      __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 5;
      return __pyx_r;
      __pyx_L11_resume_from_yield:;
      __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
      __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
      __pyx_cur_scope->__pyx_t_1 = 0;
      __Pyx_XGOTREF(__pyx_t_3);
      __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 88, __pyx_L1_error)

      /* "extensions/lib/rparser.pyx":85
 *             if value is None:
 *                 yield nil
 *             elif isinstance(value, bytes):             # <<<<<<<<<<<<<<
 *                 yield ('$%d\r\n' % len(value)).encode('utf-8')
 *                 yield value
 */
      goto __pyx_L7;
    }

    /* "extensions/lib/rparser.pyx":89
 *                 yield value
 *                 yield CRLF
 *             elif isinstance(value, str):             # <<<<<<<<<<<<<<
 *                 value = value.encode('utf-8')
 *                 yield ('$%d\r\n' % len(value)).encode('utf-8')
 */
    __pyx_t_5 = PyString_Check(__pyx_cur_scope->__pyx_v_value); 
    __pyx_t_6 = (__pyx_t_5 != 0);
    if (__pyx_t_6) {

      /* "extensions/lib/rparser.pyx":90
 *                 yield CRLF
 *             elif isinstance(value, str):
 *                 value = value.encode('utf-8')             # <<<<<<<<<<<<<<
 *                 yield ('$%d\r\n' % len(value)).encode('utf-8')
 *                 yield value
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_value, __pyx_n_s_encode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 90, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 90, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_value);
      __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_value, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      __pyx_t_2 = 0;

      /* "extensions/lib/rparser.pyx":91
 *             elif isinstance(value, str):
 *                 value = value.encode('utf-8')
 *                 yield ('$%d\r\n' % len(value)).encode('utf-8')             # <<<<<<<<<<<<<<
 *                 yield value
 *                 yield CRLF
 */
      __pyx_t_7 = PyObject_Length(__pyx_cur_scope->__pyx_v_value); if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 91, __pyx_L1_error)
      __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 91, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_8 = __Pyx_PyString_Format(__pyx_kp_s_d, __pyx_t_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 91, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 91, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 91, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_8;
      __pyx_t_8 = 0;
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
      __Pyx_XGIVEREF(__pyx_t_3);
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
      __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 6;
      return __pyx_r;
      __pyx_L12_resume_from_yield:;
      __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
      __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
      __pyx_cur_scope->__pyx_t_1 = 0;
      __Pyx_XGOTREF(__pyx_t_3);
      __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 91, __pyx_L1_error)

      /* "extensions/lib/rparser.pyx":92
 *                 value = value.encode('utf-8')
 *                 yield ('$%d\r\n' % len(value)).encode('utf-8')
 *                 yield value             # <<<<<<<<<<<<<<
 *                 yield CRLF
 *             elif hasattr(value, 'items'):
 */
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_value);
      __pyx_r = __pyx_cur_scope->__pyx_v_value;
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
      __Pyx_XGIVEREF(__pyx_t_3);
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
      __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 7;
      return __pyx_r;
      __pyx_L13_resume_from_yield:;
      __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
      __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
      __pyx_cur_scope->__pyx_t_1 = 0;
      __Pyx_XGOTREF(__pyx_t_3);
      __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 92, __pyx_L1_error)

      /* "extensions/lib/rparser.pyx":93
 *                 yield ('$%d\r\n' % len(value)).encode('utf-8')
 *                 yield value
 *                 yield CRLF             # <<<<<<<<<<<<<<
 *             elif hasattr(value, 'items'):
 *                 for value in self._pack(tuple(self._lua_dict(value))):
 */
      __Pyx_INCREF(__pyx_v_3lib_CRLF);
      __pyx_r = __pyx_v_3lib_CRLF;
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
      __Pyx_XGIVEREF(__pyx_t_3);
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
      __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 8;
      return __pyx_r;
      __pyx_L14_resume_from_yield:;
      __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
      __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
      __pyx_cur_scope->__pyx_t_1 = 0;
      __Pyx_XGOTREF(__pyx_t_3);
      __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 93, __pyx_L1_error)

      /* "extensions/lib/rparser.pyx":89
 *                 yield value
 *                 yield CRLF
 *             elif isinstance(value, str):             # <<<<<<<<<<<<<<
 *                 value = value.encode('utf-8')
 *                 yield ('$%d\r\n' % len(value)).encode('utf-8')
 */
      goto __pyx_L7;
    }

    /* "extensions/lib/rparser.pyx":94
 *                 yield value
 *                 yield CRLF
 *             elif hasattr(value, 'items'):             # <<<<<<<<<<<<<<
 *                 for value in self._pack(tuple(self._lua_dict(value))):
 *                     yield value
 */
    __pyx_t_6 = PyObject_HasAttr(__pyx_cur_scope->__pyx_v_value, __pyx_n_s_items); if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(0, 94, __pyx_L1_error)
    __pyx_t_5 = (__pyx_t_6 != 0);
    if (__pyx_t_5) {

      /* "extensions/lib/rparser.pyx":95
 *                 yield CRLF
 *             elif hasattr(value, 'items'):
 *                 for value in self._pack(tuple(self._lua_dict(value))):             # <<<<<<<<<<<<<<
 *                     yield value
 *             elif hasattr(value, '__len__'):
 */
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_pack); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 95, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_lua_dict); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 95, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_11 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_10))) {
        __pyx_t_11 = PyMethod_GET_SELF(__pyx_t_10);
        if (likely(__pyx_t_11)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_11);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_10, function);
        }
      }
      if (!__pyx_t_11) {
        __pyx_t_9 = __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_cur_scope->__pyx_v_value); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 95, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
      } else {
        __pyx_t_12 = PyTuple_New(1+1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 95, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_GIVEREF(__pyx_t_11); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_11); __pyx_t_11 = NULL;
        __Pyx_INCREF(__pyx_cur_scope->__pyx_v_value);
        __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_value);
        PyTuple_SET_ITEM(__pyx_t_12, 0+1, __pyx_cur_scope->__pyx_v_value);
        __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_12, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 95, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      }
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = PySequence_Tuple(__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 95, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_9 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_2))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_2);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_2, function);
        }
      }
      if (!__pyx_t_9) {
        __pyx_t_8 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_10); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 95, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_GOTREF(__pyx_t_8);
      } else {
        __pyx_t_12 = PyTuple_New(1+1); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 95, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_t_9); __pyx_t_9 = NULL;
        __Pyx_GIVEREF(__pyx_t_10);
        PyTuple_SET_ITEM(__pyx_t_12, 0+1, __pyx_t_10);
        __pyx_t_10 = 0;
        __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_12, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 95, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (likely(PyList_CheckExact(__pyx_t_8)) || PyTuple_CheckExact(__pyx_t_8)) {
        __pyx_t_2 = __pyx_t_8; __Pyx_INCREF(__pyx_t_2); __pyx_t_7 = 0;
        __pyx_t_13 = NULL;
      } else {
        __pyx_t_7 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 95, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __pyx_t_13 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 95, __pyx_L1_error)
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      for (;;) {
        if (likely(!__pyx_t_13)) {
          if (likely(PyList_CheckExact(__pyx_t_2))) {
            if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_2)) break;
            #if CYTHON_COMPILING_IN_CPYTHON
            __pyx_t_8 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_7); __Pyx_INCREF(__pyx_t_8); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 95, __pyx_L1_error)
            #else
            __pyx_t_8 = PySequence_ITEM(__pyx_t_2, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 95, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_8);
            #endif
          } else {
            if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
            #if CYTHON_COMPILING_IN_CPYTHON
            __pyx_t_8 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_7); __Pyx_INCREF(__pyx_t_8); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 95, __pyx_L1_error)
            #else
            __pyx_t_8 = PySequence_ITEM(__pyx_t_2, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 95, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_8);
            #endif
          }
        } else {
          __pyx_t_8 = __pyx_t_13(__pyx_t_2);
          if (unlikely(!__pyx_t_8)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 95, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_8);
        }
        __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_value);
        __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_value, __pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_8);
        __pyx_t_8 = 0;

        /* "extensions/lib/rparser.pyx":96
 *             elif hasattr(value, 'items'):
 *                 for value in self._pack(tuple(self._lua_dict(value))):
 *                     yield value             # <<<<<<<<<<<<<<
 *             elif hasattr(value, '__len__'):
 *                 for value in self._pack(value):
 */
        __Pyx_INCREF(__pyx_cur_scope->__pyx_v_value);
        __pyx_r = __pyx_cur_scope->__pyx_v_value;
        __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
        __Pyx_XGIVEREF(__pyx_t_2);
        __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
        __Pyx_XGIVEREF(__pyx_t_3);
        __pyx_cur_scope->__pyx_t_3 = __pyx_t_3;
        __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
        __pyx_cur_scope->__pyx_t_4 = __pyx_t_7;
        __pyx_cur_scope->__pyx_t_5 = __pyx_t_13;
        __Pyx_XGIVEREF(__pyx_r);
        __Pyx_RefNannyFinishContext();
        /* return from generator, yielding value */
        __pyx_generator->resume_label = 9;
        return __pyx_r;
        __pyx_L17_resume_from_yield:;
        __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
        __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
        __pyx_cur_scope->__pyx_t_1 = 0;
        __Pyx_XGOTREF(__pyx_t_2);
        __pyx_t_3 = __pyx_cur_scope->__pyx_t_3;
        __pyx_cur_scope->__pyx_t_3 = 0;
        __Pyx_XGOTREF(__pyx_t_3);
        __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
        __pyx_t_7 = __pyx_cur_scope->__pyx_t_4;
        __pyx_t_13 = __pyx_cur_scope->__pyx_t_5;
        if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 96, __pyx_L1_error)

        /* "extensions/lib/rparser.pyx":95
 *                 yield CRLF
 *             elif hasattr(value, 'items'):
 *                 for value in self._pack(tuple(self._lua_dict(value))):             # <<<<<<<<<<<<<<
 *                     yield value
 *             elif hasattr(value, '__len__'):
 */
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "extensions/lib/rparser.pyx":94
 *                 yield value
 *                 yield CRLF
 *             elif hasattr(value, 'items'):             # <<<<<<<<<<<<<<
 *                 for value in self._pack(tuple(self._lua_dict(value))):
 *                     yield value
 */
      goto __pyx_L7;
    }

    /* "extensions/lib/rparser.pyx":97
 *                 for value in self._pack(tuple(self._lua_dict(value))):
 *                     yield value
 *             elif hasattr(value, '__len__'):             # <<<<<<<<<<<<<<
 *                 for value in self._pack(value):
 *                     yield value
 */
    __pyx_t_5 = PyObject_HasAttr(__pyx_cur_scope->__pyx_v_value, __pyx_n_s_len); if (unlikely(__pyx_t_5 == -1)) __PYX_ERR(0, 97, __pyx_L1_error)
    __pyx_t_6 = (__pyx_t_5 != 0);
    if (__pyx_t_6) {

      /* "extensions/lib/rparser.pyx":98
 *                     yield value
 *             elif hasattr(value, '__len__'):
 *                 for value in self._pack(value):             # <<<<<<<<<<<<<<
 *                     yield value
 *             else:
 */
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_self), __pyx_n_s_pack); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 98, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_12 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_12 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_12)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_12);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      if (!__pyx_t_12) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_cur_scope->__pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 98, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        __pyx_t_10 = PyTuple_New(1+1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 98, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_GIVEREF(__pyx_t_12); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_12); __pyx_t_12 = NULL;
        __Pyx_INCREF(__pyx_cur_scope->__pyx_v_value);
        __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_value);
        PyTuple_SET_ITEM(__pyx_t_10, 0+1, __pyx_cur_scope->__pyx_v_value);
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_10, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 98, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
        __pyx_t_8 = __pyx_t_2; __Pyx_INCREF(__pyx_t_8); __pyx_t_7 = 0;
        __pyx_t_13 = NULL;
      } else {
        __pyx_t_7 = -1; __pyx_t_8 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 98, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_13 = Py_TYPE(__pyx_t_8)->tp_iternext; if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 98, __pyx_L1_error)
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      for (;;) {
        if (likely(!__pyx_t_13)) {
          if (likely(PyList_CheckExact(__pyx_t_8))) {
            if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_8)) break;
            #if CYTHON_COMPILING_IN_CPYTHON
            __pyx_t_2 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_7); __Pyx_INCREF(__pyx_t_2); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 98, __pyx_L1_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_8, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 98, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          } else {
            if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_8)) break;
            #if CYTHON_COMPILING_IN_CPYTHON
            __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_7); __Pyx_INCREF(__pyx_t_2); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(0, 98, __pyx_L1_error)
            #else
            __pyx_t_2 = PySequence_ITEM(__pyx_t_8, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 98, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_2);
            #endif
          }
        } else {
          __pyx_t_2 = __pyx_t_13(__pyx_t_8);
          if (unlikely(!__pyx_t_2)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 98, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_2);
        }
        __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_value);
        __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_value, __pyx_t_2);
        __Pyx_GIVEREF(__pyx_t_2);
        __pyx_t_2 = 0;

        /* "extensions/lib/rparser.pyx":99
 *             elif hasattr(value, '__len__'):
 *                 for value in self._pack(value):
 *                     yield value             # <<<<<<<<<<<<<<
 *             else:
 *                 value = str(value).encode('utf-8')
 */
        __Pyx_INCREF(__pyx_cur_scope->__pyx_v_value);
        __pyx_r = __pyx_cur_scope->__pyx_v_value;
        __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
        __Pyx_XGIVEREF(__pyx_t_3);
        __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
        __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
        __pyx_cur_scope->__pyx_t_4 = __pyx_t_7;
        __Pyx_XGIVEREF(__pyx_t_8);
        __pyx_cur_scope->__pyx_t_3 = __pyx_t_8;
        __pyx_cur_scope->__pyx_t_5 = __pyx_t_13;
        __Pyx_XGIVEREF(__pyx_r);
        __Pyx_RefNannyFinishContext();
        /* return from generator, yielding value */
        __pyx_generator->resume_label = 10;
        return __pyx_r;
        __pyx_L20_resume_from_yield:;
        __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
        __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
        __pyx_cur_scope->__pyx_t_1 = 0;
        __Pyx_XGOTREF(__pyx_t_3);
        __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
        __pyx_t_7 = __pyx_cur_scope->__pyx_t_4;
        __pyx_t_8 = __pyx_cur_scope->__pyx_t_3;
        __pyx_cur_scope->__pyx_t_3 = 0;
        __Pyx_XGOTREF(__pyx_t_8);
        __pyx_t_13 = __pyx_cur_scope->__pyx_t_5;
        if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 99, __pyx_L1_error)

        /* "extensions/lib/rparser.pyx":98
 *                     yield value
 *             elif hasattr(value, '__len__'):
 *                 for value in self._pack(value):             # <<<<<<<<<<<<<<
 *                     yield value
 *             else:
 */
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "extensions/lib/rparser.pyx":97
 *                 for value in self._pack(tuple(self._lua_dict(value))):
 *                     yield value
 *             elif hasattr(value, '__len__'):             # <<<<<<<<<<<<<<
 *                 for value in self._pack(value):
 *                     yield value
 */
      goto __pyx_L7;
    }

    /* "extensions/lib/rparser.pyx":101
 *                     yield value
 *             else:
 *                 value = str(value).encode('utf-8')             # <<<<<<<<<<<<<<
 *                 yield ('$%d\r\n' % len(value)).encode('utf-8')
 *                 yield value
 */
    /*else*/ {
      __pyx_t_8 = PyTuple_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 101, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_value);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_value);
      PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_cur_scope->__pyx_v_value);
      __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)(&PyString_Type)), __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 101, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_encode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 101, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 101, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_value);
      __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_value, __pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_2);
      __pyx_t_2 = 0;

      /* "extensions/lib/rparser.pyx":102
 *             else:
 *                 value = str(value).encode('utf-8')
 *                 yield ('$%d\r\n' % len(value)).encode('utf-8')             # <<<<<<<<<<<<<<
 *                 yield value
 *                 yield CRLF
 */
      __pyx_t_7 = PyObject_Length(__pyx_cur_scope->__pyx_v_value); if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 102, __pyx_L1_error)
      __pyx_t_2 = PyInt_FromSsize_t(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 102, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_8 = __Pyx_PyString_Format(__pyx_kp_s_d, __pyx_t_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 102, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_encode); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 102, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 102, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_8;
      __pyx_t_8 = 0;
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
      __Pyx_XGIVEREF(__pyx_t_3);
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
      __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 11;
      return __pyx_r;
      __pyx_L21_resume_from_yield:;
      __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
      __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
      __pyx_cur_scope->__pyx_t_1 = 0;
      __Pyx_XGOTREF(__pyx_t_3);
      __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 102, __pyx_L1_error)

      /* "extensions/lib/rparser.pyx":103
 *                 value = str(value).encode('utf-8')
 *                 yield ('$%d\r\n' % len(value)).encode('utf-8')
 *                 yield value             # <<<<<<<<<<<<<<
 *                 yield CRLF
 * 
 */
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_value);
      __pyx_r = __pyx_cur_scope->__pyx_v_value;
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
      __Pyx_XGIVEREF(__pyx_t_3);
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
      __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 12;
      return __pyx_r;
      __pyx_L22_resume_from_yield:;
      __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
      __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
      __pyx_cur_scope->__pyx_t_1 = 0;
      __Pyx_XGOTREF(__pyx_t_3);
      __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 103, __pyx_L1_error)

      /* "extensions/lib/rparser.pyx":104
 *                 yield ('$%d\r\n' % len(value)).encode('utf-8')
 *                 yield value
 *                 yield CRLF             # <<<<<<<<<<<<<<
 * 
 *     def _lua_dict(self, d):
 */
      __Pyx_INCREF(__pyx_v_3lib_CRLF);
      __pyx_r = __pyx_v_3lib_CRLF;
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
      __Pyx_XGIVEREF(__pyx_t_3);
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_3;
      __pyx_cur_scope->__pyx_t_2 = __pyx_t_4;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 13;
      return __pyx_r;
      __pyx_L23_resume_from_yield:;
      __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
      __pyx_t_3 = __pyx_cur_scope->__pyx_t_1;
      __pyx_cur_scope->__pyx_t_1 = 0;
      __Pyx_XGOTREF(__pyx_t_3);
      __pyx_t_4 = __pyx_cur_scope->__pyx_t_2;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 104, __pyx_L1_error)
    }
    __pyx_L7:;

    /* "extensions/lib/rparser.pyx":82
 *     def _pack(self, args):
 *         yield ('*%d\r\n' % len(args)).encode('utf-8')
 *         for value in args:             # <<<<<<<<<<<<<<
 *             if value is None:
 *                 yield nil
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "extensions/lib/rparser.pyx":80
 *             yield CRLF
 * 
 *     def _pack(self, args):             # <<<<<<<<<<<<<<
 *         yield ('*%d\r\n' % len(args)).encode('utf-8')
 *         for value in args:
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_AddTraceback("_pack", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_3lib_11RedisParser_30generator2(__pyx_CoroutineObject *__pyx_generator, PyObject *__pyx_sent_value); /* proto */

/* "extensions/lib/rparser.pyx":106
 *                 yield CRLF
 * 
 *     def _lua_dict(self, d):             # <<<<<<<<<<<<<<
 *         index = 0
 *         while True:
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_11RedisParser_29_lua_dict(PyObject *__pyx_v_self, PyObject *__pyx_v_d); /*proto*/
static PyObject *__pyx_pw_3lib_11RedisParser_29_lua_dict(PyObject *__pyx_v_self, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_lua_dict (wrapper)", 0);
  __pyx_r = __pyx_pf_3lib_11RedisParser_28_lua_dict(((struct __pyx_obj_3lib_RedisParser *)__pyx_v_self), ((PyObject *)__pyx_v_d));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_11RedisParser_28_lua_dict(CYTHON_UNUSED struct __pyx_obj_3lib_RedisParser *__pyx_v_self, PyObject *__pyx_v_d) {
  struct __pyx_obj_3lib___pyx_scope_struct_4__lua_dict *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_lua_dict", 0);
  __pyx_cur_scope = (struct __pyx_obj_3lib___pyx_scope_struct_4__lua_dict *)__pyx_tp_new_3lib___pyx_scope_struct_4__lua_dict(__pyx_ptype_3lib___pyx_scope_struct_4__lua_dict, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __Pyx_GOTREF(__pyx_cur_scope);
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_d = __pyx_v_d;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_d);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_d);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_3lib_11RedisParser_30generator2, (PyObject *) __pyx_cur_scope, __pyx_n_s_lua_dict, __pyx_n_s_RedisParser__lua_dict); if (unlikely(!gen)) __PYX_ERR(0, 106, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("lib.RedisParser._lua_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_3lib_11RedisParser_30generator2(__pyx_CoroutineObject *__pyx_generator, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_3lib___pyx_scope_struct_4__lua_dict *__pyx_cur_scope = ((struct __pyx_obj_3lib___pyx_scope_struct_4__lua_dict *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("None", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L7_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 106, __pyx_L1_error)

  /* "extensions/lib/rparser.pyx":107
 * 
 *     def _lua_dict(self, d):
 *         index = 0             # <<<<<<<<<<<<<<
 *         while True:
 *             index += 1
 */
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  __pyx_cur_scope->__pyx_v_index = __pyx_int_0;

  /* "extensions/lib/rparser.pyx":108
 *     def _lua_dict(self, d):
 *         index = 0
 *         while True:             # <<<<<<<<<<<<<<
 *             index += 1
 *             v = d.get(index)
 */
  while (1) {

    /* "extensions/lib/rparser.pyx":109
 *         index = 0
 *         while True:
 *             index += 1             # <<<<<<<<<<<<<<
 *             v = d.get(index)
 *             if v is None:
 */
    __pyx_t_1 = __Pyx_PyInt_AddObjC(__pyx_cur_scope->__pyx_v_index, __pyx_int_1, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 109, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_index);
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_index, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;

    /* "extensions/lib/rparser.pyx":110
 *         while True:
 *             index += 1
 *             v = d.get(index)             # <<<<<<<<<<<<<<
 *             if v is None:
 *                 break
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_d, __pyx_n_s_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 110, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_3)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_3);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (!__pyx_t_3) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_cur_scope->__pyx_v_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 110, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    } else {
      __pyx_t_4 = PyTuple_New(1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 110, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_3); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3); __pyx_t_3 = NULL;
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_index);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_index);
      PyTuple_SET_ITEM(__pyx_t_4, 0+1, __pyx_cur_scope->__pyx_v_index);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 110, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_v);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_v, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __pyx_t_1 = 0;

    /* "extensions/lib/rparser.pyx":111
 *             index += 1
 *             v = d.get(index)
 *             if v is None:             # <<<<<<<<<<<<<<
 *                 break
 *             yield v
 */
    __pyx_t_5 = (__pyx_cur_scope->__pyx_v_v == Py_None);
    __pyx_t_6 = (__pyx_t_5 != 0);
    if (__pyx_t_6) {

      /* "extensions/lib/rparser.pyx":112
 *             v = d.get(index)
 *             if v is None:
 *                 break             # <<<<<<<<<<<<<<
 *             yield v
 * 
 */
      goto __pyx_L5_break;

      /* "extensions/lib/rparser.pyx":111
 *             index += 1
 *             v = d.get(index)
 *             if v is None:             # <<<<<<<<<<<<<<
 *                 break
 *             yield v
 */
    }

    /* "extensions/lib/rparser.pyx":113
 *             if v is None:
 *                 break
 *             yield v             # <<<<<<<<<<<<<<
 * 
 *     cdef object _get(self, Task next):
 */
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_v);
    __pyx_r = __pyx_cur_scope->__pyx_v_v;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L7_resume_from_yield:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 113, __pyx_L1_error)
  }
  __pyx_L5_break:;

  /* "extensions/lib/rparser.pyx":106
 *                 yield CRLF
 * 
 *     def _lua_dict(self, d):             # <<<<<<<<<<<<<<
 *         index = 0
 *         while True:
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("_lua_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/rparser.pyx":115
 *             yield v
 * 
 *     cdef object _get(self, Task next):             # <<<<<<<<<<<<<<
 *         b = self._inbuffer
 *         cdef int length = b.find(b'\r\n')
 */

static PyObject *__pyx_f_3lib_11RedisParser__get(struct __pyx_obj_3lib_RedisParser *__pyx_v_self, struct __pyx_obj_3lib_Task *__pyx_v_next) {
  PyObject *__pyx_v_b = NULL;
  int __pyx_v_length;
  PyObject *__pyx_v_response = NULL;
  PyObject *__pyx_v_rtype = NULL;
  struct __pyx_obj_3lib_Task *__pyx_v_task = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("_get", 0);

  /* "extensions/lib/rparser.pyx":116
 * 
 *     cdef object _get(self, Task next):
 *         b = self._inbuffer             # <<<<<<<<<<<<<<
 *         cdef int length = b.find(b'\r\n')
 *         if length >= 0:
 */
  __pyx_t_1 = __pyx_v_self->_inbuffer;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_b = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "extensions/lib/rparser.pyx":117
 *     cdef object _get(self, Task next):
 *         b = self._inbuffer
 *         cdef int length = b.find(b'\r\n')             # <<<<<<<<<<<<<<
 *         if length >= 0:
 *             self._inbuffer, response = b[length+2:], bytes(b[:length])
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_b, __pyx_n_s_find); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_length = __pyx_t_3;

  /* "extensions/lib/rparser.pyx":118
 *         b = self._inbuffer
 *         cdef int length = b.find(b'\r\n')
 *         if length >= 0:             # <<<<<<<<<<<<<<
 *             self._inbuffer, response = b[length+2:], bytes(b[:length])
 *             rtype, response = response[:1], response[1:]
 */
  __pyx_t_4 = ((__pyx_v_length >= 0) != 0);
  if (__pyx_t_4) {

    /* "extensions/lib/rparser.pyx":119
 *         cdef int length = b.find(b'\r\n')
 *         if length >= 0:
 *             self._inbuffer, response = b[length+2:], bytes(b[:length])             # <<<<<<<<<<<<<<
 *             rtype, response = response[:1], response[1:]
 *             if rtype == RESPONSE_ERROR:
 */
    __pyx_t_2 = __Pyx_PyObject_GetSlice(__pyx_v_b, (__pyx_v_length + 2), 0, NULL, NULL, NULL, 1, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_b, 0, __pyx_v_length, NULL, NULL, NULL, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)(&PyBytes_Type)), __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->_inbuffer);
    __Pyx_DECREF(__pyx_v_self->_inbuffer);
    __pyx_v_self->_inbuffer = __pyx_t_2;
    __pyx_t_2 = 0;
    __pyx_v_response = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "extensions/lib/rparser.pyx":120
 *         if length >= 0:
 *             self._inbuffer, response = b[length+2:], bytes(b[:length])
 *             rtype, response = response[:1], response[1:]             # <<<<<<<<<<<<<<
 *             if rtype == RESPONSE_ERROR:
 *                 return self._responseError(response.decode('utf-8'))
 */
    __pyx_t_1 = PySequence_GetSlice(__pyx_v_response, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 120, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PySequence_GetSlice(__pyx_v_response, 1, PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 120, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_rtype = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_response, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "extensions/lib/rparser.pyx":121
 *             self._inbuffer, response = b[length+2:], bytes(b[:length])
 *             rtype, response = response[:1], response[1:]
 *             if rtype == RESPONSE_ERROR:             # <<<<<<<<<<<<<<
 *                 return self._responseError(response.decode('utf-8'))
 *             elif rtype == RESPONSE_INTEGER:
 */
    __pyx_t_4 = (__Pyx_PyBytes_Equals(__pyx_v_rtype, __pyx_v_3lib_RESPONSE_ERROR, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 121, __pyx_L1_error)
    __pyx_t_6 = (__pyx_t_4 != 0);
    if (__pyx_t_6) {

      /* "extensions/lib/rparser.pyx":122
 *             rtype, response = response[:1], response[1:]
 *             if rtype == RESPONSE_ERROR:
 *                 return self._responseError(response.decode('utf-8'))             # <<<<<<<<<<<<<<
 *             elif rtype == RESPONSE_INTEGER:
 *                 return long(response)
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_1 = __Pyx_decode_bytes(__pyx_v_response, 0, PY_SSIZE_T_MAX, NULL, NULL, PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 122, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_INCREF(__pyx_v_self->_responseError);
      __pyx_t_5 = __pyx_v_self->_responseError; __pyx_t_7 = NULL;
      if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
        }
      }
      if (!__pyx_t_7) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 122, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else {
        __pyx_t_8 = PyTuple_New(1+1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 122, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_7); __pyx_t_7 = NULL;
        __Pyx_GIVEREF(__pyx_t_1);
        PyTuple_SET_ITEM(__pyx_t_8, 0+1, __pyx_t_1);
        __pyx_t_1 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 122, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      goto __pyx_L0;

      /* "extensions/lib/rparser.pyx":121
 *             self._inbuffer, response = b[length+2:], bytes(b[:length])
 *             rtype, response = response[:1], response[1:]
 *             if rtype == RESPONSE_ERROR:             # <<<<<<<<<<<<<<
 *                 return self._responseError(response.decode('utf-8'))
 *             elif rtype == RESPONSE_INTEGER:
 */
    }

    /* "extensions/lib/rparser.pyx":123
 *             if rtype == RESPONSE_ERROR:
 *                 return self._responseError(response.decode('utf-8'))
 *             elif rtype == RESPONSE_INTEGER:             # <<<<<<<<<<<<<<
 *                 return long(response)
 *             elif rtype == RESPONSE_STATUS:
 */
    __pyx_t_6 = (__Pyx_PyBytes_Equals(__pyx_v_rtype, __pyx_v_3lib_RESPONSE_INTEGER, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 123, __pyx_L1_error)
    __pyx_t_4 = (__pyx_t_6 != 0);
    if (__pyx_t_4) {

      /* "extensions/lib/rparser.pyx":124
 *                 return self._responseError(response.decode('utf-8'))
 *             elif rtype == RESPONSE_INTEGER:
 *                 return long(response)             # <<<<<<<<<<<<<<
 *             elif rtype == RESPONSE_STATUS:
 *                 return response
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 124, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_v_response);
      __Pyx_GIVEREF(__pyx_v_response);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_response);
      __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)(&PyLong_Type)), __pyx_t_2, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 124, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_5;
      __pyx_t_5 = 0;
      goto __pyx_L0;

      /* "extensions/lib/rparser.pyx":123
 *             if rtype == RESPONSE_ERROR:
 *                 return self._responseError(response.decode('utf-8'))
 *             elif rtype == RESPONSE_INTEGER:             # <<<<<<<<<<<<<<
 *                 return long(response)
 *             elif rtype == RESPONSE_STATUS:
 */
    }

    /* "extensions/lib/rparser.pyx":125
 *             elif rtype == RESPONSE_INTEGER:
 *                 return long(response)
 *             elif rtype == RESPONSE_STATUS:             # <<<<<<<<<<<<<<
 *                 return response
 *             elif rtype == RESPONSE_STRING:
 */
    __pyx_t_4 = (__Pyx_PyBytes_Equals(__pyx_v_rtype, __pyx_v_3lib_RESPONSE_STATUS, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 125, __pyx_L1_error)
    __pyx_t_6 = (__pyx_t_4 != 0);
    if (__pyx_t_6) {

      /* "extensions/lib/rparser.pyx":126
 *                 return long(response)
 *             elif rtype == RESPONSE_STATUS:
 *                 return response             # <<<<<<<<<<<<<<
 *             elif rtype == RESPONSE_STRING:
 *                 task = Task(long(response), next)
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_response);
      __pyx_r = __pyx_v_response;
      goto __pyx_L0;

      /* "extensions/lib/rparser.pyx":125
 *             elif rtype == RESPONSE_INTEGER:
 *                 return long(response)
 *             elif rtype == RESPONSE_STATUS:             # <<<<<<<<<<<<<<
 *                 return response
 *             elif rtype == RESPONSE_STRING:
 */
    }

    /* "extensions/lib/rparser.pyx":127
 *             elif rtype == RESPONSE_STATUS:
 *                 return response
 *             elif rtype == RESPONSE_STRING:             # <<<<<<<<<<<<<<
 *                 task = Task(long(response), next)
 *                 return task.decode(self, False)
 */
    __pyx_t_6 = (__Pyx_PyBytes_Equals(__pyx_v_rtype, __pyx_v_3lib_RESPONSE_STRING, Py_EQ)); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(0, 127, __pyx_L1_error)
    __pyx_t_4 = (__pyx_t_6 != 0);
    if (__pyx_t_4) {

      /* "extensions/lib/rparser.pyx":128
 *                 return response
 *             elif rtype == RESPONSE_STRING:
 *                 task = Task(long(response), next)             # <<<<<<<<<<<<<<
 *                 return task.decode(self, False)
 *             elif rtype == RESPONSE_ARRAY:
 */
      __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 128, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_v_response);
      __Pyx_GIVEREF(__pyx_v_response);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_response);
      __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)(&PyLong_Type)), __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 128, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = PyTuple_New(2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 128, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_2);
      __Pyx_INCREF(((PyObject *)__pyx_v_next));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_next));
      PyTuple_SET_ITEM(__pyx_t_5, 1, ((PyObject *)__pyx_v_next));
      __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_3lib_Task), __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 128, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_v_task = ((struct __pyx_obj_3lib_Task *)__pyx_t_2);
      __pyx_t_2 = 0;

      /* "extensions/lib/rparser.pyx":129
 *             elif rtype == RESPONSE_STRING:
 *                 task = Task(long(response), next)
 *                 return task.decode(self, False)             # <<<<<<<<<<<<<<
 *             elif rtype == RESPONSE_ARRAY:
 *                 task = ArrayTask(long(response), next)
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_2 = ((struct __pyx_vtabstruct_3lib_Task *)__pyx_v_task->__pyx_vtab)->decode(__pyx_v_task, __pyx_v_self, Py_False); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 129, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      goto __pyx_L0;

      /* "extensions/lib/rparser.pyx":127
 *             elif rtype == RESPONSE_STATUS:
 *                 return response
 *             elif rtype == RESPONSE_STRING:             # <<<<<<<<<<<<<<
 *                 task = Task(long(response), next)
 *                 return task.decode(self, False)
 */
    }

    /* "extensions/lib/rparser.pyx":130
 *                 task = Task(long(response), next)
 *                 return task.decode(self, False)
 *             elif rtype == RESPONSE_ARRAY:             # <<<<<<<<<<<<<<
 *                 task = ArrayTask(long(response), next)
 *                 return task.decode(self, False)
 */
    __pyx_t_4 = (__Pyx_PyBytes_Equals(__pyx_v_rtype, __pyx_v_3lib_RESPONSE_ARRAY, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 130, __pyx_L1_error)
    __pyx_t_6 = (__pyx_t_4 != 0);
    if (__pyx_t_6) {

      /* "extensions/lib/rparser.pyx":131
 *                 return task.decode(self, False)
 *             elif rtype == RESPONSE_ARRAY:
 *                 task = ArrayTask(long(response), next)             # <<<<<<<<<<<<<<
 *                 return task.decode(self, False)
 *             else:
 */
      __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 131, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_v_response);
      __Pyx_GIVEREF(__pyx_v_response);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_response);
      __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)(&PyLong_Type)), __pyx_t_2, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 131, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 131, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5);
      __Pyx_INCREF(((PyObject *)__pyx_v_next));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_next));
      PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)__pyx_v_next));
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_3lib_ArrayTask), __pyx_t_2, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 131, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_v_task = ((struct __pyx_obj_3lib_Task *)__pyx_t_5);
      __pyx_t_5 = 0;

      /* "extensions/lib/rparser.pyx":132
 *             elif rtype == RESPONSE_ARRAY:
 *                 task = ArrayTask(long(response), next)
 *                 return task.decode(self, False)             # <<<<<<<<<<<<<<
 *             else:
 *                 # Clear the buffer and raise
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_t_5 = ((struct __pyx_vtabstruct_3lib_Task *)__pyx_v_task->__pyx_vtab)->decode(__pyx_v_task, __pyx_v_self, Py_False); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 132, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_r = __pyx_t_5;
      __pyx_t_5 = 0;
      goto __pyx_L0;

      /* "extensions/lib/rparser.pyx":130
 *                 task = Task(long(response), next)
 *                 return task.decode(self, False)
 *             elif rtype == RESPONSE_ARRAY:             # <<<<<<<<<<<<<<
 *                 task = ArrayTask(long(response), next)
 *                 return task.decode(self, False)
 */
    }

    /* "extensions/lib/rparser.pyx":135
 *             else:
 *                 # Clear the buffer and raise
 *                 self._inbuffer = bytearray()             # <<<<<<<<<<<<<<
 *                 raise self._protocolError('Protocol Error')
 *         else:
 */
    /*else*/ {
      __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)(&PyByteArray_Type)), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 135, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_v_self->_inbuffer);
      __Pyx_DECREF(__pyx_v_self->_inbuffer);
      __pyx_v_self->_inbuffer = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "extensions/lib/rparser.pyx":136
 *                 # Clear the buffer and raise
 *                 self._inbuffer = bytearray()
 *                 raise self._protocolError('Protocol Error')             # <<<<<<<<<<<<<<
 *         else:
 *             return False
 */
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_v_self->_protocolError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 136, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(0, 136, __pyx_L1_error)
    }

    /* "extensions/lib/rparser.pyx":118
 *         b = self._inbuffer
 *         cdef int length = b.find(b'\r\n')
 *         if length >= 0:             # <<<<<<<<<<<<<<
 *             self._inbuffer, response = b[length+2:], bytes(b[:length])
 *             rtype, response = response[:1], response[1:]
 */
  }

  /* "extensions/lib/rparser.pyx":138
 *                 raise self._protocolError('Protocol Error')
 *         else:
 *             return False             # <<<<<<<<<<<<<<
 * 
 *     cdef object _resume(self, Task task, object result):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;
  }

  /* "extensions/lib/rparser.pyx":115
 *             yield v
 * 
 *     cdef object _get(self, Task next):             # <<<<<<<<<<<<<<
 *         b = self._inbuffer
 *         cdef int length = b.find(b'\r\n')
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("lib.RedisParser._get", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_b);
  __Pyx_XDECREF(__pyx_v_response);
  __Pyx_XDECREF(__pyx_v_rtype);
  __Pyx_XDECREF((PyObject *)__pyx_v_task);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/rparser.pyx":140
 *             return False
 * 
 *     cdef object _resume(self, Task task, object result):             # <<<<<<<<<<<<<<
 *         result = task.decode(self, result)
 *         if result is not False and task._next:
 */

static PyObject *__pyx_f_3lib_11RedisParser__resume(struct __pyx_obj_3lib_RedisParser *__pyx_v_self, struct __pyx_obj_3lib_Task *__pyx_v_task, PyObject *__pyx_v_result) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  __Pyx_RefNannySetupContext("_resume", 0);
  __Pyx_INCREF(__pyx_v_result);

  /* "extensions/lib/rparser.pyx":141
 * 
 *     cdef object _resume(self, Task task, object result):
 *         result = task.decode(self, result)             # <<<<<<<<<<<<<<
 *         if result is not False and task._next:
 *             return self._resume(task._next, result)
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_3lib_Task *)__pyx_v_task->__pyx_vtab)->decode(__pyx_v_task, __pyx_v_self, __pyx_v_result); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "extensions/lib/rparser.pyx":142
 *     cdef object _resume(self, Task task, object result):
 *         result = task.decode(self, result)
 *         if result is not False and task._next:             # <<<<<<<<<<<<<<
 *             return self._resume(task._next, result)
 *         else:
 */
  __pyx_t_3 = (__pyx_v_result != Py_False);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_task->_next)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 142, __pyx_L1_error)
  __pyx_t_2 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "extensions/lib/rparser.pyx":143
 *         result = task.decode(self, result)
 *         if result is not False and task._next:
 *             return self._resume(task._next, result)             # <<<<<<<<<<<<<<
 *         else:
 *             return result
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_1 = ((PyObject *)__pyx_v_task->_next);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_5 = ((struct __pyx_vtabstruct_3lib_RedisParser *)__pyx_v_self->__pyx_vtab)->_resume(__pyx_v_self, ((struct __pyx_obj_3lib_Task *)__pyx_t_1), __pyx_v_result); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 143, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "extensions/lib/rparser.pyx":142
 *     cdef object _resume(self, Task task, object result):
 *         result = task.decode(self, result)
 *         if result is not False and task._next:             # <<<<<<<<<<<<<<
 *             return self._resume(task._next, result)
 *         else:
 */
  }

  /* "extensions/lib/rparser.pyx":145
 *             return self._resume(task._next, result)
 *         else:
 *             return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(__pyx_v_result);
    __pyx_r = __pyx_v_result;
    goto __pyx_L0;
  }

  /* "extensions/lib/rparser.pyx":140
 *             return False
 * 
 *     cdef object _resume(self, Task task, object result):             # <<<<<<<<<<<<<<
 *         result = task.decode(self, result)
 *         if result is not False and task._next:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("lib.RedisParser._resume", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/rparser.pyx":152
 *     cdef Task _next
 * 
 *     def __cinit__(self, long length, Task next):             # <<<<<<<<<<<<<<
 *         self._length = length
 *         self._next = next
 */

/* Python wrapper */
static int __pyx_pw_3lib_4Task_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_3lib_4Task_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  long __pyx_v_length;
  struct __pyx_obj_3lib_Task *__pyx_v_next = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_length,&__pyx_n_s_next,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_length)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_next)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, 1); __PYX_ERR(0, 152, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(0, 152, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_length = __Pyx_PyInt_As_long(values[0]); if (unlikely((__pyx_v_length == (long)-1) && PyErr_Occurred())) __PYX_ERR(0, 152, __pyx_L3_error)
    __pyx_v_next = ((struct __pyx_obj_3lib_Task *)values[1]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 152, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("lib.Task.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_next), __pyx_ptype_3lib_Task, 1, "next", 0))) __PYX_ERR(0, 152, __pyx_L1_error)
  __pyx_r = __pyx_pf_3lib_4Task___cinit__(((struct __pyx_obj_3lib_Task *)__pyx_v_self), __pyx_v_length, __pyx_v_next);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_3lib_4Task___cinit__(struct __pyx_obj_3lib_Task *__pyx_v_self, long __pyx_v_length, struct __pyx_obj_3lib_Task *__pyx_v_next) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "extensions/lib/rparser.pyx":153
 * 
 *     def __cinit__(self, long length, Task next):
 *         self._length = length             # <<<<<<<<<<<<<<
 *         self._next = next
 * 
 */
  __pyx_v_self->_length = __pyx_v_length;

  /* "extensions/lib/rparser.pyx":154
 *     def __cinit__(self, long length, Task next):
 *         self._length = length
 *         self._next = next             # <<<<<<<<<<<<<<
 * 
 *     cdef object decode(self, RedisParser parser, object result):
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_next));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_next));
  __Pyx_GOTREF(__pyx_v_self->_next);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->_next));
  __pyx_v_self->_next = __pyx_v_next;

  /* "extensions/lib/rparser.pyx":152
 *     cdef Task _next
 * 
 *     def __cinit__(self, long length, Task next):             # <<<<<<<<<<<<<<
 *         self._length = length
 *         self._next = next
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/rparser.pyx":156
 *         self._next = next
 * 
 *     cdef object decode(self, RedisParser parser, object result):             # <<<<<<<<<<<<<<
 *         cdef long length = self._length
 *         cdef bytes chunk
 */

static PyObject *__pyx_f_3lib_4Task_decode(struct __pyx_obj_3lib_Task *__pyx_v_self, struct __pyx_obj_3lib_RedisParser *__pyx_v_parser, CYTHON_UNUSED PyObject *__pyx_v_result) {
  long __pyx_v_length;
  PyObject *__pyx_v_chunk = 0;
  PyObject *__pyx_v_b = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannySetupContext("decode", 0);

  /* "extensions/lib/rparser.pyx":157
 * 
 *     cdef object decode(self, RedisParser parser, object result):
 *         cdef long length = self._length             # <<<<<<<<<<<<<<
 *         cdef bytes chunk
 *         parser._current = None
 */
  __pyx_t_1 = __pyx_v_self->_length;
  __pyx_v_length = __pyx_t_1;

  /* "extensions/lib/rparser.pyx":159
 *         cdef long length = self._length
 *         cdef bytes chunk
 *         parser._current = None             # <<<<<<<<<<<<<<
 *         if length >= 0:
 *             b = parser._inbuffer
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_parser->_current);
  __Pyx_DECREF(((PyObject *)__pyx_v_parser->_current));
  __pyx_v_parser->_current = ((struct __pyx_obj_3lib_Task *)Py_None);

  /* "extensions/lib/rparser.pyx":160
 *         cdef bytes chunk
 *         parser._current = None
 *         if length >= 0:             # <<<<<<<<<<<<<<
 *             b = parser._inbuffer
 *             if len(b) >= length+2:
 */
  __pyx_t_2 = ((__pyx_v_length >= 0) != 0);
  if (__pyx_t_2) {

    /* "extensions/lib/rparser.pyx":161
 *         parser._current = None
 *         if length >= 0:
 *             b = parser._inbuffer             # <<<<<<<<<<<<<<
 *             if len(b) >= length+2:
 *                 parser._inbuffer, chunk = b[length+2:], bytes(b[:length])
 */
    __pyx_t_3 = __pyx_v_parser->_inbuffer;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_b = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "extensions/lib/rparser.pyx":162
 *         if length >= 0:
 *             b = parser._inbuffer
 *             if len(b) >= length+2:             # <<<<<<<<<<<<<<
 *                 parser._inbuffer, chunk = b[length+2:], bytes(b[:length])
 *                 if parser._encoding:
 */
    __pyx_t_4 = PyObject_Length(__pyx_v_b); if (unlikely(__pyx_t_4 == -1)) __PYX_ERR(0, 162, __pyx_L1_error)
    __pyx_t_2 = ((__pyx_t_4 >= (__pyx_v_length + 2)) != 0);
    if (__pyx_t_2) {

      /* "extensions/lib/rparser.pyx":163
 *             b = parser._inbuffer
 *             if len(b) >= length+2:
 *                 parser._inbuffer, chunk = b[length+2:], bytes(b[:length])             # <<<<<<<<<<<<<<
 *                 if parser._encoding:
 *                     return chunk.decode(parser._encoding)
 */
      __pyx_t_3 = __Pyx_PyObject_GetSlice(__pyx_v_b, (__pyx_v_length + 2), 0, NULL, NULL, NULL, 1, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 163, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = __Pyx_PyObject_GetSlice(__pyx_v_b, 0, __pyx_v_length, NULL, NULL, NULL, 0, 1, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 163, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 163, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5);
      __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)(&PyBytes_Type)), __pyx_t_6, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 163, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GIVEREF(__pyx_t_3);
      __Pyx_GOTREF(__pyx_v_parser->_inbuffer);
      __Pyx_DECREF(__pyx_v_parser->_inbuffer);
      __pyx_v_parser->_inbuffer = __pyx_t_3;
      __pyx_t_3 = 0;
      __pyx_v_chunk = ((PyObject*)__pyx_t_5);
      __pyx_t_5 = 0;

      /* "extensions/lib/rparser.pyx":164
 *             if len(b) >= length+2:
 *                 parser._inbuffer, chunk = b[length+2:], bytes(b[:length])
 *                 if parser._encoding:             # <<<<<<<<<<<<<<
 *                     return chunk.decode(parser._encoding)
 *                 else:
 */
      __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_parser->_encoding); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 164, __pyx_L1_error)
      if (__pyx_t_2) {

        /* "extensions/lib/rparser.pyx":165
 *                 parser._inbuffer, chunk = b[length+2:], bytes(b[:length])
 *                 if parser._encoding:
 *                     return chunk.decode(parser._encoding)             # <<<<<<<<<<<<<<
 *                 else:
 *                     return chunk
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_chunk, __pyx_n_s_decode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 165, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_6 = NULL;
        if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_3))) {
          __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
          if (likely(__pyx_t_6)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
            __Pyx_INCREF(__pyx_t_6);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_3, function);
          }
        }
        if (!__pyx_t_6) {
          __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_parser->_encoding); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 165, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
        } else {
          __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 165, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_6); __pyx_t_6 = NULL;
          __Pyx_INCREF(__pyx_v_parser->_encoding);
          __Pyx_GIVEREF(__pyx_v_parser->_encoding);
          PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_v_parser->_encoding);
          __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 165, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        }
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_r = __pyx_t_5;
        __pyx_t_5 = 0;
        goto __pyx_L0;

        /* "extensions/lib/rparser.pyx":164
 *             if len(b) >= length+2:
 *                 parser._inbuffer, chunk = b[length+2:], bytes(b[:length])
 *                 if parser._encoding:             # <<<<<<<<<<<<<<
 *                     return chunk.decode(parser._encoding)
 *                 else:
 */
      }

      /* "extensions/lib/rparser.pyx":167
 *                     return chunk.decode(parser._encoding)
 *                 else:
 *                     return chunk             # <<<<<<<<<<<<<<
 *             else:
 *                 parser._current = self
 */
      /*else*/ {
        __Pyx_XDECREF(__pyx_r);
        __Pyx_INCREF(__pyx_v_chunk);
        __pyx_r = __pyx_v_chunk;
        goto __pyx_L0;
      }

      /* "extensions/lib/rparser.pyx":162
 *         if length >= 0:
 *             b = parser._inbuffer
 *             if len(b) >= length+2:             # <<<<<<<<<<<<<<
 *                 parser._inbuffer, chunk = b[length+2:], bytes(b[:length])
 *                 if parser._encoding:
 */
    }

    /* "extensions/lib/rparser.pyx":169
 *                     return chunk
 *             else:
 *                 parser._current = self             # <<<<<<<<<<<<<<
 *                 return False
 * 
 */
    /*else*/ {
      __Pyx_INCREF(((PyObject *)__pyx_v_self));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
      __Pyx_GOTREF(__pyx_v_parser->_current);
      __Pyx_DECREF(((PyObject *)__pyx_v_parser->_current));
      __pyx_v_parser->_current = __pyx_v_self;

      /* "extensions/lib/rparser.pyx":170
 *             else:
 *                 parser._current = self
 *                 return False             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(Py_False);
      __pyx_r = Py_False;
      goto __pyx_L0;
    }

    /* "extensions/lib/rparser.pyx":160
 *         cdef bytes chunk
 *         parser._current = None
 *         if length >= 0:             # <<<<<<<<<<<<<<
 *             b = parser._inbuffer
 *             if len(b) >= length+2:
 */
  }

  /* "extensions/lib/rparser.pyx":156
 *         self._next = next
 * 
 *     cdef object decode(self, RedisParser parser, object result):             # <<<<<<<<<<<<<<
 *         cdef long length = self._length
 *         cdef bytes chunk
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_AddTraceback("lib.Task.decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_chunk);
  __Pyx_XDECREF(__pyx_v_b);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/rparser.pyx":176
 *     cdef list _response
 * 
 *     cdef object decode(self, RedisParser parser, object result):             # <<<<<<<<<<<<<<
 *         cdef long length = self._length
 *         cdef list response = self._response
 */

static PyObject *__pyx_f_3lib_9ArrayTask_decode(struct __pyx_obj_3lib_ArrayTask *__pyx_v_self, struct __pyx_obj_3lib_RedisParser *__pyx_v_parser, PyObject *__pyx_v_result) {
  long __pyx_v_length;
  PyObject *__pyx_v_response = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  __Pyx_RefNannySetupContext("decode", 0);
  __Pyx_INCREF(__pyx_v_result);

  /* "extensions/lib/rparser.pyx":177
 * 
 *     cdef object decode(self, RedisParser parser, object result):
 *         cdef long length = self._length             # <<<<<<<<<<<<<<
 *         cdef list response = self._response
 *         parser._current = None
 */
  __pyx_t_1 = __pyx_v_self->__pyx_base._length;
  __pyx_v_length = __pyx_t_1;

  /* "extensions/lib/rparser.pyx":178
 *     cdef object decode(self, RedisParser parser, object result):
 *         cdef long length = self._length
 *         cdef list response = self._response             # <<<<<<<<<<<<<<
 *         parser._current = None
 *         if response is None:
 */
  __pyx_t_2 = __pyx_v_self->_response;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_v_response = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "extensions/lib/rparser.pyx":179
 *         cdef long length = self._length
 *         cdef list response = self._response
 *         parser._current = None             # <<<<<<<<<<<<<<
 *         if response is None:
 *             self._response = response = []
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_parser->_current);
  __Pyx_DECREF(((PyObject *)__pyx_v_parser->_current));
  __pyx_v_parser->_current = ((struct __pyx_obj_3lib_Task *)Py_None);

  /* "extensions/lib/rparser.pyx":180
 *         cdef list response = self._response
 *         parser._current = None
 *         if response is None:             # <<<<<<<<<<<<<<
 *             self._response = response = []
 *         if length >= 0:
 */
  __pyx_t_3 = (__pyx_v_response == ((PyObject*)Py_None));
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "extensions/lib/rparser.pyx":181
 *         parser._current = None
 *         if response is None:
 *             self._response = response = []             # <<<<<<<<<<<<<<
 *         if length >= 0:
 *             if result is not False:
 */
    __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->_response);
    __Pyx_DECREF(__pyx_v_self->_response);
    __pyx_v_self->_response = __pyx_t_2;
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_response, __pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "extensions/lib/rparser.pyx":180
 *         cdef list response = self._response
 *         parser._current = None
 *         if response is None:             # <<<<<<<<<<<<<<
 *             self._response = response = []
 *         if length >= 0:
 */
  }

  /* "extensions/lib/rparser.pyx":182
 *         if response is None:
 *             self._response = response = []
 *         if length >= 0:             # <<<<<<<<<<<<<<
 *             if result is not False:
 *                 response.append(result)
 */
  __pyx_t_4 = ((__pyx_v_length >= 0) != 0);
  if (__pyx_t_4) {

    /* "extensions/lib/rparser.pyx":183
 *             self._response = response = []
 *         if length >= 0:
 *             if result is not False:             # <<<<<<<<<<<<<<
 *                 response.append(result)
 *             while len(response) < length:
 */
    __pyx_t_4 = (__pyx_v_result != Py_False);
    __pyx_t_3 = (__pyx_t_4 != 0);
    if (__pyx_t_3) {

      /* "extensions/lib/rparser.pyx":184
 *         if length >= 0:
 *             if result is not False:
 *                 response.append(result)             # <<<<<<<<<<<<<<
 *             while len(response) < length:
 *                 result = parser._get(self)
 */
      if (unlikely(__pyx_v_response == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "append");
        __PYX_ERR(0, 184, __pyx_L1_error)
      }
      __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_response, __pyx_v_result); if (unlikely(__pyx_t_5 == -1)) __PYX_ERR(0, 184, __pyx_L1_error)

      /* "extensions/lib/rparser.pyx":183
 *             self._response = response = []
 *         if length >= 0:
 *             if result is not False:             # <<<<<<<<<<<<<<
 *                 response.append(result)
 *             while len(response) < length:
 */
    }

    /* "extensions/lib/rparser.pyx":185
 *             if result is not False:
 *                 response.append(result)
 *             while len(response) < length:             # <<<<<<<<<<<<<<
 *                 result = parser._get(self)
 *                 if result is False:
 */
    while (1) {
      if (unlikely(__pyx_v_response == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
        __PYX_ERR(0, 185, __pyx_L1_error)
      }
      __pyx_t_6 = PyList_GET_SIZE(__pyx_v_response); if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(0, 185, __pyx_L1_error)
      __pyx_t_3 = ((__pyx_t_6 < __pyx_v_length) != 0);
      if (!__pyx_t_3) break;

      /* "extensions/lib/rparser.pyx":186
 *                 response.append(result)
 *             while len(response) < length:
 *                 result = parser._get(self)             # <<<<<<<<<<<<<<
 *                 if result is False:
 *                     break
 */
      __pyx_t_2 = ((struct __pyx_vtabstruct_3lib_RedisParser *)__pyx_v_parser->__pyx_vtab)->_get(__pyx_v_parser, ((struct __pyx_obj_3lib_Task *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 186, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "extensions/lib/rparser.pyx":187
 *             while len(response) < length:
 *                 result = parser._get(self)
 *                 if result is False:             # <<<<<<<<<<<<<<
 *                     break
 *                 response.append(result)
 */
      __pyx_t_3 = (__pyx_v_result == Py_False);
      __pyx_t_4 = (__pyx_t_3 != 0);
      if (__pyx_t_4) {

        /* "extensions/lib/rparser.pyx":188
 *                 result = parser._get(self)
 *                 if result is False:
 *                     break             # <<<<<<<<<<<<<<
 *                 response.append(result)
 *             if len(response) == length:
 */
        goto __pyx_L7_break;

        /* "extensions/lib/rparser.pyx":187
 *             while len(response) < length:
 *                 result = parser._get(self)
 *                 if result is False:             # <<<<<<<<<<<<<<
 *                     break
 *                 response.append(result)
 */
      }

      /* "extensions/lib/rparser.pyx":189
 *                 if result is False:
 *                     break
 *                 response.append(result)             # <<<<<<<<<<<<<<
 *             if len(response) == length:
 *                 parser._current = None
 */
      if (unlikely(__pyx_v_response == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "append");
        __PYX_ERR(0, 189, __pyx_L1_error)
      }
      __pyx_t_5 = __Pyx_PyList_Append(__pyx_v_response, __pyx_v_result); if (unlikely(__pyx_t_5 == -1)) __PYX_ERR(0, 189, __pyx_L1_error)
    }
    __pyx_L7_break:;

    /* "extensions/lib/rparser.pyx":190
 *                     break
 *                 response.append(result)
 *             if len(response) == length:             # <<<<<<<<<<<<<<
 *                 parser._current = None
 *                 return response
 */
    if (unlikely(__pyx_v_response == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 190, __pyx_L1_error)
    }
    __pyx_t_6 = PyList_GET_SIZE(__pyx_v_response); if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(0, 190, __pyx_L1_error)
    __pyx_t_4 = ((__pyx_t_6 == __pyx_v_length) != 0);
    if (__pyx_t_4) {

      /* "extensions/lib/rparser.pyx":191
 *                 response.append(result)
 *             if len(response) == length:
 *                 parser._current = None             # <<<<<<<<<<<<<<
 *                 return response
 *             elif not parser._current:
 */
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      __Pyx_GOTREF(__pyx_v_parser->_current);
      __Pyx_DECREF(((PyObject *)__pyx_v_parser->_current));
      __pyx_v_parser->_current = ((struct __pyx_obj_3lib_Task *)Py_None);

      /* "extensions/lib/rparser.pyx":192
 *             if len(response) == length:
 *                 parser._current = None
 *                 return response             # <<<<<<<<<<<<<<
 *             elif not parser._current:
 *                 parser._current = self
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_response);
      __pyx_r = __pyx_v_response;
      goto __pyx_L0;

      /* "extensions/lib/rparser.pyx":190
 *                     break
 *                 response.append(result)
 *             if len(response) == length:             # <<<<<<<<<<<<<<
 *                 parser._current = None
 *                 return response
 */
    }

    /* "extensions/lib/rparser.pyx":193
 *                 parser._current = None
 *                 return response
 *             elif not parser._current:             # <<<<<<<<<<<<<<
 *                 parser._current = self
 *             return False
 */
    __pyx_t_4 = __Pyx_PyObject_IsTrue(((PyObject *)__pyx_v_parser->_current)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 193, __pyx_L1_error)
    __pyx_t_3 = ((!__pyx_t_4) != 0);
    if (__pyx_t_3) {

      /* "extensions/lib/rparser.pyx":194
 *                 return response
 *             elif not parser._current:
 *                 parser._current = self             # <<<<<<<<<<<<<<
 *             return False
 */
      __Pyx_INCREF(((PyObject *)__pyx_v_self));
      __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
      __Pyx_GOTREF(__pyx_v_parser->_current);
      __Pyx_DECREF(((PyObject *)__pyx_v_parser->_current));
      __pyx_v_parser->_current = ((struct __pyx_obj_3lib_Task *)__pyx_v_self);

      /* "extensions/lib/rparser.pyx":193
 *                 parser._current = None
 *                 return response
 *             elif not parser._current:             # <<<<<<<<<<<<<<
 *                 parser._current = self
 *             return False
 */
    }

    /* "extensions/lib/rparser.pyx":195
 *             elif not parser._current:
 *                 parser._current = self
 *             return False             # <<<<<<<<<<<<<<
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_False);
    __pyx_r = Py_False;
    goto __pyx_L0;

    /* "extensions/lib/rparser.pyx":182
 *         if response is None:
 *             self._response = response = []
 *         if length >= 0:             # <<<<<<<<<<<<<<
 *             if result is not False:
 *                 response.append(result)
 */
  }

  /* "extensions/lib/rparser.pyx":176
 *     cdef list _response
 * 
 *     cdef object decode(self, RedisParser parser, object result):             # <<<<<<<<<<<<<<
 *         cdef long length = self._length
 *         cdef list response = self._response
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("lib.ArrayTask.decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_response);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/websocket.pyx":15
 *     cdef object _body
 * 
 *     def __cinit__(self, int opcode, bint final, int payload_length):             # <<<<<<<<<<<<<<
 *         self._opcode = opcode
 *         self._final = final
 */

/* Python wrapper */
static int __pyx_pw_3lib_5Frame_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_3lib_5Frame_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_opcode;
  int __pyx_v_final;
  int __pyx_v_payload_length;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_opcode,&__pyx_n_s_final,&__pyx_n_s_payload_length,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_opcode)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_final)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 1); __PYX_ERR(1, 15, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_payload_length)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, 2); __PYX_ERR(1, 15, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(1, 15, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_opcode = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_opcode == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 15, __pyx_L3_error)
    __pyx_v_final = __Pyx_PyObject_IsTrue(values[1]); if (unlikely((__pyx_v_final == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 15, __pyx_L3_error)
    __pyx_v_payload_length = __Pyx_PyInt_As_int(values[2]); if (unlikely((__pyx_v_payload_length == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 15, __pyx_L3_error)
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 15, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("lib.Frame.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_3lib_5Frame___cinit__(((struct __pyx_obj_3lib_Frame *)__pyx_v_self), __pyx_v_opcode, __pyx_v_final, __pyx_v_payload_length);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_3lib_5Frame___cinit__(struct __pyx_obj_3lib_Frame *__pyx_v_self, int __pyx_v_opcode, int __pyx_v_final, int __pyx_v_payload_length) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "extensions/lib/websocket.pyx":16
 * 
 *     def __cinit__(self, int opcode, bint final, int payload_length):
 *         self._opcode = opcode             # <<<<<<<<<<<<<<
 *         self._final = final
 *         self._payload_length = payload_length
 */
  __pyx_v_self->_opcode = __pyx_v_opcode;

  /* "extensions/lib/websocket.pyx":17
 *     def __cinit__(self, int opcode, bint final, int payload_length):
 *         self._opcode = opcode
 *         self._final = final             # <<<<<<<<<<<<<<
 *         self._payload_length = payload_length
 * 
 */
  __pyx_v_self->_final = __pyx_v_final;

  /* "extensions/lib/websocket.pyx":18
 *         self._opcode = opcode
 *         self._final = final
 *         self._payload_length = payload_length             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_v_self->_payload_length = __pyx_v_payload_length;

  /* "extensions/lib/websocket.pyx":15
 *     cdef object _body
 * 
 *     def __cinit__(self, int opcode, bint final, int payload_length):             # <<<<<<<<<<<<<<
 *         self._opcode = opcode
 *         self._final = final
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/websocket.pyx":21
 * 
 *     @property
 *     def final(self):             # <<<<<<<<<<<<<<
 *         return self._final
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_5Frame_5final_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_3lib_5Frame_5final_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_3lib_5Frame_5final___get__(((struct __pyx_obj_3lib_Frame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_5Frame_5final___get__(struct __pyx_obj_3lib_Frame *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "extensions/lib/websocket.pyx":22
 *     @property
 *     def final(self):
 *         return self._final             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong(__pyx_v_self->_final); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "extensions/lib/websocket.pyx":21
 * 
 *     @property
 *     def final(self):             # <<<<<<<<<<<<<<
 *         return self._final
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("lib.Frame.final.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/websocket.pyx":25
 * 
 *     @property
 *     def opcode(self):             # <<<<<<<<<<<<<<
 *         return self._opcode
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_5Frame_6opcode_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_3lib_5Frame_6opcode_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_3lib_5Frame_6opcode___get__(((struct __pyx_obj_3lib_Frame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_5Frame_6opcode___get__(struct __pyx_obj_3lib_Frame *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "extensions/lib/websocket.pyx":26
 *     @property
 *     def opcode(self):
 *         return self._opcode             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->_opcode); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "extensions/lib/websocket.pyx":25
 * 
 *     @property
 *     def opcode(self):             # <<<<<<<<<<<<<<
 *         return self._opcode
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("lib.Frame.opcode.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/websocket.pyx":29
 * 
 *     @property
 *     def body(self):             # <<<<<<<<<<<<<<
 *         return self._body
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_5Frame_4body_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_3lib_5Frame_4body_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_3lib_5Frame_4body___get__(((struct __pyx_obj_3lib_Frame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_5Frame_4body___get__(struct __pyx_obj_3lib_Frame *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "extensions/lib/websocket.pyx":30
 *     @property
 *     def body(self):
 *         return self._body             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_body);
  __pyx_r = __pyx_v_self->_body;
  goto __pyx_L0;

  /* "extensions/lib/websocket.pyx":29
 * 
 *     @property
 *     def body(self):             # <<<<<<<<<<<<<<
 *         return self._body
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/websocket.pyx":33
 * 
 *     @property
 *     def masking_key(self):             # <<<<<<<<<<<<<<
 *         return self._masking_key
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_5Frame_11masking_key_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_3lib_5Frame_11masking_key_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_3lib_5Frame_11masking_key___get__(((struct __pyx_obj_3lib_Frame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_5Frame_11masking_key___get__(struct __pyx_obj_3lib_Frame *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "extensions/lib/websocket.pyx":34
 *     @property
 *     def masking_key(self):
 *         return self._masking_key             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_masking_key);
  __pyx_r = __pyx_v_self->_masking_key;
  goto __pyx_L0;

  /* "extensions/lib/websocket.pyx":33
 * 
 *     @property
 *     def masking_key(self):             # <<<<<<<<<<<<<<
 *         return self._masking_key
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/websocket.pyx":37
 * 
 *     @property
 *     def is_message(self):             # <<<<<<<<<<<<<<
 *         return self._opcode == 1
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_5Frame_10is_message_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_3lib_5Frame_10is_message_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_3lib_5Frame_10is_message___get__(((struct __pyx_obj_3lib_Frame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_5Frame_10is_message___get__(struct __pyx_obj_3lib_Frame *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "extensions/lib/websocket.pyx":38
 *     @property
 *     def is_message(self):
 *         return self._opcode == 1             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong((__pyx_v_self->_opcode == 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "extensions/lib/websocket.pyx":37
 * 
 *     @property
 *     def is_message(self):             # <<<<<<<<<<<<<<
 *         return self._opcode == 1
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("lib.Frame.is_message.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/websocket.pyx":41
 * 
 *     @property
 *     def is_bytes(self):             # <<<<<<<<<<<<<<
 *         return self._opcode == 2
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_5Frame_8is_bytes_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_3lib_5Frame_8is_bytes_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_3lib_5Frame_8is_bytes___get__(((struct __pyx_obj_3lib_Frame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_5Frame_8is_bytes___get__(struct __pyx_obj_3lib_Frame *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "extensions/lib/websocket.pyx":42
 *     @property
 *     def is_bytes(self):
 *         return self._opcode == 2             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong((__pyx_v_self->_opcode == 2)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "extensions/lib/websocket.pyx":41
 * 
 *     @property
 *     def is_bytes(self):             # <<<<<<<<<<<<<<
 *         return self._opcode == 2
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("lib.Frame.is_bytes.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/websocket.pyx":45
 * 
 *     @property
 *     def is_close(self):             # <<<<<<<<<<<<<<
 *         return self._opcode == 8
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_5Frame_8is_close_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_3lib_5Frame_8is_close_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_3lib_5Frame_8is_close___get__(((struct __pyx_obj_3lib_Frame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_5Frame_8is_close___get__(struct __pyx_obj_3lib_Frame *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "extensions/lib/websocket.pyx":46
 *     @property
 *     def is_close(self):
 *         return self._opcode == 8             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong((__pyx_v_self->_opcode == 8)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "extensions/lib/websocket.pyx":45
 * 
 *     @property
 *     def is_close(self):             # <<<<<<<<<<<<<<
 *         return self._opcode == 8
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("lib.Frame.is_close.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/websocket.pyx":49
 * 
 *     @property
 *     def is_ping(self):             # <<<<<<<<<<<<<<
 *         return self._opcode == 9
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_5Frame_7is_ping_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_3lib_5Frame_7is_ping_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_3lib_5Frame_7is_ping___get__(((struct __pyx_obj_3lib_Frame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_5Frame_7is_ping___get__(struct __pyx_obj_3lib_Frame *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "extensions/lib/websocket.pyx":50
 *     @property
 *     def is_ping(self):
 *         return self._opcode == 9             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong((__pyx_v_self->_opcode == 9)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "extensions/lib/websocket.pyx":49
 * 
 *     @property
 *     def is_ping(self):             # <<<<<<<<<<<<<<
 *         return self._opcode == 9
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("lib.Frame.is_ping.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/websocket.pyx":53
 * 
 *     @property
 *     def is_pong(self):             # <<<<<<<<<<<<<<
 *         return self._opcode == 10
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_5Frame_7is_pong_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_3lib_5Frame_7is_pong_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_3lib_5Frame_7is_pong___get__(((struct __pyx_obj_3lib_Frame *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_5Frame_7is_pong___get__(struct __pyx_obj_3lib_Frame *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "extensions/lib/websocket.pyx":54
 *     @property
 *     def is_pong(self):
 *         return self._opcode == 10             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyBool_FromLong((__pyx_v_self->_opcode == 10)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "extensions/lib/websocket.pyx":53
 * 
 *     @property
 *     def is_pong(self):             # <<<<<<<<<<<<<<
 *         return self._opcode == 10
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("lib.Frame.is_pong.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/websocket.pyx":77
 *     cdef object _close_codes
 * 
 *     def __cinit__(self, int version, int kind, object ProtocolError,             # <<<<<<<<<<<<<<
 *                   extensions=None, protocols=None, close_codes=None):
 *         self._version = version
 */

/* Python wrapper */
static int __pyx_pw_3lib_11FrameParser_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_3lib_11FrameParser_1__cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  int __pyx_v_version;
  int __pyx_v_kind;
  PyObject *__pyx_v_ProtocolError = 0;
  PyObject *__pyx_v_extensions = 0;
  PyObject *__pyx_v_protocols = 0;
  PyObject *__pyx_v_close_codes = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_version,&__pyx_n_s_kind,&__pyx_n_s_ProtocolError,&__pyx_n_s_extensions,&__pyx_n_s_protocols,&__pyx_n_s_close_codes,0};
    PyObject* values[6] = {0,0,0,0,0,0};

    /* "extensions/lib/websocket.pyx":78
 * 
 *     def __cinit__(self, int version, int kind, object ProtocolError,
 *                   extensions=None, protocols=None, close_codes=None):             # <<<<<<<<<<<<<<
 *         self._version = version
 *         self.kind = kind
 */
    values[3] = ((PyObject *)Py_None);
    values[4] = ((PyObject *)Py_None);
    values[5] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_version)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_kind)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 6, 1); __PYX_ERR(1, 77, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ProtocolError)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 6, 2); __PYX_ERR(1, 77, __pyx_L3_error)
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_extensions);
          if (value) { values[3] = value; kw_args--; }
        }
        case  4:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_protocols);
          if (value) { values[4] = value; kw_args--; }
        }
        case  5:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_close_codes);
          if (value) { values[5] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(1, 77, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_version = __Pyx_PyInt_As_int(values[0]); if (unlikely((__pyx_v_version == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 77, __pyx_L3_error)
    __pyx_v_kind = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_kind == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 77, __pyx_L3_error)
    __pyx_v_ProtocolError = values[2];
    __pyx_v_extensions = values[3];
    __pyx_v_protocols = values[4];
    __pyx_v_close_codes = values[5];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 6, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 77, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("lib.FrameParser.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_3lib_11FrameParser___cinit__(((struct __pyx_obj_3lib_FrameParser *)__pyx_v_self), __pyx_v_version, __pyx_v_kind, __pyx_v_ProtocolError, __pyx_v_extensions, __pyx_v_protocols, __pyx_v_close_codes);

  /* "extensions/lib/websocket.pyx":77
 *     cdef object _close_codes
 * 
 *     def __cinit__(self, int version, int kind, object ProtocolError,             # <<<<<<<<<<<<<<
 *                   extensions=None, protocols=None, close_codes=None):
 *         self._version = version
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_3lib_11FrameParser___cinit__(struct __pyx_obj_3lib_FrameParser *__pyx_v_self, int __pyx_v_version, int __pyx_v_kind, PyObject *__pyx_v_ProtocolError, PyObject *__pyx_v_extensions, PyObject *__pyx_v_protocols, PyObject *__pyx_v_close_codes) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "extensions/lib/websocket.pyx":79
 *     def __cinit__(self, int version, int kind, object ProtocolError,
 *                   extensions=None, protocols=None, close_codes=None):
 *         self._version = version             # <<<<<<<<<<<<<<
 *         self.kind = kind
 *         self.frame = None
 */
  __pyx_v_self->_version = __pyx_v_version;

  /* "extensions/lib/websocket.pyx":80
 *                   extensions=None, protocols=None, close_codes=None):
 *         self._version = version
 *         self.kind = kind             # <<<<<<<<<<<<<<
 *         self.frame = None
 *         self.buffer = bytearray()
 */
  __pyx_v_self->kind = __pyx_v_kind;

  /* "extensions/lib/websocket.pyx":81
 *         self._version = version
 *         self.kind = kind
 *         self.frame = None             # <<<<<<<<<<<<<<
 *         self.buffer = bytearray()
 *         self.ProtocolError = ProtocolError
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->frame);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->frame));
  __pyx_v_self->frame = ((struct __pyx_obj_3lib_Frame *)Py_None);

  /* "extensions/lib/websocket.pyx":82
 *         self.kind = kind
 *         self.frame = None
 *         self.buffer = bytearray()             # <<<<<<<<<<<<<<
 *         self.ProtocolError = ProtocolError
 *         self._opcodes = (0, 1, 2, 8, 9, 10)
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)(&PyByteArray_Type)), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->buffer);
  __Pyx_DECREF(__pyx_v_self->buffer);
  __pyx_v_self->buffer = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "extensions/lib/websocket.pyx":83
 *         self.frame = None
 *         self.buffer = bytearray()
 *         self.ProtocolError = ProtocolError             # <<<<<<<<<<<<<<
 *         self._opcodes = (0, 1, 2, 8, 9, 10)
 *         self._encode_mask_length = 0
 */
  __Pyx_INCREF(__pyx_v_ProtocolError);
  __Pyx_GIVEREF(__pyx_v_ProtocolError);
  __Pyx_GOTREF(__pyx_v_self->ProtocolError);
  __Pyx_DECREF(__pyx_v_self->ProtocolError);
  __pyx_v_self->ProtocolError = __pyx_v_ProtocolError;

  /* "extensions/lib/websocket.pyx":84
 *         self.buffer = bytearray()
 *         self.ProtocolError = ProtocolError
 *         self._opcodes = (0, 1, 2, 8, 9, 10)             # <<<<<<<<<<<<<<
 *         self._encode_mask_length = 0
 *         self._decode_mask_length = 0
 */
  __Pyx_INCREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);
  __Pyx_GOTREF(__pyx_v_self->_opcodes);
  __Pyx_DECREF(__pyx_v_self->_opcodes);
  __pyx_v_self->_opcodes = __pyx_tuple__17;

  /* "extensions/lib/websocket.pyx":85
 *         self.ProtocolError = ProtocolError
 *         self._opcodes = (0, 1, 2, 8, 9, 10)
 *         self._encode_mask_length = 0             # <<<<<<<<<<<<<<
 *         self._decode_mask_length = 0
 *         if kind == 0:
 */
  __pyx_v_self->_encode_mask_length = 0;

  /* "extensions/lib/websocket.pyx":86
 *         self._opcodes = (0, 1, 2, 8, 9, 10)
 *         self._encode_mask_length = 0
 *         self._decode_mask_length = 0             # <<<<<<<<<<<<<<
 *         if kind == 0:
 *             self._decode_mask_length = 4
 */
  __pyx_v_self->_decode_mask_length = 0;

  /* "extensions/lib/websocket.pyx":87
 *         self._encode_mask_length = 0
 *         self._decode_mask_length = 0
 *         if kind == 0:             # <<<<<<<<<<<<<<
 *             self._decode_mask_length = 4
 *         elif kind == 1:
 */
  switch (__pyx_v_kind) {
    case 0:

    /* "extensions/lib/websocket.pyx":88
 *         self._decode_mask_length = 0
 *         if kind == 0:
 *             self._decode_mask_length = 4             # <<<<<<<<<<<<<<
 *         elif kind == 1:
 *             self._encode_mask_length = 4
 */
    __pyx_v_self->_decode_mask_length = 4;

    /* "extensions/lib/websocket.pyx":87
 *         self._encode_mask_length = 0
 *         self._decode_mask_length = 0
 *         if kind == 0:             # <<<<<<<<<<<<<<
 *             self._decode_mask_length = 4
 *         elif kind == 1:
 */
    break;

    /* "extensions/lib/websocket.pyx":89
 *         if kind == 0:
 *             self._decode_mask_length = 4
 *         elif kind == 1:             # <<<<<<<<<<<<<<
 *             self._encode_mask_length = 4
 *         elif kind == 3:
 */
    case 1:

    /* "extensions/lib/websocket.pyx":90
 *             self._decode_mask_length = 4
 *         elif kind == 1:
 *             self._encode_mask_length = 4             # <<<<<<<<<<<<<<
 *         elif kind == 3:
 *             self._decode_mask_length = 4
 */
    __pyx_v_self->_encode_mask_length = 4;

    /* "extensions/lib/websocket.pyx":89
 *         if kind == 0:
 *             self._decode_mask_length = 4
 *         elif kind == 1:             # <<<<<<<<<<<<<<
 *             self._encode_mask_length = 4
 *         elif kind == 3:
 */
    break;

    /* "extensions/lib/websocket.pyx":91
 *         elif kind == 1:
 *             self._encode_mask_length = 4
 *         elif kind == 3:             # <<<<<<<<<<<<<<
 *             self._decode_mask_length = 4
 *             self._encode_mask_length = 4
 */
    case 3:

    /* "extensions/lib/websocket.pyx":92
 *             self._encode_mask_length = 4
 *         elif kind == 3:
 *             self._decode_mask_length = 4             # <<<<<<<<<<<<<<
 *             self._encode_mask_length = 4
 *         self._extensions = extensions
 */
    __pyx_v_self->_decode_mask_length = 4;

    /* "extensions/lib/websocket.pyx":93
 *         elif kind == 3:
 *             self._decode_mask_length = 4
 *             self._encode_mask_length = 4             # <<<<<<<<<<<<<<
 *         self._extensions = extensions
 *         self._protocols = protocols
 */
    __pyx_v_self->_encode_mask_length = 4;

    /* "extensions/lib/websocket.pyx":91
 *         elif kind == 1:
 *             self._encode_mask_length = 4
 *         elif kind == 3:             # <<<<<<<<<<<<<<
 *             self._decode_mask_length = 4
 *             self._encode_mask_length = 4
 */
    break;
    default: break;
  }

  /* "extensions/lib/websocket.pyx":94
 *             self._decode_mask_length = 4
 *             self._encode_mask_length = 4
 *         self._extensions = extensions             # <<<<<<<<<<<<<<
 *         self._protocols = protocols
 *         self._close_codes = close_codes
 */
  __Pyx_INCREF(__pyx_v_extensions);
  __Pyx_GIVEREF(__pyx_v_extensions);
  __Pyx_GOTREF(__pyx_v_self->_extensions);
  __Pyx_DECREF(__pyx_v_self->_extensions);
  __pyx_v_self->_extensions = __pyx_v_extensions;

  /* "extensions/lib/websocket.pyx":95
 *             self._encode_mask_length = 4
 *         self._extensions = extensions
 *         self._protocols = protocols             # <<<<<<<<<<<<<<
 *         self._close_codes = close_codes
 * 
 */
  __Pyx_INCREF(__pyx_v_protocols);
  __Pyx_GIVEREF(__pyx_v_protocols);
  __Pyx_GOTREF(__pyx_v_self->_protocols);
  __Pyx_DECREF(__pyx_v_self->_protocols);
  __pyx_v_self->_protocols = __pyx_v_protocols;

  /* "extensions/lib/websocket.pyx":96
 *         self._extensions = extensions
 *         self._protocols = protocols
 *         self._close_codes = close_codes             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_INCREF(__pyx_v_close_codes);
  __Pyx_GIVEREF(__pyx_v_close_codes);
  __Pyx_GOTREF(__pyx_v_self->_close_codes);
  __Pyx_DECREF(__pyx_v_self->_close_codes);
  __pyx_v_self->_close_codes = __pyx_v_close_codes;

  /* "extensions/lib/websocket.pyx":77
 *     cdef object _close_codes
 * 
 *     def __cinit__(self, int version, int kind, object ProtocolError,             # <<<<<<<<<<<<<<
 *                   extensions=None, protocols=None, close_codes=None):
 *         self._version = version
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("lib.FrameParser.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/websocket.pyx":99
 * 
 *     @property
 *     def version(self):             # <<<<<<<<<<<<<<
 *         return self._version
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_11FrameParser_7version_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_3lib_11FrameParser_7version_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_3lib_11FrameParser_7version___get__(((struct __pyx_obj_3lib_FrameParser *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_11FrameParser_7version___get__(struct __pyx_obj_3lib_FrameParser *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "extensions/lib/websocket.pyx":100
 *     @property
 *     def version(self):
 *         return self._version             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->_version); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "extensions/lib/websocket.pyx":99
 * 
 *     @property
 *     def version(self):             # <<<<<<<<<<<<<<
 *         return self._version
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("lib.FrameParser.version.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/websocket.pyx":103
 * 
 *     @property
 *     def max_payload(self):             # <<<<<<<<<<<<<<
 *         return 1 << 63
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_11FrameParser_11max_payload_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_3lib_11FrameParser_11max_payload_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_3lib_11FrameParser_11max_payload___get__(((struct __pyx_obj_3lib_FrameParser *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_11FrameParser_11max_payload___get__(CYTHON_UNUSED struct __pyx_obj_3lib_FrameParser *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "extensions/lib/websocket.pyx":104
 *     @property
 *     def max_payload(self):
 *         return 1 << 63             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_int_9223372036854775808);
  __pyx_r = __pyx_int_9223372036854775808;
  goto __pyx_L0;

  /* "extensions/lib/websocket.pyx":103
 * 
 *     @property
 *     def max_payload(self):             # <<<<<<<<<<<<<<
 *         return 1 << 63
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/websocket.pyx":107
 * 
 *     @property
 *     def decode_mask_length(self):             # <<<<<<<<<<<<<<
 *         return self._decode_mask_length
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_11FrameParser_18decode_mask_length_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_3lib_11FrameParser_18decode_mask_length_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_3lib_11FrameParser_18decode_mask_length___get__(((struct __pyx_obj_3lib_FrameParser *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_11FrameParser_18decode_mask_length___get__(struct __pyx_obj_3lib_FrameParser *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "extensions/lib/websocket.pyx":108
 *     @property
 *     def decode_mask_length(self):
 *         return self._decode_mask_length             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->_decode_mask_length); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "extensions/lib/websocket.pyx":107
 * 
 *     @property
 *     def decode_mask_length(self):             # <<<<<<<<<<<<<<
 *         return self._decode_mask_length
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("lib.FrameParser.decode_mask_length.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/websocket.pyx":111
 * 
 *     @property
 *     def encode_mask_length(self):             # <<<<<<<<<<<<<<
 *         return self._encode_mask_length
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_11FrameParser_18encode_mask_length_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_3lib_11FrameParser_18encode_mask_length_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_3lib_11FrameParser_18encode_mask_length___get__(((struct __pyx_obj_3lib_FrameParser *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_11FrameParser_18encode_mask_length___get__(struct __pyx_obj_3lib_FrameParser *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "extensions/lib/websocket.pyx":112
 *     @property
 *     def encode_mask_length(self):
 *         return self._encode_mask_length             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->_encode_mask_length); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "extensions/lib/websocket.pyx":111
 * 
 *     @property
 *     def encode_mask_length(self):             # <<<<<<<<<<<<<<
 *         return self._encode_mask_length
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("lib.FrameParser.encode_mask_length.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/websocket.pyx":115
 * 
 *     @property
 *     def extensions(self):             # <<<<<<<<<<<<<<
 *         return self._extensions
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_11FrameParser_10extensions_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_3lib_11FrameParser_10extensions_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_3lib_11FrameParser_10extensions___get__(((struct __pyx_obj_3lib_FrameParser *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_11FrameParser_10extensions___get__(struct __pyx_obj_3lib_FrameParser *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "extensions/lib/websocket.pyx":116
 *     @property
 *     def extensions(self):
 *         return self._extensions             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_extensions);
  __pyx_r = __pyx_v_self->_extensions;
  goto __pyx_L0;

  /* "extensions/lib/websocket.pyx":115
 * 
 *     @property
 *     def extensions(self):             # <<<<<<<<<<<<<<
 *         return self._extensions
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/websocket.pyx":119
 * 
 *     @property
 *     def protocols(self):             # <<<<<<<<<<<<<<
 *         return self._protocols
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_11FrameParser_9protocols_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_3lib_11FrameParser_9protocols_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_3lib_11FrameParser_9protocols___get__(((struct __pyx_obj_3lib_FrameParser *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_11FrameParser_9protocols___get__(struct __pyx_obj_3lib_FrameParser *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "extensions/lib/websocket.pyx":120
 *     @property
 *     def protocols(self):
 *         return self._protocols             # <<<<<<<<<<<<<<
 * 
 *     def ping(self, body=None):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_protocols);
  __pyx_r = __pyx_v_self->_protocols;
  goto __pyx_L0;

  /* "extensions/lib/websocket.pyx":119
 * 
 *     @property
 *     def protocols(self):             # <<<<<<<<<<<<<<
 *         return self._protocols
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/websocket.pyx":122
 *         return self._protocols
 * 
 *     def ping(self, body=None):             # <<<<<<<<<<<<<<
 *         '''return a `ping` :class:`Frame`.'''
 *         return self.encode(body, opcode=0x9)
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_11FrameParser_3ping(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_3lib_11FrameParser_2ping[] = "return a `ping` :class:`Frame`.";
static PyObject *__pyx_pw_3lib_11FrameParser_3ping(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_body = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("ping (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_body,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_body);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "ping") < 0)) __PYX_ERR(1, 122, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_body = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("ping", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 122, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("lib.FrameParser.ping", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_3lib_11FrameParser_2ping(((struct __pyx_obj_3lib_FrameParser *)__pyx_v_self), __pyx_v_body);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_11FrameParser_2ping(struct __pyx_obj_3lib_FrameParser *__pyx_v_self, PyObject *__pyx_v_body) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("ping", 0);

  /* "extensions/lib/websocket.pyx":124
 *     def ping(self, body=None):
 *         '''return a `ping` :class:`Frame`.'''
 *         return self.encode(body, opcode=0x9)             # <<<<<<<<<<<<<<
 * 
 *     def pong(self, body=None):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_encode); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_body);
  __Pyx_GIVEREF(__pyx_v_body);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_body);
  __pyx_t_3 = PyDict_New(); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_opcode, __pyx_int_9) < 0) __PYX_ERR(1, 124, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "extensions/lib/websocket.pyx":122
 *         return self._protocols
 * 
 *     def ping(self, body=None):             # <<<<<<<<<<<<<<
 *         '''return a `ping` :class:`Frame`.'''
 *         return self.encode(body, opcode=0x9)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("lib.FrameParser.ping", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/websocket.pyx":126
 *         return self.encode(body, opcode=0x9)
 * 
 *     def pong(self, body=None):             # <<<<<<<<<<<<<<
 *         '''return a `pong` :class:`Frame`.'''
 *         return self.encode(body, opcode=0xA)
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_11FrameParser_5pong(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_3lib_11FrameParser_4pong[] = "return a `pong` :class:`Frame`.";
static PyObject *__pyx_pw_3lib_11FrameParser_5pong(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_body = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("pong (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_body,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_body);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "pong") < 0)) __PYX_ERR(1, 126, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_body = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("pong", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 126, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("lib.FrameParser.pong", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_3lib_11FrameParser_4pong(((struct __pyx_obj_3lib_FrameParser *)__pyx_v_self), __pyx_v_body);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_11FrameParser_4pong(struct __pyx_obj_3lib_FrameParser *__pyx_v_self, PyObject *__pyx_v_body) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("pong", 0);

  /* "extensions/lib/websocket.pyx":128
 *     def pong(self, body=None):
 *         '''return a `pong` :class:`Frame`.'''
 *         return self.encode(body, opcode=0xA)             # <<<<<<<<<<<<<<
 * 
 *     def close(self, code=None):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_encode); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_body);
  __Pyx_GIVEREF(__pyx_v_body);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_body);
  __pyx_t_3 = PyDict_New(); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_opcode, __pyx_int_10) < 0) __PYX_ERR(1, 128, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "extensions/lib/websocket.pyx":126
 *         return self.encode(body, opcode=0x9)
 * 
 *     def pong(self, body=None):             # <<<<<<<<<<<<<<
 *         '''return a `pong` :class:`Frame`.'''
 *         return self.encode(body, opcode=0xA)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("lib.FrameParser.pong", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/websocket.pyx":130
 *         return self.encode(body, opcode=0xA)
 * 
 *     def close(self, code=None):             # <<<<<<<<<<<<<<
 *         '''return a `close` :class:`Frame`.'''
 *         code = code or 1000
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_11FrameParser_7close(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_3lib_11FrameParser_6close[] = "return a `close` :class:`Frame`.";
static PyObject *__pyx_pw_3lib_11FrameParser_7close(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_code = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("close (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_code,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_code);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "close") < 0)) __PYX_ERR(1, 130, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_code = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("close", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 130, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("lib.FrameParser.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_3lib_11FrameParser_6close(((struct __pyx_obj_3lib_FrameParser *)__pyx_v_self), __pyx_v_code);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_11FrameParser_6close(struct __pyx_obj_3lib_FrameParser *__pyx_v_self, PyObject *__pyx_v_code) {
  PyObject *__pyx_v_body = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("close", 0);
  __Pyx_INCREF(__pyx_v_code);

  /* "extensions/lib/websocket.pyx":132
 *     def close(self, code=None):
 *         '''return a `close` :class:`Frame`.'''
 *         code = code or 1000             # <<<<<<<<<<<<<<
 *         body = pack('!H', code)
 *         if self._close_codes is not None:
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_code); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 132, __pyx_L1_error)
  if (!__pyx_t_2) {
  } else {
    __Pyx_INCREF(__pyx_v_code);
    __pyx_t_1 = __pyx_v_code;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_PyInt_From_long(0x3E8); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = __pyx_t_3;
  __pyx_t_3 = 0;
  __pyx_L3_bool_binop_done:;
  __Pyx_DECREF_SET(__pyx_v_code, __pyx_t_1);
  __pyx_t_1 = 0;

  /* "extensions/lib/websocket.pyx":133
 *         '''return a `close` :class:`Frame`.'''
 *         code = code or 1000
 *         body = pack('!H', code)             # <<<<<<<<<<<<<<
 *         if self._close_codes is not None:
 *             body += self._close_codes.get(code, '').encode('utf-8')
 */
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_pack_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  __pyx_t_5 = 0;
  if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_5 = 1;
    }
  }
  __pyx_t_6 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  if (__pyx_t_4) {
    __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4); __pyx_t_4 = NULL;
  }
  __Pyx_INCREF(__pyx_kp_s_H);
  __Pyx_GIVEREF(__pyx_kp_s_H);
  PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_5, __pyx_kp_s_H);
  __Pyx_INCREF(__pyx_v_code);
  __Pyx_GIVEREF(__pyx_v_code);
  PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_5, __pyx_v_code);
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_body = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "extensions/lib/websocket.pyx":134
 *         code = code or 1000
 *         body = pack('!H', code)
 *         if self._close_codes is not None:             # <<<<<<<<<<<<<<
 *             body += self._close_codes.get(code, '').encode('utf-8')
 *         return self.encode(body, opcode=0x8)
 */
  __pyx_t_2 = (__pyx_v_self->_close_codes != Py_None);
  __pyx_t_7 = (__pyx_t_2 != 0);
  if (__pyx_t_7) {

    /* "extensions/lib/websocket.pyx":135
 *         body = pack('!H', code)
 *         if self._close_codes is not None:
 *             body += self._close_codes.get(code, '').encode('utf-8')             # <<<<<<<<<<<<<<
 *         return self.encode(body, opcode=0x8)
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_close_codes, __pyx_n_s_get); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = NULL;
    __pyx_t_5 = 0;
    if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_5 = 1;
      }
    }
    __pyx_t_4 = PyTuple_New(2+__pyx_t_5); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__pyx_t_6) {
      __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_6); __pyx_t_6 = NULL;
    }
    __Pyx_INCREF(__pyx_v_code);
    __Pyx_GIVEREF(__pyx_v_code);
    PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_5, __pyx_v_code);
    __Pyx_INCREF(__pyx_kp_s_);
    __Pyx_GIVEREF(__pyx_kp_s_);
    PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_5, __pyx_kp_s_);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_encode); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_body, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_body, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "extensions/lib/websocket.pyx":134
 *         code = code or 1000
 *         body = pack('!H', code)
 *         if self._close_codes is not None:             # <<<<<<<<<<<<<<
 *             body += self._close_codes.get(code, '').encode('utf-8')
 *         return self.encode(body, opcode=0x8)
 */
  }

  /* "extensions/lib/websocket.pyx":136
 *         if self._close_codes is not None:
 *             body += self._close_codes.get(code, '').encode('utf-8')
 *         return self.encode(body, opcode=0x8)             # <<<<<<<<<<<<<<
 * 
 *     def continuation(self, body=None, final=True):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_encode); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_body);
  __Pyx_GIVEREF(__pyx_v_body);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_body);
  __pyx_t_4 = PyDict_New(); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_n_s_opcode, __pyx_int_8) < 0) __PYX_ERR(1, 136, __pyx_L1_error)
  __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_1, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "extensions/lib/websocket.pyx":130
 *         return self.encode(body, opcode=0xA)
 * 
 *     def close(self, code=None):             # <<<<<<<<<<<<<<
 *         '''return a `close` :class:`Frame`.'''
 *         code = code or 1000
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("lib.FrameParser.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_body);
  __Pyx_XDECREF(__pyx_v_code);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/websocket.pyx":138
 *         return self.encode(body, opcode=0x8)
 * 
 *     def continuation(self, body=None, final=True):             # <<<<<<<<<<<<<<
 *         '''return a `continuation` :class:`Frame`.'''
 *         return self.encode(body, opcode=0, final=final)
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_11FrameParser_9continuation(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_3lib_11FrameParser_8continuation[] = "return a `continuation` :class:`Frame`.";
static PyObject *__pyx_pw_3lib_11FrameParser_9continuation(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_body = 0;
  PyObject *__pyx_v_final = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("continuation (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_body,&__pyx_n_s_final,0};
    PyObject* values[2] = {0,0};
    values[0] = ((PyObject *)Py_None);
    values[1] = ((PyObject *)Py_True);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_body);
          if (value) { values[0] = value; kw_args--; }
        }
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_final);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "continuation") < 0)) __PYX_ERR(1, 138, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_body = values[0];
    __pyx_v_final = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("continuation", 0, 0, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 138, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("lib.FrameParser.continuation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_3lib_11FrameParser_8continuation(((struct __pyx_obj_3lib_FrameParser *)__pyx_v_self), __pyx_v_body, __pyx_v_final);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_11FrameParser_8continuation(struct __pyx_obj_3lib_FrameParser *__pyx_v_self, PyObject *__pyx_v_body, PyObject *__pyx_v_final) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("continuation", 0);

  /* "extensions/lib/websocket.pyx":140
 *     def continuation(self, body=None, final=True):
 *         '''return a `continuation` :class:`Frame`.'''
 *         return self.encode(body, opcode=0, final=final)             # <<<<<<<<<<<<<<
 * 
 *     def encode(self, message, final=True, bytes masking_key=None,
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_encode); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_v_body);
  __Pyx_GIVEREF(__pyx_v_body);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_body);
  __pyx_t_3 = PyDict_New(); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_opcode, __pyx_int_0) < 0) __PYX_ERR(1, 140, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_final, __pyx_v_final) < 0) __PYX_ERR(1, 140, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_4;
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "extensions/lib/websocket.pyx":138
 *         return self.encode(body, opcode=0x8)
 * 
 *     def continuation(self, body=None, final=True):             # <<<<<<<<<<<<<<
 *         '''return a `continuation` :class:`Frame`.'''
 *         return self.encode(body, opcode=0, final=final)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("lib.FrameParser.continuation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/websocket.pyx":142
 *         return self.encode(body, opcode=0, final=final)
 * 
 *     def encode(self, message, final=True, bytes masking_key=None,             # <<<<<<<<<<<<<<
 *                opcode=None, int rsv1=0, int rsv2=0, int rsv3=0):
 *         '''Encode a ``message`` for writing into the wire.
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_11FrameParser_11encode(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_3lib_11FrameParser_10encode[] = "Encode a ``message`` for writing into the wire.\n\n        The message length cannot exceed :attr:`max_payload`\n        To produce several frames for a given large message use\n        :meth:`multi_encode'.\n        ";
static PyObject *__pyx_pw_3lib_11FrameParser_11encode(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_message = 0;
  PyObject *__pyx_v_final = 0;
  PyObject *__pyx_v_masking_key = 0;
  PyObject *__pyx_v_opcode = 0;
  int __pyx_v_rsv1;
  int __pyx_v_rsv2;
  int __pyx_v_rsv3;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("encode (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_message,&__pyx_n_s_final,&__pyx_n_s_masking_key,&__pyx_n_s_opcode,&__pyx_n_s_rsv1,&__pyx_n_s_rsv2,&__pyx_n_s_rsv3,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    values[1] = ((PyObject *)Py_True);
    values[2] = ((PyObject*)Py_None);

    /* "extensions/lib/websocket.pyx":143
 * 
 *     def encode(self, message, final=True, bytes masking_key=None,
 *                opcode=None, int rsv1=0, int rsv2=0, int rsv3=0):             # <<<<<<<<<<<<<<
 *         '''Encode a ``message`` for writing into the wire.
 * 
 */
    values[3] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_message)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_final);
          if (value) { values[1] = value; kw_args--; }
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_masking_key);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_opcode);
          if (value) { values[3] = value; kw_args--; }
        }
        case  4:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_rsv1);
          if (value) { values[4] = value; kw_args--; }
        }
        case  5:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_rsv2);
          if (value) { values[5] = value; kw_args--; }
        }
        case  6:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_rsv3);
          if (value) { values[6] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "encode") < 0)) __PYX_ERR(1, 142, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_message = values[0];
    __pyx_v_final = values[1];
    __pyx_v_masking_key = ((PyObject*)values[2]);
    __pyx_v_opcode = values[3];
    if (values[4]) {
      __pyx_v_rsv1 = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_rsv1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 143, __pyx_L3_error)
    } else {
      __pyx_v_rsv1 = ((int)0);
    }
    if (values[5]) {
      __pyx_v_rsv2 = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_rsv2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 143, __pyx_L3_error)
    } else {
      __pyx_v_rsv2 = ((int)0);
    }
    if (values[6]) {
      __pyx_v_rsv3 = __Pyx_PyInt_As_int(values[6]); if (unlikely((__pyx_v_rsv3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 143, __pyx_L3_error)
    } else {
      __pyx_v_rsv3 = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("encode", 0, 1, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 142, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("lib.FrameParser.encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_masking_key), (&PyBytes_Type), 1, "masking_key", 1))) __PYX_ERR(1, 142, __pyx_L1_error)
  __pyx_r = __pyx_pf_3lib_11FrameParser_10encode(((struct __pyx_obj_3lib_FrameParser *)__pyx_v_self), __pyx_v_message, __pyx_v_final, __pyx_v_masking_key, __pyx_v_opcode, __pyx_v_rsv1, __pyx_v_rsv2, __pyx_v_rsv3);

  /* "extensions/lib/websocket.pyx":142
 *         return self.encode(body, opcode=0, final=final)
 * 
 *     def encode(self, message, final=True, bytes masking_key=None,             # <<<<<<<<<<<<<<
 *                opcode=None, int rsv1=0, int rsv2=0, int rsv3=0):
 *         '''Encode a ``message`` for writing into the wire.
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_11FrameParser_10encode(struct __pyx_obj_3lib_FrameParser *__pyx_v_self, PyObject *__pyx_v_message, PyObject *__pyx_v_final, PyObject *__pyx_v_masking_key, PyObject *__pyx_v_opcode, int __pyx_v_rsv1, int __pyx_v_rsv2, int __pyx_v_rsv3) {
  PyObject *__pyx_v_data = 0;
  int __pyx_v_fin;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("encode", 0);
  __Pyx_INCREF(__pyx_v_masking_key);
  __Pyx_INCREF(__pyx_v_opcode);

  /* "extensions/lib/websocket.pyx":151
 *         '''
 *         cdef bytes data
 *         cdef int fin = 1 if final else 0             # <<<<<<<<<<<<<<
 *         opcode, masking_key, data = self._info(message, opcode, masking_key)
 *         return self._encode(data, opcode, masking_key, fin,
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_final); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 151, __pyx_L1_error)
  if (__pyx_t_2) {
    __pyx_t_1 = 1;
  } else {
    __pyx_t_1 = 0;
  }
  __pyx_v_fin = __pyx_t_1;

  /* "extensions/lib/websocket.pyx":152
 *         cdef bytes data
 *         cdef int fin = 1 if final else 0
 *         opcode, masking_key, data = self._info(message, opcode, masking_key)             # <<<<<<<<<<<<<<
 *         return self._encode(data, opcode, masking_key, fin,
 *                             rsv1, rsv2, rsv3)
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_3lib_FrameParser *)__pyx_v_self->__pyx_vtab)->_info(__pyx_v_self, __pyx_v_message, __pyx_v_opcode, __pyx_v_masking_key); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (likely(__pyx_t_3 != Py_None)) {
    PyObject* sequence = __pyx_t_3;
    #if CYTHON_COMPILING_IN_CPYTHON
    Py_ssize_t size = Py_SIZE(sequence);
    #else
    Py_ssize_t size = PySequence_Size(sequence);
    #endif
    if (unlikely(size != 3)) {
      if (size > 3) __Pyx_RaiseTooManyValuesError(3);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(1, 152, __pyx_L1_error)
    }
    #if CYTHON_COMPILING_IN_CPYTHON
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
    __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
    __pyx_t_6 = PyTuple_GET_ITEM(sequence, 2); 
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_t_6);
    #else
    __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    #endif
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(1, 152, __pyx_L1_error)
  }
  if (!(likely(PyBytes_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_5)->tp_name), 0))) __PYX_ERR(1, 152, __pyx_L1_error)
  if (!(likely(PyBytes_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_6)->tp_name), 0))) __PYX_ERR(1, 152, __pyx_L1_error)
  __Pyx_DECREF_SET(__pyx_v_opcode, __pyx_t_4);
  __pyx_t_4 = 0;
  __Pyx_DECREF_SET(__pyx_v_masking_key, ((PyObject*)__pyx_t_5));
  __pyx_t_5 = 0;
  __pyx_v_data = ((PyObject*)__pyx_t_6);
  __pyx_t_6 = 0;

  /* "extensions/lib/websocket.pyx":153
 *         cdef int fin = 1 if final else 0
 *         opcode, masking_key, data = self._info(message, opcode, masking_key)
 *         return self._encode(data, opcode, masking_key, fin,             # <<<<<<<<<<<<<<
 *                             rsv1, rsv2, rsv3)
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_As_int(__pyx_v_opcode); if (unlikely((__pyx_t_1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 153, __pyx_L1_error)

  /* "extensions/lib/websocket.pyx":154
 *         opcode, masking_key, data = self._info(message, opcode, masking_key)
 *         return self._encode(data, opcode, masking_key, fin,
 *                             rsv1, rsv2, rsv3)             # <<<<<<<<<<<<<<
 * 
 *     def multi_encode(self, message, bytes masking_key=None, opcode=None,
 */
  __pyx_t_3 = ((struct __pyx_vtabstruct_3lib_FrameParser *)__pyx_v_self->__pyx_vtab)->_encode(__pyx_v_self, __pyx_v_data, __pyx_t_1, __pyx_v_masking_key, __pyx_v_fin, __pyx_v_rsv1, __pyx_v_rsv2, __pyx_v_rsv3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "extensions/lib/websocket.pyx":142
 *         return self.encode(body, opcode=0, final=final)
 * 
 *     def encode(self, message, final=True, bytes masking_key=None,             # <<<<<<<<<<<<<<
 *                opcode=None, int rsv1=0, int rsv2=0, int rsv3=0):
 *         '''Encode a ``message`` for writing into the wire.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("lib.FrameParser.encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_masking_key);
  __Pyx_XDECREF(__pyx_v_opcode);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_3lib_11FrameParser_14generator3(__pyx_CoroutineObject *__pyx_generator, PyObject *__pyx_sent_value); /* proto */

/* "extensions/lib/websocket.pyx":156
 *                             rsv1, rsv2, rsv3)
 * 
 *     def multi_encode(self, message, bytes masking_key=None, opcode=None,             # <<<<<<<<<<<<<<
 *                      int rsv1=0, int rsv2=0, int rsv3=0,
 *                      cython.ulonglong max_payload=0):
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_11FrameParser_13multi_encode(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_3lib_11FrameParser_12multi_encode[] = "Encode a ``message`` into several frames depending on size.\n\n        Returns a generator of bytes to be sent over the wire.\n        ";
static PyObject *__pyx_pw_3lib_11FrameParser_13multi_encode(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_message = 0;
  PyObject *__pyx_v_masking_key = 0;
  PyObject *__pyx_v_opcode = 0;
  int __pyx_v_rsv1;
  int __pyx_v_rsv2;
  int __pyx_v_rsv3;
  unsigned PY_LONG_LONG __pyx_v_max_payload;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("multi_encode (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_message,&__pyx_n_s_masking_key,&__pyx_n_s_opcode,&__pyx_n_s_rsv1,&__pyx_n_s_rsv2,&__pyx_n_s_rsv3,&__pyx_n_s_max_payload,0};
    PyObject* values[7] = {0,0,0,0,0,0,0};
    values[1] = ((PyObject*)Py_None);
    values[2] = ((PyObject *)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_message)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_masking_key);
          if (value) { values[1] = value; kw_args--; }
        }
        case  2:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_opcode);
          if (value) { values[2] = value; kw_args--; }
        }
        case  3:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_rsv1);
          if (value) { values[3] = value; kw_args--; }
        }
        case  4:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_rsv2);
          if (value) { values[4] = value; kw_args--; }
        }
        case  5:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_rsv3);
          if (value) { values[5] = value; kw_args--; }
        }
        case  6:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_max_payload);
          if (value) { values[6] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "multi_encode") < 0)) __PYX_ERR(1, 156, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  7: values[6] = PyTuple_GET_ITEM(__pyx_args, 6);
        case  6: values[5] = PyTuple_GET_ITEM(__pyx_args, 5);
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_message = values[0];
    __pyx_v_masking_key = ((PyObject*)values[1]);
    __pyx_v_opcode = values[2];
    if (values[3]) {
      __pyx_v_rsv1 = __Pyx_PyInt_As_int(values[3]); if (unlikely((__pyx_v_rsv1 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 157, __pyx_L3_error)
    } else {
      __pyx_v_rsv1 = ((int)0);
    }
    if (values[4]) {
      __pyx_v_rsv2 = __Pyx_PyInt_As_int(values[4]); if (unlikely((__pyx_v_rsv2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 157, __pyx_L3_error)
    } else {
      __pyx_v_rsv2 = ((int)0);
    }
    if (values[5]) {
      __pyx_v_rsv3 = __Pyx_PyInt_As_int(values[5]); if (unlikely((__pyx_v_rsv3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 157, __pyx_L3_error)
    } else {
      __pyx_v_rsv3 = ((int)0);
    }
    if (values[6]) {
      __pyx_v_max_payload = __Pyx_PyInt_As_unsigned_PY_LONG_LONG(values[6]); if (unlikely((__pyx_v_max_payload == (unsigned PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(1, 158, __pyx_L3_error)
    } else {
      __pyx_v_max_payload = ((unsigned PY_LONG_LONG)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("multi_encode", 0, 1, 7, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 156, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("lib.FrameParser.multi_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_masking_key), (&PyBytes_Type), 1, "masking_key", 1))) __PYX_ERR(1, 156, __pyx_L1_error)
  __pyx_r = __pyx_pf_3lib_11FrameParser_12multi_encode(((struct __pyx_obj_3lib_FrameParser *)__pyx_v_self), __pyx_v_message, __pyx_v_masking_key, __pyx_v_opcode, __pyx_v_rsv1, __pyx_v_rsv2, __pyx_v_rsv3, __pyx_v_max_payload);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_11FrameParser_12multi_encode(struct __pyx_obj_3lib_FrameParser *__pyx_v_self, PyObject *__pyx_v_message, PyObject *__pyx_v_masking_key, PyObject *__pyx_v_opcode, int __pyx_v_rsv1, int __pyx_v_rsv2, int __pyx_v_rsv3, unsigned PY_LONG_LONG __pyx_v_max_payload) {
  struct __pyx_obj_3lib___pyx_scope_struct_5_multi_encode *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("multi_encode", 0);
  __pyx_cur_scope = (struct __pyx_obj_3lib___pyx_scope_struct_5_multi_encode *)__pyx_tp_new_3lib___pyx_scope_struct_5_multi_encode(__pyx_ptype_3lib___pyx_scope_struct_5_multi_encode, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __Pyx_GOTREF(__pyx_cur_scope);
  __pyx_cur_scope->__pyx_v_self = __pyx_v_self;
  __Pyx_INCREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __Pyx_GIVEREF((PyObject *)__pyx_cur_scope->__pyx_v_self);
  __pyx_cur_scope->__pyx_v_message = __pyx_v_message;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_message);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_message);
  __pyx_cur_scope->__pyx_v_masking_key = __pyx_v_masking_key;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_masking_key);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_masking_key);
  __pyx_cur_scope->__pyx_v_opcode = __pyx_v_opcode;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_opcode);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_opcode);
  __pyx_cur_scope->__pyx_v_rsv1 = __pyx_v_rsv1;
  __pyx_cur_scope->__pyx_v_rsv2 = __pyx_v_rsv2;
  __pyx_cur_scope->__pyx_v_rsv3 = __pyx_v_rsv3;
  __pyx_cur_scope->__pyx_v_max_payload = __pyx_v_max_payload;
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_3lib_11FrameParser_14generator3, (PyObject *) __pyx_cur_scope, __pyx_n_s_multi_encode, __pyx_n_s_FrameParser_multi_encode); if (unlikely(!gen)) __PYX_ERR(1, 156, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("lib.FrameParser.multi_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_3lib_11FrameParser_14generator3(__pyx_CoroutineObject *__pyx_generator, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_3lib___pyx_scope_struct_5_multi_encode *__pyx_cur_scope = ((struct __pyx_obj_3lib___pyx_scope_struct_5_multi_encode *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  long __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  unsigned PY_LONG_LONG __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("None", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L9_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(1, 156, __pyx_L1_error)

  /* "extensions/lib/websocket.pyx":166
 *         cdef bytes chunk
 *         cdef int fin
 *         max_payload = max(2, max_payload or 1 << 63)             # <<<<<<<<<<<<<<
 *         opcode, masking_key, data = self._info(message, opcode, masking_key)
 *         #
 */
  if (!__pyx_cur_scope->__pyx_v_max_payload) {
  } else {
    __pyx_t_2 = __Pyx_PyInt_From_unsigned_PY_LONG_LONG(__pyx_cur_scope->__pyx_v_max_payload); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 166, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L4_bool_binop_done;
  }
  __Pyx_INCREF(__pyx_int_9223372036854775808);
  __pyx_t_1 = __pyx_int_9223372036854775808;
  __pyx_L4_bool_binop_done:;
  __pyx_t_3 = 2;
  __pyx_t_4 = __Pyx_PyInt_From_long(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = PyObject_RichCompare(__pyx_t_1, __pyx_t_4, Py_GT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 166, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 166, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_6) {
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_2 = __pyx_t_1;
  } else {
    __pyx_t_5 = __Pyx_PyInt_From_long(__pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 166, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_2 = __pyx_t_5;
    __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = __Pyx_PyInt_As_unsigned_PY_LONG_LONG(__pyx_t_2); if (unlikely((__pyx_t_7 == (unsigned PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(1, 166, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_cur_scope->__pyx_v_max_payload = __pyx_t_7;

  /* "extensions/lib/websocket.pyx":167
 *         cdef int fin
 *         max_payload = max(2, max_payload or 1 << 63)
 *         opcode, masking_key, data = self._info(message, opcode, masking_key)             # <<<<<<<<<<<<<<
 *         #
 *         while data:
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_3lib_FrameParser *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_info(__pyx_cur_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_message, __pyx_cur_scope->__pyx_v_opcode, __pyx_cur_scope->__pyx_v_masking_key); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(__pyx_t_2 != Py_None)) {
    PyObject* sequence = __pyx_t_2;
    #if CYTHON_COMPILING_IN_CPYTHON
    Py_ssize_t size = Py_SIZE(sequence);
    #else
    Py_ssize_t size = PySequence_Size(sequence);
    #endif
    if (unlikely(size != 3)) {
      if (size > 3) __Pyx_RaiseTooManyValuesError(3);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(1, 167, __pyx_L1_error)
    }
    #if CYTHON_COMPILING_IN_CPYTHON
    __pyx_t_1 = PyTuple_GET_ITEM(sequence, 0); 
    __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 2); 
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_t_4);
    #else
    __pyx_t_1 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(1, 167, __pyx_L1_error)
  }
  if (!(likely(PyBytes_CheckExact(__pyx_t_5))||((__pyx_t_5) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_5)->tp_name), 0))) __PYX_ERR(1, 167, __pyx_L1_error)
  if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(1, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_opcode);
  __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_opcode, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;
  __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_masking_key);
  __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_masking_key, ((PyObject*)__pyx_t_5));
  __Pyx_GIVEREF(__pyx_t_5);
  __pyx_t_5 = 0;
  __Pyx_GIVEREF(__pyx_t_4);
  __pyx_cur_scope->__pyx_v_data = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "extensions/lib/websocket.pyx":169
 *         opcode, masking_key, data = self._info(message, opcode, masking_key)
 *         #
 *         while data:             # <<<<<<<<<<<<<<
 *             if len(data) >= max_payload:
 *                 chunk, data, fin = (data[:max_payload],
 */
  while (1) {
    __pyx_t_6 = (__pyx_cur_scope->__pyx_v_data != Py_None) && (PyBytes_GET_SIZE(__pyx_cur_scope->__pyx_v_data) != 0);
    if (!__pyx_t_6) break;

    /* "extensions/lib/websocket.pyx":170
 *         #
 *         while data:
 *             if len(data) >= max_payload:             # <<<<<<<<<<<<<<
 *                 chunk, data, fin = (data[:max_payload],
 *                                     data[max_payload:], 0)
 */
    if (unlikely(__pyx_cur_scope->__pyx_v_data == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(1, 170, __pyx_L1_error)
    }
    __pyx_t_8 = PyBytes_GET_SIZE(__pyx_cur_scope->__pyx_v_data); if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(1, 170, __pyx_L1_error)
    __pyx_t_6 = ((__pyx_t_8 >= __pyx_cur_scope->__pyx_v_max_payload) != 0);
    if (__pyx_t_6) {

      /* "extensions/lib/websocket.pyx":171
 *         while data:
 *             if len(data) >= max_payload:
 *                 chunk, data, fin = (data[:max_payload],             # <<<<<<<<<<<<<<
 *                                     data[max_payload:], 0)
 *             else:
 */
      if (unlikely(__pyx_cur_scope->__pyx_v_data == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(1, 171, __pyx_L1_error)
      }
      __pyx_t_2 = PySequence_GetSlice(__pyx_cur_scope->__pyx_v_data, 0, __pyx_cur_scope->__pyx_v_max_payload); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 171, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);

      /* "extensions/lib/websocket.pyx":172
 *             if len(data) >= max_payload:
 *                 chunk, data, fin = (data[:max_payload],
 *                                     data[max_payload:], 0)             # <<<<<<<<<<<<<<
 *             else:
 *                 chunk, data, fin = data, b'', 1
 */
      if (unlikely(__pyx_cur_scope->__pyx_v_data == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(1, 172, __pyx_L1_error)
      }
      __pyx_t_4 = PySequence_GetSlice(__pyx_cur_scope->__pyx_v_data, __pyx_cur_scope->__pyx_v_max_payload, PY_SSIZE_T_MAX); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 172, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_9 = 0;
      __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_chunk);
      __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_chunk, ((PyObject*)__pyx_t_2));
      __Pyx_GIVEREF(__pyx_t_2);
      __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_data);
      __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_data, ((PyObject*)__pyx_t_4));
      __Pyx_GIVEREF(__pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_cur_scope->__pyx_v_fin = __pyx_t_9;

      /* "extensions/lib/websocket.pyx":170
 *         #
 *         while data:
 *             if len(data) >= max_payload:             # <<<<<<<<<<<<<<
 *                 chunk, data, fin = (data[:max_payload],
 *                                     data[max_payload:], 0)
 */
      goto __pyx_L8;
    }

    /* "extensions/lib/websocket.pyx":174
 *                                     data[max_payload:], 0)
 *             else:
 *                 chunk, data, fin = data, b'', 1             # <<<<<<<<<<<<<<
 *             yield self._encode(chunk, opcode, masking_key, fin,
 *                                rsv1, rsv2, rsv3)
 */
    /*else*/ {
      __pyx_t_4 = __pyx_cur_scope->__pyx_v_data;
      __Pyx_INCREF(__pyx_t_4);
      __pyx_t_2 = __pyx_kp_b_;
      __Pyx_INCREF(__pyx_t_2);
      __pyx_t_9 = 1;
      __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_chunk);
      __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_chunk, ((PyObject*)__pyx_t_4));
      __Pyx_GIVEREF(__pyx_t_4);
      __pyx_t_4 = 0;
      __Pyx_GOTREF(__pyx_cur_scope->__pyx_v_data);
      __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_data, ((PyObject*)__pyx_t_2));
      __Pyx_GIVEREF(__pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_cur_scope->__pyx_v_fin = __pyx_t_9;
    }
    __pyx_L8:;

    /* "extensions/lib/websocket.pyx":175
 *             else:
 *                 chunk, data, fin = data, b'', 1
 *             yield self._encode(chunk, opcode, masking_key, fin,             # <<<<<<<<<<<<<<
 *                                rsv1, rsv2, rsv3)
 * 
 */
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_cur_scope->__pyx_v_opcode); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 175, __pyx_L1_error)

    /* "extensions/lib/websocket.pyx":176
 *                 chunk, data, fin = data, b'', 1
 *             yield self._encode(chunk, opcode, masking_key, fin,
 *                                rsv1, rsv2, rsv3)             # <<<<<<<<<<<<<<
 * 
 *     def decode(self, bytes data=None):
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_3lib_FrameParser *)__pyx_cur_scope->__pyx_v_self->__pyx_vtab)->_encode(__pyx_cur_scope->__pyx_v_self, __pyx_cur_scope->__pyx_v_chunk, __pyx_t_9, __pyx_cur_scope->__pyx_v_masking_key, __pyx_cur_scope->__pyx_v_fin, __pyx_cur_scope->__pyx_v_rsv1, __pyx_cur_scope->__pyx_v_rsv2, __pyx_cur_scope->__pyx_v_rsv3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L9_resume_from_yield:;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(1, 175, __pyx_L1_error)
  }

  /* "extensions/lib/websocket.pyx":156
 *                             rsv1, rsv2, rsv3)
 * 
 *     def multi_encode(self, message, bytes masking_key=None, opcode=None,             # <<<<<<<<<<<<<<
 *                      int rsv1=0, int rsv2=0, int rsv3=0,
 *                      cython.ulonglong max_payload=0):
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("multi_encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/websocket.pyx":178
 *                                rsv1, rsv2, rsv3)
 * 
 *     def decode(self, bytes data=None):             # <<<<<<<<<<<<<<
 *         cdef int fin, rsv1, rsv2, rsv3, opcode, payload_length
 *         cdef Frame frame = self.frame
 */

/* Python wrapper */
static PyObject *__pyx_pw_3lib_11FrameParser_16decode(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyObject *__pyx_pw_3lib_11FrameParser_16decode(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_data = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("decode (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_data,0};
    PyObject* values[1] = {0};
    values[0] = ((PyObject*)Py_None);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = PyDict_GetItem(__pyx_kwds, __pyx_n_s_data);
          if (value) { values[0] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "decode") < 0)) __PYX_ERR(1, 178, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_data = ((PyObject*)values[0]);
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("decode", 0, 0, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 178, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("lib.FrameParser.decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_data), (&PyBytes_Type), 1, "data", 1))) __PYX_ERR(1, 178, __pyx_L1_error)
  __pyx_r = __pyx_pf_3lib_11FrameParser_15decode(((struct __pyx_obj_3lib_FrameParser *)__pyx_v_self), __pyx_v_data);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3lib_11FrameParser_15decode(struct __pyx_obj_3lib_FrameParser *__pyx_v_self, PyObject *__pyx_v_data) {
  int __pyx_v_fin;
  CYTHON_UNUSED int __pyx_v_rsv1;
  CYTHON_UNUSED int __pyx_v_rsv2;
  CYTHON_UNUSED int __pyx_v_rsv3;
  int __pyx_v_opcode;
  int __pyx_v_payload_length;
  struct __pyx_obj_3lib_Frame *__pyx_v_frame = 0;
  int __pyx_v_mask_length;
  PyObject *__pyx_v_chunk = 0;
  PyObject *__pyx_v_first_byte = NULL;
  PyObject *__pyx_v_second_byte = NULL;
  PyObject *__pyx_v_extension = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *(*__pyx_t_10)(PyObject *);
  PyObject *(*__pyx_t_11)(PyObject *);
  Py_ssize_t __pyx_t_12;
  char const *__pyx_t_13;
  char const *__pyx_t_14;
  __Pyx_RefNannySetupContext("decode", 0);

  /* "extensions/lib/websocket.pyx":180
 *     def decode(self, bytes data=None):
 *         cdef int fin, rsv1, rsv2, rsv3, opcode, payload_length
 *         cdef Frame frame = self.frame             # <<<<<<<<<<<<<<
 *         cdef int mask_length = self._decode_mask_length
 *         cdef bytes chunk
 */
  __pyx_t_1 = ((PyObject *)__pyx_v_self->frame);
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_frame = ((struct __pyx_obj_3lib_Frame *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "extensions/lib/websocket.pyx":181
 *         cdef int fin, rsv1, rsv2, rsv3, opcode, payload_length
 *         cdef Frame frame = self.frame
 *         cdef int mask_length = self._decode_mask_length             # <<<<<<<<<<<<<<
 *         cdef bytes chunk
 *         #
 */
  __pyx_t_2 = __pyx_v_self->_decode_mask_length;
  __pyx_v_mask_length = __pyx_t_2;

  /* "extensions/lib/websocket.pyx":184
 *         cdef bytes chunk
 *         #
 *         if data:             # <<<<<<<<<<<<<<
 *             self.buffer.extend(data)
 *         if frame is None:
 */
  __pyx_t_3 = (__pyx_v_data != Py_None) && (PyBytes_GET_SIZE(__pyx_v_data) != 0);
  if (__pyx_t_3) {

    /* "extensions/lib/websocket.pyx":185
 *         #
 *         if data:
 *             self.buffer.extend(data)             # <<<<<<<<<<<<<<
 *         if frame is None:
 *             if len(self.buffer) < 2:
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->buffer, __pyx_n_s_extend); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_data); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 185, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    } else {
      __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 185, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
      __Pyx_INCREF(__pyx_v_data);
      __Pyx_GIVEREF(__pyx_v_data);
      PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_data);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 185, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "extensions/lib/websocket.pyx":184
 *         cdef bytes chunk
 *         #
 *         if data:             # <<<<<<<<<<<<<<
 *             self.buffer.extend(data)
 *         if frame is None:
 */
  }

  /* "extensions/lib/websocket.pyx":186
 *         if data:
 *             self.buffer.extend(data)
 *         if frame is None:             # <<<<<<<<<<<<<<
 *             if len(self.buffer) < 2:
 *                 return
 */
  __pyx_t_3 = (((PyObject *)__pyx_v_frame) == Py_None);
  __pyx_t_7 = (__pyx_t_3 != 0);
  if (__pyx_t_7) {

    /* "extensions/lib/websocket.pyx":187
 *             self.buffer.extend(data)
 *         if frame is None:
 *             if len(self.buffer) < 2:             # <<<<<<<<<<<<<<
 *                 return
 *             first_byte, second_byte = unpack("BB", self.buffer[:2])
 */
    __pyx_t_1 = __pyx_v_self->buffer;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_8 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(1, 187, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = ((__pyx_t_8 < 2) != 0);
    if (__pyx_t_7) {

      /* "extensions/lib/websocket.pyx":188
 *         if frame is None:
 *             if len(self.buffer) < 2:
 *                 return             # <<<<<<<<<<<<<<
 *             first_byte, second_byte = unpack("BB", self.buffer[:2])
 *             fin = (first_byte >> 7) & 1
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "extensions/lib/websocket.pyx":187
 *             self.buffer.extend(data)
 *         if frame is None:
 *             if len(self.buffer) < 2:             # <<<<<<<<<<<<<<
 *                 return
 *             first_byte, second_byte = unpack("BB", self.buffer[:2])
 */
    }

    /* "extensions/lib/websocket.pyx":189
 *             if len(self.buffer) < 2:
 *                 return
 *             first_byte, second_byte = unpack("BB", self.buffer[:2])             # <<<<<<<<<<<<<<
 *             fin = (first_byte >> 7) & 1
 *             rsv1 = (first_byte >> 6) & 1
 */
    __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 189, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = __Pyx_PyObject_GetSlice(__pyx_v_self->buffer, 0, 2, NULL, NULL, &__pyx_slice__19, 0, 1, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 189, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
        __pyx_t_8 = 1;
      }
    }
    __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 189, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    if (__pyx_t_5) {
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_5); __pyx_t_5 = NULL;
    }
    __Pyx_INCREF(__pyx_n_s_BB);
    __Pyx_GIVEREF(__pyx_n_s_BB);
    PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_n_s_BB);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_t_6);
    __pyx_t_6 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 189, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
      PyObject* sequence = __pyx_t_1;
      #if CYTHON_COMPILING_IN_CPYTHON
      Py_ssize_t size = Py_SIZE(sequence);
      #else
      Py_ssize_t size = PySequence_Size(sequence);
      #endif
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 189, __pyx_L1_error)
      }
      #if CYTHON_COMPILING_IN_CPYTHON
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 189, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_9 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 189, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_6 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 189, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_10 = Py_TYPE(__pyx_t_6)->tp_iternext;
      index = 0; __pyx_t_4 = __pyx_t_10(__pyx_t_6); if (unlikely(!__pyx_t_4)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      index = 1; __pyx_t_9 = __pyx_t_10(__pyx_t_6); if (unlikely(!__pyx_t_9)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_10(__pyx_t_6), 2) < 0) __PYX_ERR(1, 189, __pyx_L1_error)
      __pyx_t_10 = NULL;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      goto __pyx_L7_unpacking_done;
      __pyx_L6_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_10 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(1, 189, __pyx_L1_error)
      __pyx_L7_unpacking_done:;
    }
    __pyx_v_first_byte = __pyx_t_4;
    __pyx_t_4 = 0;
    __pyx_v_second_byte = __pyx_t_9;
    __pyx_t_9 = 0;

    /* "extensions/lib/websocket.pyx":190
 *                 return
 *             first_byte, second_byte = unpack("BB", self.buffer[:2])
 *             fin = (first_byte >> 7) & 1             # <<<<<<<<<<<<<<
 *             rsv1 = (first_byte >> 6) & 1
 *             rsv2 = (first_byte >> 5) & 1
 */
    __pyx_t_1 = __Pyx_PyInt_RshiftObjC(__pyx_v_first_byte, __pyx_int_7, 7, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = __Pyx_PyInt_AndObjC(__pyx_t_1, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_9); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 190, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_v_fin = __pyx_t_2;

    /* "extensions/lib/websocket.pyx":191
 *             first_byte, second_byte = unpack("BB", self.buffer[:2])
 *             fin = (first_byte >> 7) & 1
 *             rsv1 = (first_byte >> 6) & 1             # <<<<<<<<<<<<<<
 *             rsv2 = (first_byte >> 5) & 1
 *             rsv3 = (first_byte >> 4) & 1
 */
    __pyx_t_9 = __Pyx_PyInt_RshiftObjC(__pyx_v_first_byte, __pyx_int_6, 6, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 191, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_1 = __Pyx_PyInt_AndObjC(__pyx_t_9, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 191, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 191, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_rsv1 = __pyx_t_2;

    /* "extensions/lib/websocket.pyx":192
 *             fin = (first_byte >> 7) & 1
 *             rsv1 = (first_byte >> 6) & 1
 *             rsv2 = (first_byte >> 5) & 1             # <<<<<<<<<<<<<<
 *             rsv3 = (first_byte >> 4) & 1
 *             opcode = first_byte & 0xf
 */
    __pyx_t_1 = __Pyx_PyInt_RshiftObjC(__pyx_v_first_byte, __pyx_int_5, 5, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 192, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = __Pyx_PyInt_AndObjC(__pyx_t_1, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 192, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_9); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 192, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_v_rsv2 = __pyx_t_2;

    /* "extensions/lib/websocket.pyx":193
 *             rsv1 = (first_byte >> 6) & 1
 *             rsv2 = (first_byte >> 5) & 1
 *             rsv3 = (first_byte >> 4) & 1             # <<<<<<<<<<<<<<
 *             opcode = first_byte & 0xf
 *             if fin not in (0, 1):
 */
    __pyx_t_9 = __Pyx_PyInt_RshiftObjC(__pyx_v_first_byte, __pyx_int_4, 4, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_1 = __Pyx_PyInt_AndObjC(__pyx_t_9, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 193, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_rsv3 = __pyx_t_2;

    /* "extensions/lib/websocket.pyx":194
 *             rsv2 = (first_byte >> 5) & 1
 *             rsv3 = (first_byte >> 4) & 1
 *             opcode = first_byte & 0xf             # <<<<<<<<<<<<<<
 *             if fin not in (0, 1):
 *                 raise self.ProtocolError('FIN must be 0 or 1')
 */
    __pyx_t_1 = __Pyx_PyInt_AndObjC(__pyx_v_first_byte, __pyx_int_15, 0xf, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 194, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 194, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_opcode = __pyx_t_2;

    /* "extensions/lib/websocket.pyx":195
 *             rsv3 = (first_byte >> 4) & 1
 *             opcode = first_byte & 0xf
 *             if fin not in (0, 1):             # <<<<<<<<<<<<<<
 *                 raise self.ProtocolError('FIN must be 0 or 1')
 *             if bool(mask_length) != bool(second_byte & 0x80):
 */
    switch (__pyx_v_fin) {
      case 0:
      case 1:
      __pyx_t_7 = 0;
      break;
      default:
      __pyx_t_7 = 1;
      break;
    }
    __pyx_t_3 = (__pyx_t_7 != 0);
    if (__pyx_t_3) {

      /* "extensions/lib/websocket.pyx":196
 *             opcode = first_byte & 0xf
 *             if fin not in (0, 1):
 *                 raise self.ProtocolError('FIN must be 0 or 1')             # <<<<<<<<<<<<<<
 *             if bool(mask_length) != bool(second_byte & 0x80):
 *                 if mask_length:
 */
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_v_self->ProtocolError, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 196, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_Raise(__pyx_t_1, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __PYX_ERR(1, 196, __pyx_L1_error)

      /* "extensions/lib/websocket.pyx":195
 *             rsv3 = (first_byte >> 4) & 1
 *             opcode = first_byte & 0xf
 *             if fin not in (0, 1):             # <<<<<<<<<<<<<<
 *                 raise self.ProtocolError('FIN must be 0 or 1')
 *             if bool(mask_length) != bool(second_byte & 0x80):
 */
    }

    /* "extensions/lib/websocket.pyx":197
 *             if fin not in (0, 1):
 *                 raise self.ProtocolError('FIN must be 0 or 1')
 *             if bool(mask_length) != bool(second_byte & 0x80):             # <<<<<<<<<<<<<<
 *                 if mask_length:
 *                     raise self.ProtocolError('unmasked client frame.')
 */
    __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_mask_length); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 197, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 197, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyBool_FromLong((!(!__pyx_t_3))); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 197, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_9 = __Pyx_PyInt_AndObjC(__pyx_v_second_byte, __pyx_int_128, 0x80, 0); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 197, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 197, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = __Pyx_PyBool_FromLong((!(!__pyx_t_3))); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 197, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_4 = PyObject_RichCompare(__pyx_t_1, __pyx_t_9, Py_NE); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 197, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(1, 197, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_3) {

      /* "extensions/lib/websocket.pyx":198
 *                 raise self.ProtocolError('FIN must be 0 or 1')
 *             if bool(mask_length) != bool(second_byte & 0x80):
 *                 if mask_length:             # <<<<<<<<<<<<<<
 *                     raise self.ProtocolError('unmasked client frame.')
 *                 else:
 */
      __pyx_t_3 = (__pyx_v_mask_length != 0);
      if (__pyx_t_3) {

        /* "extensions/lib/websocket.pyx":199
 *             if bool(mask_length) != bool(second_byte & 0x80):
 *                 if mask_length:
 *                     raise self.ProtocolError('unmasked client frame.')             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise self.ProtocolError('masked server frame.')
 */
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_v_self->ProtocolError, __pyx_tuple__21, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 199, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(1, 199, __pyx_L1_error)

        /* "extensions/lib/websocket.pyx":198
 *                 raise self.ProtocolError('FIN must be 0 or 1')
 *             if bool(mask_length) != bool(second_byte & 0x80):
 *                 if mask_length:             # <<<<<<<<<<<<<<
 *                     raise self.ProtocolError('unmasked client frame.')
 *                 else:
 */
      }

      /* "extensions/lib/websocket.pyx":201
 *                     raise self.ProtocolError('unmasked client frame.')
 *                 else:
 *                     raise self.ProtocolError('masked server frame.')             # <<<<<<<<<<<<<<
 *             payload_length = second_byte & 0x7f
 *             # All control frames MUST have a payload length of 125 bytes
 */
      /*else*/ {
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_v_self->ProtocolError, __pyx_tuple__22, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 201, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(1, 201, __pyx_L1_error)
      }

      /* "extensions/lib/websocket.pyx":197
 *             if fin not in (0, 1):
 *                 raise self.ProtocolError('FIN must be 0 or 1')
 *             if bool(mask_length) != bool(second_byte & 0x80):             # <<<<<<<<<<<<<<
 *                 if mask_length:
 *                     raise self.ProtocolError('unmasked client frame.')
 */
    }

    /* "extensions/lib/websocket.pyx":202
 *                 else:
 *                     raise self.ProtocolError('masked server frame.')
 *             payload_length = second_byte & 0x7f             # <<<<<<<<<<<<<<
 *             # All control frames MUST have a payload length of 125 bytes
 *             # or less
 */
    __pyx_t_4 = __Pyx_PyInt_AndObjC(__pyx_v_second_byte, __pyx_int_127, 0x7f, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_4); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 202, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_payload_length = __pyx_t_2;

    /* "extensions/lib/websocket.pyx":205
 *             # All control frames MUST have a payload length of 125 bytes
 *             # or less
 *             if opcode > 7:             # <<<<<<<<<<<<<<
 *                 if payload_length > 125:
 *                     raise self.ProtocolError(
 */
    __pyx_t_3 = ((__pyx_v_opcode > 7) != 0);
    if (__pyx_t_3) {

      /* "extensions/lib/websocket.pyx":206
 *             # or less
 *             if opcode > 7:
 *                 if payload_length > 125:             # <<<<<<<<<<<<<<
 *                     raise self.ProtocolError(
 *                         'WEBSOCKET control frame too large')
 */
      __pyx_t_3 = ((__pyx_v_payload_length > 0x7D) != 0);
      if (__pyx_t_3) {

        /* "extensions/lib/websocket.pyx":207
 *             if opcode > 7:
 *                 if payload_length > 125:
 *                     raise self.ProtocolError(             # <<<<<<<<<<<<<<
 *                         'WEBSOCKET control frame too large')
 *                 elif not fin:
 */
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_v_self->ProtocolError, __pyx_tuple__23, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 207, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(1, 207, __pyx_L1_error)

        /* "extensions/lib/websocket.pyx":206
 *             # or less
 *             if opcode > 7:
 *                 if payload_length > 125:             # <<<<<<<<<<<<<<
 *                     raise self.ProtocolError(
 *                         'WEBSOCKET control frame too large')
 */
      }

      /* "extensions/lib/websocket.pyx":209
 *                     raise self.ProtocolError(
 *                         'WEBSOCKET control frame too large')
 *                 elif not fin:             # <<<<<<<<<<<<<<
 *                     raise self.ProtocolError(
 *                         'WEBSOCKET control frame fragmented')
 */
      __pyx_t_3 = ((!(__pyx_v_fin != 0)) != 0);
      if (__pyx_t_3) {

        /* "extensions/lib/websocket.pyx":210
 *                         'WEBSOCKET control frame too large')
 *                 elif not fin:
 *                     raise self.ProtocolError(             # <<<<<<<<<<<<<<
 *                         'WEBSOCKET control frame fragmented')
 *             chunk = self._chunk(2)
 */
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_v_self->ProtocolError, __pyx_tuple__24, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 210, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(1, 210, __pyx_L1_error)

        /* "extensions/lib/websocket.pyx":209
 *                     raise self.ProtocolError(
 *                         'WEBSOCKET control frame too large')
 *                 elif not fin:             # <<<<<<<<<<<<<<
 *                     raise self.ProtocolError(
 *                         'WEBSOCKET control frame fragmented')
 */
      }

      /* "extensions/lib/websocket.pyx":205
 *             # All control frames MUST have a payload length of 125 bytes
 *             # or less
 *             if opcode > 7:             # <<<<<<<<<<<<<<
 *                 if payload_length > 125:
 *                     raise self.ProtocolError(
 */
    }

    /* "extensions/lib/websocket.pyx":212
 *                     raise self.ProtocolError(
 *                         'WEBSOCKET control frame fragmented')
 *             chunk = self._chunk(2)             # <<<<<<<<<<<<<<
 *             self.frame = frame = Frame(opcode, <bint>fin, payload_length)
 * 
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_3lib_FrameParser *)__pyx_v_self->__pyx_vtab)->_chunk(__pyx_v_self, 2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 212, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_v_chunk = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "extensions/lib/websocket.pyx":213
 *                         'WEBSOCKET control frame fragmented')
 *             chunk = self._chunk(2)
 *             self.frame = frame = Frame(opcode, <bint>fin, payload_length)             # <<<<<<<<<<<<<<
 * 
 *         if frame._masking_key is None:
 */
    __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_opcode); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_9 = __Pyx_PyBool_FromLong((__pyx_v_fin != 0)); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_payload_length); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = PyTuple_New(3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_6, 1, __pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_6, 2, __pyx_t_1);
    __pyx_t_4 = 0;
    __pyx_t_9 = 0;
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_3lib_Frame), __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_1);
    __Pyx_GOTREF(__pyx_v_self->frame);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->frame));
    __pyx_v_self->frame = ((struct __pyx_obj_3lib_Frame *)__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_frame, ((struct __pyx_obj_3lib_Frame *)__pyx_t_1));
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "extensions/lib/websocket.pyx":186
 *         if data:
 *             self.buffer.extend(data)
 *         if frame is None:             # <<<<<<<<<<<<<<
 *             if len(self.buffer) < 2:
 *                 return
 */
  }

  /* "extensions/lib/websocket.pyx":215
 *             self.frame = frame = Frame(opcode, <bint>fin, payload_length)
 * 
 *         if frame._masking_key is None:             # <<<<<<<<<<<<<<
 *             if frame._payload_length == 126:
 *                 if len(self.buffer) < 2 + mask_length:  # 2 + 4 for mask
 */
  __pyx_t_3 = (__pyx_v_frame->_masking_key == ((PyObject*)Py_None));
  __pyx_t_7 = (__pyx_t_3 != 0);
  if (__pyx_t_7) {

    /* "extensions/lib/websocket.pyx":216
 * 
 *         if frame._masking_key is None:
 *             if frame._payload_length == 126:             # <<<<<<<<<<<<<<
 *                 if len(self.buffer) < 2 + mask_length:  # 2 + 4 for mask
 *                     return
 */
    __pyx_t_7 = ((__pyx_v_frame->_payload_length == 0x7E) != 0);
    if (__pyx_t_7) {

      /* "extensions/lib/websocket.pyx":217
 *         if frame._masking_key is None:
 *             if frame._payload_length == 126:
 *                 if len(self.buffer) < 2 + mask_length:  # 2 + 4 for mask             # <<<<<<<<<<<<<<
 *                     return
 *                 chunk = self._chunk(2)
 */
      __pyx_t_1 = __pyx_v_self->buffer;
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_8 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(1, 217, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_7 = ((__pyx_t_8 < (2 + __pyx_v_mask_length)) != 0);
      if (__pyx_t_7) {

        /* "extensions/lib/websocket.pyx":218
 *             if frame._payload_length == 126:
 *                 if len(self.buffer) < 2 + mask_length:  # 2 + 4 for mask
 *                     return             # <<<<<<<<<<<<<<
 *                 chunk = self._chunk(2)
 *                 frame._payload_length = unpack("!H", chunk)[0]
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        goto __pyx_L0;

        /* "extensions/lib/websocket.pyx":217
 *         if frame._masking_key is None:
 *             if frame._payload_length == 126:
 *                 if len(self.buffer) < 2 + mask_length:  # 2 + 4 for mask             # <<<<<<<<<<<<<<
 *                     return
 *                 chunk = self._chunk(2)
 */
      }

      /* "extensions/lib/websocket.pyx":219
 *                 if len(self.buffer) < 2 + mask_length:  # 2 + 4 for mask
 *                     return
 *                 chunk = self._chunk(2)             # <<<<<<<<<<<<<<
 *                 frame._payload_length = unpack("!H", chunk)[0]
 *             elif frame._payload_length == 127:
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_3lib_FrameParser *)__pyx_v_self->__pyx_vtab)->_chunk(__pyx_v_self, 2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 219, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_chunk, ((PyObject*)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "extensions/lib/websocket.pyx":220
 *                     return
 *                 chunk = self._chunk(2)
 *                 frame._payload_length = unpack("!H", chunk)[0]             # <<<<<<<<<<<<<<
 *             elif frame._payload_length == 127:
 *                 if len(self.buffer) < 8 + mask_length:  # 8 + 4 for mask
 */
      __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 220, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_9 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_6))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_6);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_6, function);
          __pyx_t_8 = 1;
        }
      }
      __pyx_t_4 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 220, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (__pyx_t_9) {
        __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_9); __pyx_t_9 = NULL;
      }
      __Pyx_INCREF(__pyx_kp_s_H);
      __Pyx_GIVEREF(__pyx_kp_s_H);
      PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_8, __pyx_kp_s_H);
      __Pyx_INCREF(__pyx_v_chunk);
      __Pyx_GIVEREF(__pyx_v_chunk);
      PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_8, __pyx_v_chunk);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 220, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 220, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_6); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 220, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_v_frame->_payload_length = __pyx_t_2;

      /* "extensions/lib/websocket.pyx":216
 * 
 *         if frame._masking_key is None:
 *             if frame._payload_length == 126:             # <<<<<<<<<<<<<<
 *                 if len(self.buffer) < 2 + mask_length:  # 2 + 4 for mask
 *                     return
 */
      goto __pyx_L14;
    }

    /* "extensions/lib/websocket.pyx":221
 *                 chunk = self._chunk(2)
 *                 frame._payload_length = unpack("!H", chunk)[0]
 *             elif frame._payload_length == 127:             # <<<<<<<<<<<<<<
 *                 if len(self.buffer) < 8 + mask_length:  # 8 + 4 for mask
 *                     return
 */
    __pyx_t_7 = ((__pyx_v_frame->_payload_length == 0x7F) != 0);
    if (__pyx_t_7) {

      /* "extensions/lib/websocket.pyx":222
 *                 frame._payload_length = unpack("!H", chunk)[0]
 *             elif frame._payload_length == 127:
 *                 if len(self.buffer) < 8 + mask_length:  # 8 + 4 for mask             # <<<<<<<<<<<<<<
 *                     return
 *                 chunk = self._chunk(8)
 */
      __pyx_t_6 = __pyx_v_self->buffer;
      __Pyx_INCREF(__pyx_t_6);
      __pyx_t_8 = PyObject_Length(__pyx_t_6); if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(1, 222, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_7 = ((__pyx_t_8 < (8 + __pyx_v_mask_length)) != 0);
      if (__pyx_t_7) {

        /* "extensions/lib/websocket.pyx":223
 *             elif frame._payload_length == 127:
 *                 if len(self.buffer) < 8 + mask_length:  # 8 + 4 for mask
 *                     return             # <<<<<<<<<<<<<<
 *                 chunk = self._chunk(8)
 *                 frame._payload_length = unpack("!Q", chunk)[0]
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        goto __pyx_L0;

        /* "extensions/lib/websocket.pyx":222
 *                 frame._payload_length = unpack("!H", chunk)[0]
 *             elif frame._payload_length == 127:
 *                 if len(self.buffer) < 8 + mask_length:  # 8 + 4 for mask             # <<<<<<<<<<<<<<
 *                     return
 *                 chunk = self._chunk(8)
 */
      }

      /* "extensions/lib/websocket.pyx":224
 *                 if len(self.buffer) < 8 + mask_length:  # 8 + 4 for mask
 *                     return
 *                 chunk = self._chunk(8)             # <<<<<<<<<<<<<<
 *                 frame._payload_length = unpack("!Q", chunk)[0]
 *             elif len(self.buffer) < mask_length:
 */
      __pyx_t_6 = ((struct __pyx_vtabstruct_3lib_FrameParser *)__pyx_v_self->__pyx_vtab)->_chunk(__pyx_v_self, 8); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 224, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_chunk, ((PyObject*)__pyx_t_6));
      __pyx_t_6 = 0;

      /* "extensions/lib/websocket.pyx":225
 *                     return
 *                 chunk = self._chunk(8)
 *                 frame._payload_length = unpack("!Q", chunk)[0]             # <<<<<<<<<<<<<<
 *             elif len(self.buffer) < mask_length:
 *                 return
 */
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_unpack); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 225, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_1))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_1);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_1, function);
          __pyx_t_8 = 1;
        }
      }
      __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 225, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (__pyx_t_4) {
        __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_4); __pyx_t_4 = NULL;
      }
      __Pyx_INCREF(__pyx_kp_s_Q);
      __Pyx_GIVEREF(__pyx_kp_s_Q);
      PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_kp_s_Q);
      __Pyx_INCREF(__pyx_v_chunk);
      __Pyx_GIVEREF(__pyx_v_chunk);
      PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_chunk);
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_9, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 225, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_1 = __Pyx_GetItemInt(__pyx_t_6, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 225, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 225, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_v_frame->_payload_length = __pyx_t_2;

      /* "extensions/lib/websocket.pyx":221
 *                 chunk = self._chunk(2)
 *                 frame._payload_length = unpack("!H", chunk)[0]
 *             elif frame._payload_length == 127:             # <<<<<<<<<<<<<<
 *                 if len(self.buffer) < 8 + mask_length:  # 8 + 4 for mask
 *                     return
 */
      goto __pyx_L14;
    }

    /* "extensions/lib/websocket.pyx":226
 *                 chunk = self._chunk(8)
 *                 frame._payload_length = unpack("!Q", chunk)[0]
 *             elif len(self.buffer) < mask_length:             # <<<<<<<<<<<<<<
 *                 return
 *             if mask_length:
 */
    __pyx_t_1 = __pyx_v_self->buffer;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_8 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(1, 226, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = ((__pyx_t_8 < __pyx_v_mask_length) != 0);
    if (__pyx_t_7) {

      /* "extensions/lib/websocket.pyx":227
 *                 frame._payload_length = unpack("!Q", chunk)[0]
 *             elif len(self.buffer) < mask_length:
 *                 return             # <<<<<<<<<<<<<<
 *             if mask_length:
 *                 frame._masking_key = self._chunk(mask_length)
 */
      __Pyx_XDECREF(__pyx_r);
      __pyx_r = Py_None; __Pyx_INCREF(Py_None);
      goto __pyx_L0;

      /* "extensions/lib/websocket.pyx":226
 *                 chunk = self._chunk(8)
 *                 frame._payload_length = unpack("!Q", chunk)[0]
 *             elif len(self.buffer) < mask_length:             # <<<<<<<<<<<<<<
 *                 return
 *             if mask_length:
 */
    }
    __pyx_L14:;

    /* "extensions/lib/websocket.pyx":228
 *             elif len(self.buffer) < mask_length:
 *                 return
 *             if mask_length:             # <<<<<<<<<<<<<<
 *                 frame._masking_key = self._chunk(mask_length)
 *             else:
 */
    __pyx_t_7 = (__pyx_v_mask_length != 0);
    if (__pyx_t_7) {

      /* "extensions/lib/websocket.pyx":229
 *                 return
 *             if mask_length:
 *                 frame._masking_key = self._chunk(mask_length)             # <<<<<<<<<<<<<<
 *             else:
 *                 frame._masking_key = b''
 */
      __pyx_t_1 = ((struct __pyx_vtabstruct_3lib_FrameParser *)__pyx_v_self->__pyx_vtab)->_chunk(__pyx_v_self, __pyx_v_mask_length); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 229, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_1);
      __Pyx_GOTREF(__pyx_v_frame->_masking_key);
      __Pyx_DECREF(__pyx_v_frame->_masking_key);
      __pyx_v_frame->_masking_key = ((PyObject*)__pyx_t_1);
      __pyx_t_1 = 0;

      /* "extensions/lib/websocket.pyx":228
 *             elif len(self.buffer) < mask_length:
 *                 return
 *             if mask_length:             # <<<<<<<<<<<<<<
 *                 frame._masking_key = self._chunk(mask_length)
 *             else:
 */
      goto __pyx_L17;
    }

    /* "extensions/lib/websocket.pyx":231
 *                 frame._masking_key = self._chunk(mask_length)
 *             else:
 *                 frame._masking_key = b''             # <<<<<<<<<<<<<<
 * 
 *         if len(self.buffer) >= frame._payload_length:
 */
    /*else*/ {
      __Pyx_INCREF(__pyx_kp_b_);
      __Pyx_GIVEREF(__pyx_kp_b_);
      __Pyx_GOTREF(__pyx_v_frame->_masking_key);
      __Pyx_DECREF(__pyx_v_frame->_masking_key);
      __pyx_v_frame->_masking_key = __pyx_kp_b_;
    }
    __pyx_L17:;

    /* "extensions/lib/websocket.pyx":215
 *             self.frame = frame = Frame(opcode, <bint>fin, payload_length)
 * 
 *         if frame._masking_key is None:             # <<<<<<<<<<<<<<
 *             if frame._payload_length == 126:
 *                 if len(self.buffer) < 2 + mask_length:  # 2 + 4 for mask
 */
  }

  /* "extensions/lib/websocket.pyx":233
 *                 frame._masking_key = b''
 * 
 *         if len(self.buffer) >= frame._payload_length:             # <<<<<<<<<<<<<<
 *             self.frame = None
 *             chunk = self._chunk(frame._payload_length)
 */
  __pyx_t_1 = __pyx_v_self->buffer;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_8 = PyObject_Length(__pyx_t_1); if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(1, 233, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = ((__pyx_t_8 >= __pyx_v_frame->_payload_length) != 0);
  if (__pyx_t_7) {

    /* "extensions/lib/websocket.pyx":234
 * 
 *         if len(self.buffer) >= frame._payload_length:
 *             self.frame = None             # <<<<<<<<<<<<<<
 *             chunk = self._chunk(frame._payload_length)
 *             if self._extensions:
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_self->frame);
    __Pyx_DECREF(((PyObject *)__pyx_v_self->frame));
    __pyx_v_self->frame = ((struct __pyx_obj_3lib_Frame *)Py_None);

    /* "extensions/lib/websocket.pyx":235
 *         if len(self.buffer) >= frame._payload_length:
 *             self.frame = None
 *             chunk = self._chunk(frame._payload_length)             # <<<<<<<<<<<<<<
 *             if self._extensions:
 *                 for extension in self._extensions:
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_3lib_FrameParser *)__pyx_v_self->__pyx_vtab)->_chunk(__pyx_v_self, __pyx_v_frame->_payload_length); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 235, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_chunk, ((PyObject*)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "extensions/lib/websocket.pyx":236
 *             self.frame = None
 *             chunk = self._chunk(frame._payload_length)
 *             if self._extensions:             # <<<<<<<<<<<<<<
 *                 for extension in self._extensions:
 *                     chunk = extension.receive(frame, self.buffer)
 */
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_self->_extensions); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(1, 236, __pyx_L1_error)
    if (__pyx_t_7) {

      /* "extensions/lib/websocket.pyx":237
 *             chunk = self._chunk(frame._payload_length)
 *             if self._extensions:
 *                 for extension in self._extensions:             # <<<<<<<<<<<<<<
 *                     chunk = extension.receive(frame, self.buffer)
 *             if frame._masking_key:
 */
      if (likely(PyList_CheckExact(__pyx_v_self->_extensions)) || PyTuple_CheckExact(__pyx_v_self->_extensions)) {
        __pyx_t_1 = __pyx_v_self->_extensions; __Pyx_INCREF(__pyx_t_1); __pyx_t_8 = 0;
        __pyx_t_11 = NULL;
      } else {
        __pyx_t_8 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_v_self->_extensions); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 237, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_11 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 237, __pyx_L1_error)
      }
      for (;;) {
        if (likely(!__pyx_t_11)) {
          if (likely(PyList_CheckExact(__pyx_t_1))) {
            if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_1)) break;
            #if CYTHON_COMPILING_IN_CPYTHON
            __pyx_t_6 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_8); __Pyx_INCREF(__pyx_t_6); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(1, 237, __pyx_L1_error)
            #else
            __pyx_t_6 = PySequence_ITEM(__pyx_t_1, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 237, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            #endif
          } else {
            if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
            #if CYTHON_COMPILING_IN_CPYTHON
            __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_8); __Pyx_INCREF(__pyx_t_6); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(1, 237, __pyx_L1_error)
            #else
            __pyx_t_6 = PySequence_ITEM(__pyx_t_1, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 237, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            #endif
          }
        } else {
          __pyx_t_6 = __pyx_t_11(__pyx_t_1);
          if (unlikely(!__pyx_t_6)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(1, 237, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_6);
        }
        __Pyx_XDECREF_SET(__pyx_v_extension, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "extensions/lib/websocket.pyx":238
 *             if self._extensions:
 *                 for extension in self._extensions:
 *                     chunk = extension.receive(frame, self.buffer)             # <<<<<<<<<<<<<<
 *             if frame._masking_key:
 *                 chunk = websocket_mask(chunk, frame._masking_key,
 */
        __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_extension, __pyx_n_s_receive); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 238, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        __pyx_t_4 = NULL;
        __pyx_t_12 = 0;
        if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_9))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_9);
          if (likely(__pyx_t_4)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
            __Pyx_INCREF(__pyx_t_4);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_9, function);
            __pyx_t_12 = 1;
          }
        }
        __pyx_t_5 = PyTuple_New(2+__pyx_t_12); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 238, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        if (__pyx_t_4) {
          __Pyx_GIVEREF(__pyx_t_4); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4); __pyx_t_4 = NULL;
        }
        __Pyx_INCREF(((PyObject *)__pyx_v_frame));
        __Pyx_GIVEREF(((PyObject *)__pyx_v_frame));
        PyTuple_SET_ITEM(__pyx_t_5, 0+__pyx_t_12, ((PyObject *)__pyx_v_frame));
        __Pyx_INCREF(__pyx_v_self->buffer);
        __Pyx_GIVEREF(__pyx_v_self->buffer);
        PyTuple_SET_ITEM(__pyx_t_5, 1+__pyx_t_12, __pyx_v_self->buffer);
        __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_9, __pyx_t_5, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 238, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        if (!(likely(PyBytes_CheckExact(__pyx_t_6))||((__pyx_t_6) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_6)->tp_name), 0))) __PYX_ERR(1, 238, __pyx_L1_error)
        __Pyx_DECREF_SET(__pyx_v_chunk, ((PyObject*)__pyx_t_6));
        __pyx_t_6 = 0;

        /* "extensions/lib/websocket.pyx":237
 *             chunk = self._chunk(frame._payload_length)
 *             if self._extensions:
 *                 for extension in self._extensions:             # <<<<<<<<<<<<<<
 *                     chunk = extension.receive(frame, self.buffer)
 *             if frame._masking_key:
 */
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "extensions/lib/websocket.pyx":236
 *             self.frame = None
 *             chunk = self._chunk(frame._payload_length)
 *             if self._extensions:             # <<<<<<<<<<<<<<
 *                 for extension in self._extensions:
 *                     chunk = extension.receive(frame, self.buffer)
 */
    }

    /* "extensions/lib/websocket.pyx":239
 *                 for extension in self._extensions:
 *                     chunk = extension.receive(frame, self.buffer)
 *             if frame._masking_key:             # <<<<<<<<<<<<<<
 *                 chunk = websocket_mask(chunk, frame._masking_key,
 *                                        len(chunk), len(frame._masking_key))
 */
    __pyx_t_7 = (__pyx_v_frame->_masking_key != Py_None) && (PyBytes_GET_SIZE(__pyx_v_frame->_masking_key) != 0);
    if (__pyx_t_7) {

      /* "extensions/lib/websocket.pyx":240
 *                     chunk = extension.receive(frame, self.buffer)
 *             if frame._masking_key:
 *                 chunk = websocket_mask(chunk, frame._masking_key,             # <<<<<<<<<<<<<<
 *                                        len(chunk), len(frame._masking_key))
 *             if frame.opcode == 1:
 */
      __pyx_t_13 = __Pyx_PyObject_AsString(__pyx_v_chunk); if (unlikely((!__pyx_t_13) && PyErr_Occurred())) __PYX_ERR(1, 240, __pyx_L1_error)
      __pyx_t_14 = __Pyx_PyObject_AsString(__pyx_v_frame->_masking_key); if (unlikely((!__pyx_t_14) && PyErr_Occurred())) __PYX_ERR(1, 240, __pyx_L1_error)

      /* "extensions/lib/websocket.pyx":241
 *             if frame._masking_key:
 *                 chunk = websocket_mask(chunk, frame._masking_key,
 *                                        len(chunk), len(frame._masking_key))             # <<<<<<<<<<<<<<
 *             if frame.opcode == 1:
 *                 frame._body = chunk.decode("utf-8", "replace")
 */
      if (unlikely(__pyx_v_chunk == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
        __PYX_ERR(1, 241, __pyx_L1_error)
      }
      __pyx_t_8 = PyBytes_GET_SIZE(__pyx_v_chunk); if (unlikely(__pyx_t_8 == -1)) __PYX_ERR(1, 241, __pyx_L1_error)
      __pyx_t_1 = __pyx_v_frame->_masking_key;
      __Pyx_INCREF(__pyx_t_1);
      if (unlikely(__pyx_t_1 == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
        __PYX_ERR(1, 241, __pyx_L1_error)
      }
      __pyx_t_12 = PyBytes_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_12 == -1)) __PYX_ERR(1, 241, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "extensions/lib/websocket.pyx":240
 *                     chunk = extension.receive(frame, self.buffer)
 *             if frame._masking_key:
 *                 chunk = websocket_mask(chunk, frame._masking_key,             # <<<<<<<<<<<<<<
 *                                        len(chunk), len(frame._masking_key))
 *             if frame.opcode == 1:
 */
      __pyx_t_1 = websocket_mask(__pyx_t_13, __pyx_t_14, __pyx_t_8, __pyx_t_12); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 240, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!(likely(PyBytes_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_1)->tp_name), 0))) __PYX_ERR(1, 240, __pyx_L1_error)
      __Pyx_DECREF_SET(__pyx_v_chunk, ((PyObject*)__pyx_t_1));
      __pyx_t_1 = 0;

      /* "extensions/lib/websocket.pyx":239
 *                 for extension in self._extensions:
 *                     chunk = extension.receive(frame, self.buffer)
 *             if frame._masking_key:             # <<<<<<<<<<<<<<
 *                 chunk = websocket_mask(chunk, frame._masking_key,
 *                                        len(chunk), len(frame._masking_key))
 */
    }

    /* "extensions/lib/websocket.pyx":242
 *                 chunk = websocket_mask(chunk, frame._masking_key,
 *                                        len(chunk), len(frame._masking_key))
 *             if frame.opcode == 1:             # <<<<<<<<<<<<<<
 *                 frame._body = chunk.decode("utf-8", "replace")
 *             else:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_frame), __pyx_n_s_opcode); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 242, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyInt_EqObjC(__pyx_t_1, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 242, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(1, 242, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (__pyx_t_7) {

      /* "extensions/lib/websocket.pyx":243
 *                                        len(chunk), len(frame._masking_key))
 *             if frame.opcode == 1:
 *                 frame._body = chunk.decode("utf-8", "replace")             # <<<<<<<<<<<<<<
 *             else:
 *                 frame._body = chunk
 */
      if (unlikely(__pyx_v_chunk == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "decode");
        __PYX_ERR(1, 243, __pyx_L1_error)
      }
      __pyx_t_6 = __Pyx_decode_bytes(__pyx_v_chunk, 0, PY_SSIZE_T_MAX, NULL, ((char const *)"replace"), PyUnicode_DecodeUTF8); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 243, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_v_frame->_body);
      __Pyx_DECREF(__pyx_v_frame->_body);
      __pyx_v_frame->_body = __pyx_t_6;
      __pyx_t_6 = 0;

      /* "extensions/lib/websocket.pyx":242
 *                 chunk = websocket_mask(chunk, frame._masking_key,
 *                                        len(chunk), len(frame._masking_key))
 *             if frame.opcode == 1:             # <<<<<<<<<<<<<<
 *                 frame._body = chunk.decode("utf-8", "replace")
 *             else:
 */
      goto __pyx_L23;
    }

    /* "extensions/lib/websocket.pyx":245
 *                 frame._body = chunk.decode("utf-8", "replace")
 *             else:
 *                 frame._body = chunk             # <<<<<<<<<<<<<<
 *             return frame
 * 
 */
    /*else*/ {
      __Pyx_INCREF(__pyx_v_chunk);
      __Pyx_GIVEREF(__pyx_v_chunk);
      __Pyx_GOTREF(__pyx_v_frame->_body);
      __Pyx_DECREF(__pyx_v_frame->_body);
      __pyx_v_frame->_body = __pyx_v_chunk;
    }
    __pyx_L23:;

    /* "extensions/lib/websocket.pyx":246
 *             else:
 *                 frame._body = chunk
 *             return frame             # <<<<<<<<<<<<<<
 * 
 *     cdef bytes _encode(self, bytes data, int opcode, bytes masking_key,
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_frame));
    __pyx_r = ((PyObject *)__pyx_v_frame);
    goto __pyx_L0;

    /* "extensions/lib/websocket.pyx":233
 *                 frame._masking_key = b''
 * 
 *         if len(self.buffer) >= frame._payload_length:             # <<<<<<<<<<<<<<
 *             self.frame = None
 *             chunk = self._chunk(frame._payload_length)
 */
  }

  /* "extensions/lib/websocket.pyx":178
 *                                rsv1, rsv2, rsv3)
 * 
 *     def decode(self, bytes data=None):             # <<<<<<<<<<<<<<
 *         cdef int fin, rsv1, rsv2, rsv3, opcode, payload_length
 *         cdef Frame frame = self.frame
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("lib.FrameParser.decode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_frame);
  __Pyx_XDECREF(__pyx_v_chunk);
  __Pyx_XDECREF(__pyx_v_first_byte);
  __Pyx_XDECREF(__pyx_v_second_byte);
  __Pyx_XDECREF(__pyx_v_extension);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/websocket.pyx":248
 *             return frame
 * 
 *     cdef bytes _encode(self, bytes data, int opcode, bytes masking_key,             # <<<<<<<<<<<<<<
 *                        int fin, int rsv1, int rsv2, int rsv3):
 *         cdef object buffer = bytearray()
 */

static PyObject *__pyx_f_3lib_11FrameParser__encode(struct __pyx_obj_3lib_FrameParser *__pyx_v_self, PyObject *__pyx_v_data, int __pyx_v_opcode, PyObject *__pyx_v_masking_key, int __pyx_v_fin, int __pyx_v_rsv1, int __pyx_v_rsv2, int __pyx_v_rsv3) {
  PyObject *__pyx_v_buffer = 0;
  PY_LONG_LONG __pyx_v_length;
  int __pyx_v_mask_bit;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  char const *__pyx_t_12;
  char const *__pyx_t_13;
  __Pyx_RefNannySetupContext("_encode", 0);

  /* "extensions/lib/websocket.pyx":250
 *     cdef bytes _encode(self, bytes data, int opcode, bytes masking_key,
 *                        int fin, int rsv1, int rsv2, int rsv3):
 *         cdef object buffer = bytearray()             # <<<<<<<<<<<<<<
 *         cdef cython.longlong length = len(data)
 *         cdef int mask_bit = 128 if masking_key else 0
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)(&PyByteArray_Type)), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_buffer = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "extensions/lib/websocket.pyx":251
 *                        int fin, int rsv1, int rsv2, int rsv3):
 *         cdef object buffer = bytearray()
 *         cdef cython.longlong length = len(data)             # <<<<<<<<<<<<<<
 *         cdef int mask_bit = 128 if masking_key else 0
 * 
 */
  if (unlikely(__pyx_v_data == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 251, __pyx_L1_error)
  }
  __pyx_t_2 = PyBytes_GET_SIZE(__pyx_v_data); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(1, 251, __pyx_L1_error)
  __pyx_v_length = __pyx_t_2;

  /* "extensions/lib/websocket.pyx":252
 *         cdef object buffer = bytearray()
 *         cdef cython.longlong length = len(data)
 *         cdef int mask_bit = 128 if masking_key else 0             # <<<<<<<<<<<<<<
 * 
 *         buffer.append(((fin << 7) | (rsv1 << 6) | (rsv2 << 5) |
 */
  __pyx_t_4 = (__pyx_v_masking_key != Py_None) && (PyBytes_GET_SIZE(__pyx_v_masking_key) != 0);
  if (__pyx_t_4) {
    __pyx_t_3 = 0x80;
  } else {
    __pyx_t_3 = 0;
  }
  __pyx_v_mask_bit = __pyx_t_3;

  /* "extensions/lib/websocket.pyx":255
 * 
 *         buffer.append(((fin << 7) | (rsv1 << 6) | (rsv2 << 5) |
 *                        (rsv3 << 4) | opcode))             # <<<<<<<<<<<<<<
 * 
 *         if length < 126:
 */
  __pyx_t_1 = __Pyx_PyInt_From_long((((((__pyx_v_fin << 7) | (__pyx_v_rsv1 << 6)) | (__pyx_v_rsv2 << 5)) | (__pyx_v_rsv3 << 4)) | __pyx_v_opcode)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "extensions/lib/websocket.pyx":254
 *         cdef int mask_bit = 128 if masking_key else 0
 * 
 *         buffer.append(((fin << 7) | (rsv1 << 6) | (rsv2 << 5) |             # <<<<<<<<<<<<<<
 *                        (rsv3 << 4) | opcode))
 * 
 */
  __pyx_t_5 = __Pyx_PyObject_Append(__pyx_v_buffer, __pyx_t_1); if (unlikely(__pyx_t_5 == -1)) __PYX_ERR(1, 254, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "extensions/lib/websocket.pyx":257
 *                        (rsv3 << 4) | opcode))
 * 
 *         if length < 126:             # <<<<<<<<<<<<<<
 *             buffer.append(mask_bit | length)
 *         elif length < 65536:
 */
  __pyx_t_4 = ((__pyx_v_length < 0x7E) != 0);
  if (__pyx_t_4) {

    /* "extensions/lib/websocket.pyx":258
 * 
 *         if length < 126:
 *             buffer.append(mask_bit | length)             # <<<<<<<<<<<<<<
 *         elif length < 65536:
 *             buffer.append(mask_bit | 126)
 */
    __pyx_t_1 = __Pyx_PyInt_From_PY_LONG_LONG((__pyx_v_mask_bit | __pyx_v_length)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 258, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_Append(__pyx_v_buffer, __pyx_t_1); if (unlikely(__pyx_t_5 == -1)) __PYX_ERR(1, 258, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "extensions/lib/websocket.pyx":257
 *                        (rsv3 << 4) | opcode))
 * 
 *         if length < 126:             # <<<<<<<<<<<<<<
 *             buffer.append(mask_bit | length)
 *         elif length < 65536:
 */
    goto __pyx_L3;
  }

  /* "extensions/lib/websocket.pyx":259
 *         if length < 126:
 *             buffer.append(mask_bit | length)
 *         elif length < 65536:             # <<<<<<<<<<<<<<
 *             buffer.append(mask_bit | 126)
 *             buffer.extend(pack('!H', length))
 */
  __pyx_t_4 = ((__pyx_v_length < 0x10000) != 0);
  if (__pyx_t_4) {

    /* "extensions/lib/websocket.pyx":260
 *             buffer.append(mask_bit | length)
 *         elif length < 65536:
 *             buffer.append(mask_bit | 126)             # <<<<<<<<<<<<<<
 *             buffer.extend(pack('!H', length))
 *         elif length < (1 << 63):
 */
    __pyx_t_1 = __Pyx_PyInt_From_long((__pyx_v_mask_bit | 0x7E)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 260, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = __Pyx_PyObject_Append(__pyx_v_buffer, __pyx_t_1); if (unlikely(__pyx_t_5 == -1)) __PYX_ERR(1, 260, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "extensions/lib/websocket.pyx":261
 *         elif length < 65536:
 *             buffer.append(mask_bit | 126)
 *             buffer.extend(pack('!H', length))             # <<<<<<<<<<<<<<
 *         elif length < (1 << 63):
 *             buffer.append(mask_bit | 127)
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_buffer, __pyx_n_s_extend); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 261, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_pack_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 261, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = __Pyx_PyInt_From_PY_LONG_LONG(__pyx_v_length); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 261, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_10 = NULL;
    __pyx_t_2 = 0;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
        __pyx_t_2 = 1;
      }
    }
    __pyx_t_11 = PyTuple_New(2+__pyx_t_2); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 261, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    if (__pyx_t_10) {
      __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_10); __pyx_t_10 = NULL;
    }
    __Pyx_INCREF(__pyx_kp_s_H);
    __Pyx_GIVEREF(__pyx_kp_s_H);
    PyTuple_SET_ITEM(__pyx_t_11, 0+__pyx_t_2, __pyx_kp_s_H);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_11, 1+__pyx_t_2, __pyx_t_9);
    __pyx_t_9 = 0;
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_11, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 261, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    if (!__pyx_t_8) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_7); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 261, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else {
      __pyx_t_11 = PyTuple_New(1+1); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 261, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_8); __pyx_t_8 = NULL;
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_11, 0+1, __pyx_t_7);
      __pyx_t_7 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_11, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 261, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "extensions/lib/websocket.pyx":259
 *         if length < 126:
 *             buffer.append(mask_bit | length)
 *         elif length < 65536:             # <<<<<<<<<<<<<<
 *             buffer.append(mask_bit | 126)
 *             buffer.extend(pack('!H', length))
 */
    goto __pyx_L3;
  }

  /* "extensions/lib/websocket.pyx":262
 *             buffer.append(mask_bit | 126)
 *             buffer.extend(pack('!H', length))
 *         elif length < (1 << 63):             # <<<<<<<<<<<<<<
 *             buffer.append(mask_bit | 127)
 *             buffer.extend(pack('!Q', length))
 */
  __pyx_t_1 = __Pyx_PyInt_From_PY_LONG_LONG(__pyx_v_length); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = PyObject_RichCompare(__pyx_t_1, __pyx_int_9223372036854775808, Py_LT); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 262, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 262, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (__pyx_t_4) {

    /* "extensions/lib/websocket.pyx":263
 *             buffer.extend(pack('!H', length))
 *         elif length < (1 << 63):
 *             buffer.append(mask_bit | 127)             # <<<<<<<<<<<<<<
 *             buffer.extend(pack('!Q', length))
 *         else:
 */
    __pyx_t_6 = __Pyx_PyInt_From_long((__pyx_v_mask_bit | 0x7F)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = __Pyx_PyObject_Append(__pyx_v_buffer, __pyx_t_6); if (unlikely(__pyx_t_5 == -1)) __PYX_ERR(1, 263, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "extensions/lib/websocket.pyx":264
 *         elif length < (1 << 63):
 *             buffer.append(mask_bit | 127)
 *             buffer.extend(pack('!Q', length))             # <<<<<<<<<<<<<<
 *         else:
 *             raise self.ProtocolError('WEBSOCKET frame too large')
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buffer, __pyx_n_s_extend); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_pack_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyInt_From_PY_LONG_LONG(__pyx_v_length); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = NULL;
    __pyx_t_2 = 0;
    if (CYTHON_COMPILING_IN_CPYTHON && unlikely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_2 = 1;
      }
    }
    __pyx_t_10 = PyTuple_New(2+__pyx_t_2); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (__pyx_t_9) {
      __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9); __pyx_t_9 = NULL;
    }
    __Pyx_INCREF(__pyx_kp_s_Q);
    __Pyx_GIVEREF(__pyx_kp_s_Q);
    PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_2, __pyx_kp_s_Q);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_2, __pyx_t_8);
    __pyx_t_8 = 0;
    __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_10, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 264, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    if (!__pyx_t_7) {
      __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_11); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 264, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_GOTREF(__pyx_t_6);
    } else {
      __pyx_t_10 = PyTuple_New(1+1); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 264, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_7); __pyx_t_7 = NULL;
      __Pyx_GIVEREF(__pyx_t_11);
      PyTuple_SET_ITEM(__pyx_t_10, 0+1, __pyx_t_11);
      __pyx_t_11 = 0;
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_10, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 264, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "extensions/lib/websocket.pyx":262
 *             buffer.append(mask_bit | 126)
 *             buffer.extend(pack('!H', length))
 *         elif length < (1 << 63):             # <<<<<<<<<<<<<<
 *             buffer.append(mask_bit | 127)
 *             buffer.extend(pack('!Q', length))
 */
    goto __pyx_L3;
  }

  /* "extensions/lib/websocket.pyx":266
 *             buffer.extend(pack('!Q', length))
 *         else:
 *             raise self.ProtocolError('WEBSOCKET frame too large')             # <<<<<<<<<<<<<<
 *         if masking_key:
 *             buffer.extend(masking_key)
 */
  /*else*/ {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_v_self->ProtocolError, __pyx_tuple__25, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 266, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(1, 266, __pyx_L1_error)
  }
  __pyx_L3:;

  /* "extensions/lib/websocket.pyx":267
 *         else:
 *             raise self.ProtocolError('WEBSOCKET frame too large')
 *         if masking_key:             # <<<<<<<<<<<<<<
 *             buffer.extend(masking_key)
 *             buffer.extend(websocket_mask(data, masking_key,
 */
  __pyx_t_4 = (__pyx_v_masking_key != Py_None) && (PyBytes_GET_SIZE(__pyx_v_masking_key) != 0);
  if (__pyx_t_4) {

    /* "extensions/lib/websocket.pyx":268
 *             raise self.ProtocolError('WEBSOCKET frame too large')
 *         if masking_key:
 *             buffer.extend(masking_key)             # <<<<<<<<<<<<<<
 *             buffer.extend(websocket_mask(data, masking_key,
 *                                          length, len(masking_key)))
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buffer, __pyx_n_s_extend); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 268, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_10 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    if (!__pyx_t_10) {
      __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_masking_key); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 268, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
    } else {
      __pyx_t_11 = PyTuple_New(1+1); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 268, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_10); __pyx_t_10 = NULL;
      __Pyx_INCREF(__pyx_v_masking_key);
      __Pyx_GIVEREF(__pyx_v_masking_key);
      PyTuple_SET_ITEM(__pyx_t_11, 0+1, __pyx_v_masking_key);
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_11, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 268, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "extensions/lib/websocket.pyx":269
 *         if masking_key:
 *             buffer.extend(masking_key)
 *             buffer.extend(websocket_mask(data, masking_key,             # <<<<<<<<<<<<<<
 *                                          length, len(masking_key)))
 *         else:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buffer, __pyx_n_s_extend); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 269, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_12 = __Pyx_PyObject_AsString(__pyx_v_data); if (unlikely((!__pyx_t_12) && PyErr_Occurred())) __PYX_ERR(1, 269, __pyx_L1_error)
    __pyx_t_13 = __Pyx_PyObject_AsString(__pyx_v_masking_key); if (unlikely((!__pyx_t_13) && PyErr_Occurred())) __PYX_ERR(1, 269, __pyx_L1_error)

    /* "extensions/lib/websocket.pyx":270
 *             buffer.extend(masking_key)
 *             buffer.extend(websocket_mask(data, masking_key,
 *                                          length, len(masking_key)))             # <<<<<<<<<<<<<<
 *         else:
 *             buffer.extend(data)
 */
    if (unlikely(__pyx_v_masking_key == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(1, 270, __pyx_L1_error)
    }
    __pyx_t_2 = PyBytes_GET_SIZE(__pyx_v_masking_key); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(1, 270, __pyx_L1_error)

    /* "extensions/lib/websocket.pyx":269
 *         if masking_key:
 *             buffer.extend(masking_key)
 *             buffer.extend(websocket_mask(data, masking_key,             # <<<<<<<<<<<<<<
 *                                          length, len(masking_key)))
 *         else:
 */
    __pyx_t_11 = websocket_mask(__pyx_t_12, __pyx_t_13, __pyx_v_length, __pyx_t_2); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 269, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_10 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    if (!__pyx_t_10) {
      __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_11); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 269, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_GOTREF(__pyx_t_6);
    } else {
      __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 269, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_10); __pyx_t_10 = NULL;
      __Pyx_GIVEREF(__pyx_t_11);
      PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_11);
      __pyx_t_11 = 0;
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_7, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 269, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

    /* "extensions/lib/websocket.pyx":267
 *         else:
 *             raise self.ProtocolError('WEBSOCKET frame too large')
 *         if masking_key:             # <<<<<<<<<<<<<<
 *             buffer.extend(masking_key)
 *             buffer.extend(websocket_mask(data, masking_key,
 */
    goto __pyx_L4;
  }

  /* "extensions/lib/websocket.pyx":272
 *                                          length, len(masking_key)))
 *         else:
 *             buffer.extend(data)             # <<<<<<<<<<<<<<
 *         return bytes(buffer)
 * 
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_buffer, __pyx_n_s_extend); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 272, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    if (!__pyx_t_7) {
      __pyx_t_6 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_data); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 272, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
    } else {
      __pyx_t_11 = PyTuple_New(1+1); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 272, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_7); __pyx_t_7 = NULL;
      __Pyx_INCREF(__pyx_v_data);
      __Pyx_GIVEREF(__pyx_v_data);
      PyTuple_SET_ITEM(__pyx_t_11, 0+1, __pyx_v_data);
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_11, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 272, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __pyx_L4:;

  /* "extensions/lib/websocket.pyx":273
 *         else:
 *             buffer.extend(data)
 *         return bytes(buffer)             # <<<<<<<<<<<<<<
 * 
 *     cdef tuple _info(self, message, opcode, bytes masking_key):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_INCREF(__pyx_v_buffer);
  __Pyx_GIVEREF(__pyx_v_buffer);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_v_buffer);
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)(&PyBytes_Type)), __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "extensions/lib/websocket.pyx":248
 *             return frame
 * 
 *     cdef bytes _encode(self, bytes data, int opcode, bytes masking_key,             # <<<<<<<<<<<<<<
 *                        int fin, int rsv1, int rsv2, int rsv3):
 *         cdef object buffer = bytearray()
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("lib.FrameParser._encode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_buffer);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/websocket.pyx":275
 *         return bytes(buffer)
 * 
 *     cdef tuple _info(self, message, opcode, bytes masking_key):             # <<<<<<<<<<<<<<
 *         cdef int mask_length = self._encode_mask_length
 *         #
 */

static PyObject *__pyx_f_3lib_11FrameParser__info(struct __pyx_obj_3lib_FrameParser *__pyx_v_self, PyObject *__pyx_v_message, PyObject *__pyx_v_opcode, PyObject *__pyx_v_masking_key) {
  int __pyx_v_mask_length;
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("_info", 0);
  __Pyx_INCREF(__pyx_v_opcode);
  __Pyx_INCREF(__pyx_v_masking_key);

  /* "extensions/lib/websocket.pyx":276
 * 
 *     cdef tuple _info(self, message, opcode, bytes masking_key):
 *         cdef int mask_length = self._encode_mask_length             # <<<<<<<<<<<<<<
 *         #
 *         if mask_length:
 */
  __pyx_t_1 = __pyx_v_self->_encode_mask_length;
  __pyx_v_mask_length = __pyx_t_1;

  /* "extensions/lib/websocket.pyx":278
 *         cdef int mask_length = self._encode_mask_length
 *         #
 *         if mask_length:             # <<<<<<<<<<<<<<
 *             masking_key = masking_key or os.urandom(4)
 *             assert len(masking_key) == mask_length, "bad masking key"
 */
  __pyx_t_2 = (__pyx_v_mask_length != 0);
  if (__pyx_t_2) {

    /* "extensions/lib/websocket.pyx":279
 *         #
 *         if mask_length:
 *             masking_key = masking_key or os.urandom(4)             # <<<<<<<<<<<<<<
 *             assert len(masking_key) == mask_length, "bad masking key"
 *         else:
 */
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_masking_key); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 279, __pyx_L1_error)
    if (!__pyx_t_2) {
    } else {
      __Pyx_INCREF(__pyx_v_masking_key);
      __pyx_t_3 = __pyx_v_masking_key;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_4 = __Pyx_GetModuleGlobalName(__pyx_n_s_os); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_urandom); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_tuple__26, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(1, 279, __pyx_L1_error)
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_L4_bool_binop_done:;
    __Pyx_DECREF_SET(__pyx_v_masking_key, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "extensions/lib/websocket.pyx":280
 *         if mask_length:
 *             masking_key = masking_key or os.urandom(4)
 *             assert len(masking_key) == mask_length, "bad masking key"             # <<<<<<<<<<<<<<
 *         else:
 *             masking_key = b''
 */
    #ifndef CYTHON_WITHOUT_ASSERTIONS
    if (unlikely(!Py_OptimizeFlag)) {
      if (unlikely(__pyx_v_masking_key == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
        __PYX_ERR(1, 280, __pyx_L1_error)
      }
      __pyx_t_6 = PyBytes_GET_SIZE(__pyx_v_masking_key); if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(1, 280, __pyx_L1_error)
      if (unlikely(!((__pyx_t_6 == __pyx_v_mask_length) != 0))) {
        PyErr_SetObject(PyExc_AssertionError, __pyx_kp_s_bad_masking_key);
        __PYX_ERR(1, 280, __pyx_L1_error)
      }
    }
    #endif

    /* "extensions/lib/websocket.pyx":278
 *         cdef int mask_length = self._encode_mask_length
 *         #
 *         if mask_length:             # <<<<<<<<<<<<<<
 *             masking_key = masking_key or os.urandom(4)
 *             assert len(masking_key) == mask_length, "bad masking key"
 */
    goto __pyx_L3;
  }

  /* "extensions/lib/websocket.pyx":282
 *             assert len(masking_key) == mask_length, "bad masking key"
 *         else:
 *             masking_key = b''             # <<<<<<<<<<<<<<
 *         if opcode is None:
 *             opcode = 1 if isinstance(message, str) else 2
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_kp_b_);
    __Pyx_DECREF_SET(__pyx_v_masking_key, __pyx_kp_b_);
  }
  __pyx_L3:;

  /* "extensions/lib/websocket.pyx":283
 *         else:
 *             masking_key = b''
 *         if opcode is None:             # <<<<<<<<<<<<<<
 *             opcode = 1 if isinstance(message, str) else 2
 *         data = to_bytes(message or b'', 'utf-8')
 */
  __pyx_t_2 = (__pyx_v_opcode == Py_None);
  __pyx_t_7 = (__pyx_t_2 != 0);
  if (__pyx_t_7) {

    /* "extensions/lib/websocket.pyx":284
 *             masking_key = b''
 *         if opcode is None:
 *             opcode = 1 if isinstance(message, str) else 2             # <<<<<<<<<<<<<<
 *         data = to_bytes(message or b'', 'utf-8')
 *         if opcode not in self._opcodes:
 */
    __pyx_t_7 = PyString_Check(__pyx_v_message); 
    if ((__pyx_t_7 != 0)) {
      __Pyx_INCREF(__pyx_int_1);
      __pyx_t_3 = __pyx_int_1;
    } else {
      __Pyx_INCREF(__pyx_int_2);
      __pyx_t_3 = __pyx_int_2;
    }
    __Pyx_DECREF_SET(__pyx_v_opcode, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "extensions/lib/websocket.pyx":283
 *         else:
 *             masking_key = b''
 *         if opcode is None:             # <<<<<<<<<<<<<<
 *             opcode = 1 if isinstance(message, str) else 2
 *         data = to_bytes(message or b'', 'utf-8')
 */
  }

  /* "extensions/lib/websocket.pyx":285
 *         if opcode is None:
 *             opcode = 1 if isinstance(message, str) else 2
 *         data = to_bytes(message or b'', 'utf-8')             # <<<<<<<<<<<<<<
 *         if opcode not in self._opcodes:
 *             raise self.ProtocolError('WEBSOCKET opcode a reserved value')
 */
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_v_message); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(1, 285, __pyx_L1_error)
  if (!__pyx_t_7) {
  } else {
    __Pyx_INCREF(__pyx_v_message);
    __pyx_t_3 = __pyx_v_message;
    goto __pyx_L7_bool_binop_done;
  }
  __Pyx_INCREF(__pyx_kp_b_);
  __pyx_t_3 = __pyx_kp_b_;
  __pyx_L7_bool_binop_done:;
  __pyx_t_4 = __pyx_f_9websocket_to_bytes(__pyx_t_3, __pyx_kp_s_utf_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 285, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_data = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "extensions/lib/websocket.pyx":286
 *             opcode = 1 if isinstance(message, str) else 2
 *         data = to_bytes(message or b'', 'utf-8')
 *         if opcode not in self._opcodes:             # <<<<<<<<<<<<<<
 *             raise self.ProtocolError('WEBSOCKET opcode a reserved value')
 *         elif opcode > 7:
 */
  __pyx_t_7 = (__Pyx_PySequence_ContainsTF(__pyx_v_opcode, __pyx_v_self->_opcodes, Py_NE)); if (unlikely(__pyx_t_7 < 0)) __PYX_ERR(1, 286, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_7 != 0);
  if (__pyx_t_2) {

    /* "extensions/lib/websocket.pyx":287
 *         data = to_bytes(message or b'', 'utf-8')
 *         if opcode not in self._opcodes:
 *             raise self.ProtocolError('WEBSOCKET opcode a reserved value')             # <<<<<<<<<<<<<<
 *         elif opcode > 7:
 *             if len(data) > 125:
 */
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_v_self->ProtocolError, __pyx_tuple__27, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 287, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_Raise(__pyx_t_4, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __PYX_ERR(1, 287, __pyx_L1_error)

    /* "extensions/lib/websocket.pyx":286
 *             opcode = 1 if isinstance(message, str) else 2
 *         data = to_bytes(message or b'', 'utf-8')
 *         if opcode not in self._opcodes:             # <<<<<<<<<<<<<<
 *             raise self.ProtocolError('WEBSOCKET opcode a reserved value')
 *         elif opcode > 7:
 */
  }

  /* "extensions/lib/websocket.pyx":288
 *         if opcode not in self._opcodes:
 *             raise self.ProtocolError('WEBSOCKET opcode a reserved value')
 *         elif opcode > 7:             # <<<<<<<<<<<<<<
 *             if len(data) > 125:
 *                 raise self.ProtocolError('WEBSOCKET control frame too large')
 */
  __pyx_t_4 = PyObject_RichCompare(__pyx_v_opcode, __pyx_int_7, Py_GT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 288, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 288, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_2) {

    /* "extensions/lib/websocket.pyx":289
 *             raise self.ProtocolError('WEBSOCKET opcode a reserved value')
 *         elif opcode > 7:
 *             if len(data) > 125:             # <<<<<<<<<<<<<<
 *                 raise self.ProtocolError('WEBSOCKET control frame too large')
 *             if opcode == 8:
 */
    if (unlikely(__pyx_v_data == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(1, 289, __pyx_L1_error)
    }
    __pyx_t_6 = PyBytes_GET_SIZE(__pyx_v_data); if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(1, 289, __pyx_L1_error)
    __pyx_t_2 = ((__pyx_t_6 > 0x7D) != 0);
    if (__pyx_t_2) {

      /* "extensions/lib/websocket.pyx":290
 *         elif opcode > 7:
 *             if len(data) > 125:
 *                 raise self.ProtocolError('WEBSOCKET control frame too large')             # <<<<<<<<<<<<<<
 *             if opcode == 8:
 *                 #TODO CHECK CLOSE FRAME STATUS CODE
 */
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_v_self->ProtocolError, __pyx_tuple__28, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 290, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_Raise(__pyx_t_4, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __PYX_ERR(1, 290, __pyx_L1_error)

      /* "extensions/lib/websocket.pyx":289
 *             raise self.ProtocolError('WEBSOCKET opcode a reserved value')
 *         elif opcode > 7:
 *             if len(data) > 125:             # <<<<<<<<<<<<<<
 *                 raise self.ProtocolError('WEBSOCKET control frame too large')
 *             if opcode == 8:
 */
    }

    /* "extensions/lib/websocket.pyx":291
 *             if len(data) > 125:
 *                 raise self.ProtocolError('WEBSOCKET control frame too large')
 *             if opcode == 8:             # <<<<<<<<<<<<<<
 *                 #TODO CHECK CLOSE FRAME STATUS CODE
 *                 pass
 */
    __pyx_t_4 = __Pyx_PyInt_EqObjC(__pyx_v_opcode, __pyx_int_8, 8, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 291, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 291, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_2) {
    }

    /* "extensions/lib/websocket.pyx":288
 *         if opcode not in self._opcodes:
 *             raise self.ProtocolError('WEBSOCKET opcode a reserved value')
 *         elif opcode > 7:             # <<<<<<<<<<<<<<
 *             if len(data) > 125:
 *                 raise self.ProtocolError('WEBSOCKET control frame too large')
 */
  }

  /* "extensions/lib/websocket.pyx":294
 *                 #TODO CHECK CLOSE FRAME STATUS CODE
 *                 pass
 *         return opcode, masking_key, data             # <<<<<<<<<<<<<<
 * 
 *     cdef bytes _chunk(self, int length):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 294, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_INCREF(__pyx_v_opcode);
  __Pyx_GIVEREF(__pyx_v_opcode);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_opcode);
  __Pyx_INCREF(__pyx_v_masking_key);
  __Pyx_GIVEREF(__pyx_v_masking_key);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_v_masking_key);
  __Pyx_INCREF(__pyx_v_data);
  __Pyx_GIVEREF(__pyx_v_data);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_data);
  __pyx_r = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;
  goto __pyx_L0;

  /* "extensions/lib/websocket.pyx":275
 *         return bytes(buffer)
 * 
 *     cdef tuple _info(self, message, opcode, bytes masking_key):             # <<<<<<<<<<<<<<
 *         cdef int mask_length = self._encode_mask_length
 *         #
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("lib.FrameParser._info", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_opcode);
  __Pyx_XDECREF(__pyx_v_masking_key);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "extensions/lib/websocket.pyx":296
 *         return opcode, masking_key, data
 * 
 *     cdef bytes _chunk(self, int length):             # <<<<<<<<<<<<<<
 *         cdef bytes chunk = bytes(self.buffer[:length])
 *         self.buffer = self.buffer[length:]
 */

static PyObject *__pyx_f_3lib_11FrameParser__chunk(struct __pyx_obj_3lib_FrameParser *__pyx_v_self, int __pyx_v_length) {
  PyObject *__pyx_v_chunk = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannySetupContext("_chunk", 0);

  /* "extensions/lib/websocket.pyx":297
 * 
 *     cdef bytes _chunk(self, int length):
 *         cdef bytes chunk = bytes(self.buffer[:length])             # <<<<<<<<<<<<<<
 *         self.buffer = self.buffer[length:]
 *         return chunk
 */
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_self->buffer, 0, __pyx_v_length, NULL, NULL, NULL, 0, 1, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)(&PyBytes_Type)), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 297, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_chunk = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "extensions/lib/websocket.pyx":298
 *     cdef bytes _chunk(self, int length):
 *         cdef bytes chunk = bytes(self.buffer[:length])
 *         self.buffer = self.buffer[length:]             # <<<<<<<<<<<<<<
 *         return chunk
 */
  __pyx_t_1 = __Pyx_PyObject_GetSlice(__pyx_v_self->buffer, __pyx_v_length, 0, NULL, NULL, NULL, 1, 0, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 298, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->buffer);
  __Pyx_DECREF(__pyx_v_self->buffer);
  __pyx_v_self->buffer = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "extensions/lib/websocket.pyx":299
 *         cdef bytes chunk = bytes(self.buffer[:length])
 *         self.buffer = self.buffer[length:]
 *         return chunk             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_chunk);
  __pyx_r = __pyx_v_chunk;
  goto __pyx_L0;

  /* "extensions/lib/websocket.pyx":296
 *         return opcode, masking_key, data
 * 
 *     cdef bytes _chunk(self, int length):             # <<<<<<<<<<<<<<
 *         cdef bytes chunk = bytes(self.buffer[:length])
 *         self.buffer = self.buffer[length:]
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("lib.FrameParser._chunk", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_chunk);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "websocket.pxd":7
 * 
 * 
 * cdef inline bytes to_bytes(object value, str encoding):             # <<<<<<<<<<<<<<
 *     if isinstance(value, bytes):
 *         return value
 */

static CYTHON_INLINE PyObject *__pyx_f_9websocket_to_bytes(PyObject *__pyx_v_value, PyObject *__pyx_v_encoding) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannySetupContext("to_bytes", 0);

  /* "websocket.pxd":8
 * 
 * cdef inline bytes to_bytes(object value, str encoding):
 *     if isinstance(value, bytes):             # <<<<<<<<<<<<<<
 *         return value
 *     elif isinstance(value, str):
 */
  __pyx_t_1 = PyBytes_Check(__pyx_v_value); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "websocket.pxd":9
 * cdef inline bytes to_bytes(object value, str encoding):
 *     if isinstance(value, bytes):
 *         return value             # <<<<<<<<<<<<<<
 *     elif isinstance(value, str):
 *         return value.encode(encoding)
 */
    __Pyx_XDECREF(__pyx_r);
    if (!(likely(PyBytes_CheckExact(__pyx_v_value))||((__pyx_v_value) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_v_value)->tp_name), 0))) __PYX_ERR(2, 9, __pyx_L1_error)
    __Pyx_INCREF(__pyx_v_value);
    __pyx_r = ((PyObject*)__pyx_v_value);
    goto __pyx_L0;

    /* "websocket.pxd":8
 * 
 * cdef inline bytes to_bytes(object value, str encoding):
 *     if isinstance(value, bytes):             # <<<<<<<<<<<<<<
 *         return value
 *     elif isinstance(value, str):
 */
  }

  /* "websocket.pxd":10
 *     if isinstance(value, bytes):
 *         return value
 *     elif isinstance(value, str):             # <<<<<<<<<<<<<<
 *         return value.encode(encoding)
 *     else:
 */
  __pyx_t_2 = PyString_Check(__pyx_v_value); 
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "websocket.pxd":11
 *         return value
 *     elif isinstance(value, str):
 *         return value.encode(encoding)             # <<<<<<<<<<<<<<
 *     else:
 *         raise TypeError('Requires bytes or string')
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_value, __pyx_n_s_encode); if (unlikely(!__pyx_t_4)) __PYX_ERR(2, 11, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_COMPILING_IN_CPYTHON && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_v_encoding); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 11, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(2, 11, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
      __Pyx_INCREF(__pyx_v_encoding);
      __Pyx_GIVEREF(__pyx_v_encoding);
      PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_encoding);
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 11, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(2, 11, __pyx_L1_error)
    __pyx_r = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "websocket.pxd":10
 *     if isinstance(value, bytes):
 *         return value
 *     elif isinstance(value, str):             # <<<<<<<<<<<<<<
 *         return value.encode(encoding)
 *     else:
 */
  }

  /* "websocket.pxd":13
 *         return value.encode(encoding)
 *     else:
 *         raise TypeError('Requires bytes or string')             # <<<<<<<<<<<<<<
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__29, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(2, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(2, 13, __pyx_L1_error)
  }

  /* "websocket.pxd":7
 * 
 * 
 * cdef inline bytes to_bytes(object value, str encoding):             # <<<<<<<<<<<<<<
 *     if isinstance(value, bytes):
 *         return value
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("websocket.to_bytes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "lib.pxd":9
 * cdef double clog2 = log(2.)
 * 
 * cdef inline int int_max(int a, int b): return a if a >= b else b             # <<<<<<<<<<<<<<
 * cdef inline int int_min(int a, int b): return a if a >= b else b
 * 
 */

static CYTHON_INLINE int __pyx_f_3lib_int_max(int __pyx_v_a, int __pyx_v_b) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("int_max", 0);
  if (((__pyx_v_a >= __pyx_v_b) != 0)) {
    __pyx_t_1 = __pyx_v_a;
  } else {
    __pyx_t_1 = __pyx_v_b;
  }
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "lib.pxd":10
 * 
 * cdef inline int int_max(int a, int b): return a if a >= b else b
 * cdef inline int int_min(int a, int b): return a if a >= b else b             # <<<<<<<<<<<<<<
 * 
 * 
 */

static CYTHON_INLINE int __pyx_f_3lib_int_min(int __pyx_v_a, int __pyx_v_b) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("int_min", 0);
  if (((__pyx_v_a >= __pyx_v_b) != 0)) {
    __pyx_t_1 = __pyx_v_a;
  } else {
    __pyx_t_1 = __pyx_v_b;
  }
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "lib.pxd":14
 * 
 * # MSVC does not have log2!
 * cdef inline double Log2(double x):             # <<<<<<<<<<<<<<
 *     return log(x) / clog2
 */

static CYTHON_INLINE double __pyx_f_3lib_Log2(double __pyx_v_x) {
  double __pyx_r;
  __Pyx_RefNannyDeclarations
  double __pyx_t_1;
  __Pyx_RefNannySetupContext("Log2", 0);

  /* "lib.pxd":15
 * # MSVC does not have log2!
 * cdef inline double Log2(double x):
 *     return log(x) / clog2             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = log(__pyx_v_x);
  if (unlikely(__pyx_v_3lib_clog2 == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "float division");
    __PYX_ERR(3, 15, __pyx_L1_error)
  }
  __pyx_r = (__pyx_t_1 / __pyx_v_3lib_clog2);
  goto __pyx_L0;

  /* "lib.pxd":14
 * 
 * # MSVC does not have log2!
 * cdef inline double Log2(double x):             # <<<<<<<<<<<<<<
 *     return log(x) / clog2
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_WriteUnraisable("lib.Log2", __pyx_clineno, __pyx_lineno, __pyx_filename, 0, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_3lib_Task __pyx_vtable_3lib_Task;

static PyObject *__pyx_tp_new_3lib_Task(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_3lib_Task *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_3lib_Task *)o);
  p->__pyx_vtab = __pyx_vtabptr_3lib_Task;
  p->_next = ((struct __pyx_obj_3lib_Task *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_3lib_4Task_1__cinit__(o, a, k) < 0)) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_3lib_Task(PyObject *o) {
  struct __pyx_obj_3lib_Task *p = (struct __pyx_obj_3lib_Task *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_next);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_3lib_Task(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_3lib_Task *p = (struct __pyx_obj_3lib_Task *)o;
  if (p->_next) {
    e = (*v)(((PyObject*)p->_next), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_3lib_Task(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_3lib_Task *p = (struct __pyx_obj_3lib_Task *)o;
  tmp = ((PyObject*)p->_next);
  p->_next = ((struct __pyx_obj_3lib_Task *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_3lib_Task[] = {
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_3lib_Task = {
  PyVarObject_HEAD_INIT(0, 0)
  "lib.Task", /*tp_name*/
  sizeof(struct __pyx_obj_3lib_Task), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_3lib_Task, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_3lib_Task, /*tp_traverse*/
  __pyx_tp_clear_3lib_Task, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_3lib_Task, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_3lib_Task, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_3lib_RedisParser __pyx_vtable_3lib_RedisParser;

static PyObject *__pyx_tp_new_3lib_RedisParser(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_3lib_RedisParser *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_3lib_RedisParser *)o);
  p->__pyx_vtab = __pyx_vtabptr_3lib_RedisParser;
  p->_protocolError = Py_None; Py_INCREF(Py_None);
  p->_responseError = Py_None; Py_INCREF(Py_None);
  p->_encoding = Py_None; Py_INCREF(Py_None);
  p->_inbuffer = Py_None; Py_INCREF(Py_None);
  p->_current = ((struct __pyx_obj_3lib_Task *)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_3lib_11RedisParser_1__cinit__(o, a, k) < 0)) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_3lib_RedisParser(PyObject *o) {
  struct __pyx_obj_3lib_RedisParser *p = (struct __pyx_obj_3lib_RedisParser *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_protocolError);
  Py_CLEAR(p->_responseError);
  Py_CLEAR(p->_encoding);
  Py_CLEAR(p->_inbuffer);
  Py_CLEAR(p->_current);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_3lib_RedisParser(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_3lib_RedisParser *p = (struct __pyx_obj_3lib_RedisParser *)o;
  if (p->_protocolError) {
    e = (*v)(p->_protocolError, a); if (e) return e;
  }
  if (p->_responseError) {
    e = (*v)(p->_responseError, a); if (e) return e;
  }
  if (p->_encoding) {
    e = (*v)(p->_encoding, a); if (e) return e;
  }
  if (p->_inbuffer) {
    e = (*v)(p->_inbuffer, a); if (e) return e;
  }
  if (p->_current) {
    e = (*v)(((PyObject*)p->_current), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_3lib_RedisParser(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_3lib_RedisParser *p = (struct __pyx_obj_3lib_RedisParser *)o;
  tmp = ((PyObject*)p->_protocolError);
  p->_protocolError = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_responseError);
  p->_responseError = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_encoding);
  p->_encoding = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_inbuffer);
  p->_inbuffer = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_current);
  p->_current = ((struct __pyx_obj_3lib_Task *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_3lib_RedisParser[] = {
  {"on_connect", (PyCFunction)__pyx_pw_3lib_11RedisParser_3on_connect, METH_O, 0},
  {"on_disconnect", (PyCFunction)__pyx_pw_3lib_11RedisParser_5on_disconnect, METH_NOARGS, 0},
  {"get", (PyCFunction)__pyx_pw_3lib_11RedisParser_7get, METH_NOARGS, 0},
  {"feed", (PyCFunction)__pyx_pw_3lib_11RedisParser_9feed, METH_O, 0},
  {"buffer", (PyCFunction)__pyx_pw_3lib_11RedisParser_11buffer, METH_NOARGS, 0},
  {"pack_command", (PyCFunction)__pyx_pw_3lib_11RedisParser_13pack_command, METH_O, 0},
  {"pack_pipeline", (PyCFunction)__pyx_pw_3lib_11RedisParser_15pack_pipeline, METH_O, 0},
  {"bulk", (PyCFunction)__pyx_pw_3lib_11RedisParser_17bulk, METH_VARARGS|METH_KEYWORDS, 0},
  {"multi_bulk_len", (PyCFunction)__pyx_pw_3lib_11RedisParser_19multi_bulk_len, METH_O, 0},
  {"multi_bulk", (PyCFunction)__pyx_pw_3lib_11RedisParser_21multi_bulk, METH_O, 0},
  {"_pack_command", (PyCFunction)__pyx_pw_3lib_11RedisParser_23_pack_command, METH_O, 0},
  {"_pack", (PyCFunction)__pyx_pw_3lib_11RedisParser_26_pack, METH_O, 0},
  {"_lua_dict", (PyCFunction)__pyx_pw_3lib_11RedisParser_29_lua_dict, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_3lib_RedisParser = {
  PyVarObject_HEAD_INIT(0, 0)
  "lib.RedisParser", /*tp_name*/
  sizeof(struct __pyx_obj_3lib_RedisParser), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_3lib_RedisParser, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_3lib_RedisParser, /*tp_traverse*/
  __pyx_tp_clear_3lib_RedisParser, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_3lib_RedisParser, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_3lib_RedisParser, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_3lib_ArrayTask __pyx_vtable_3lib_ArrayTask;

static PyObject *__pyx_tp_new_3lib_ArrayTask(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_3lib_ArrayTask *p;
  PyObject *o = __pyx_tp_new_3lib_Task(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_3lib_ArrayTask *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_3lib_Task*)__pyx_vtabptr_3lib_ArrayTask;
  p->_response = ((PyObject*)Py_None); Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_3lib_ArrayTask(PyObject *o) {
  struct __pyx_obj_3lib_ArrayTask *p = (struct __pyx_obj_3lib_ArrayTask *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_response);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_3lib_Task(o);
}

static int __pyx_tp_traverse_3lib_ArrayTask(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_3lib_ArrayTask *p = (struct __pyx_obj_3lib_ArrayTask *)o;
  e = __pyx_tp_traverse_3lib_Task(o, v, a); if (e) return e;
  if (p->_response) {
    e = (*v)(p->_response, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_3lib_ArrayTask(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_3lib_ArrayTask *p = (struct __pyx_obj_3lib_ArrayTask *)o;
  __pyx_tp_clear_3lib_Task(o);
  tmp = ((PyObject*)p->_response);
  p->_response = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_3lib_ArrayTask = {
  PyVarObject_HEAD_INIT(0, 0)
  "lib.ArrayTask", /*tp_name*/
  sizeof(struct __pyx_obj_3lib_ArrayTask), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_3lib_ArrayTask, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_3lib_ArrayTask, /*tp_traverse*/
  __pyx_tp_clear_3lib_ArrayTask, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_3lib_ArrayTask, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_3lib_Frame(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_3lib_Frame *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_3lib_Frame *)o);
  p->_masking_key = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_body = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_3lib_5Frame_1__cinit__(o, a, k) < 0)) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_3lib_Frame(PyObject *o) {
  struct __pyx_obj_3lib_Frame *p = (struct __pyx_obj_3lib_Frame *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_masking_key);
  Py_CLEAR(p->_body);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_3lib_Frame(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_3lib_Frame *p = (struct __pyx_obj_3lib_Frame *)o;
  if (p->_body) {
    e = (*v)(p->_body, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_3lib_Frame(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_3lib_Frame *p = (struct __pyx_obj_3lib_Frame *)o;
  tmp = ((PyObject*)p->_body);
  p->_body = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_3lib_5Frame_final(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_3lib_5Frame_5final_1__get__(o);
}

static PyObject *__pyx_getprop_3lib_5Frame_opcode(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_3lib_5Frame_6opcode_1__get__(o);
}

static PyObject *__pyx_getprop_3lib_5Frame_body(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_3lib_5Frame_4body_1__get__(o);
}

static PyObject *__pyx_getprop_3lib_5Frame_masking_key(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_3lib_5Frame_11masking_key_1__get__(o);
}

static PyObject *__pyx_getprop_3lib_5Frame_is_message(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_3lib_5Frame_10is_message_1__get__(o);
}

static PyObject *__pyx_getprop_3lib_5Frame_is_bytes(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_3lib_5Frame_8is_bytes_1__get__(o);
}

static PyObject *__pyx_getprop_3lib_5Frame_is_close(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_3lib_5Frame_8is_close_1__get__(o);
}

static PyObject *__pyx_getprop_3lib_5Frame_is_ping(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_3lib_5Frame_7is_ping_1__get__(o);
}

static PyObject *__pyx_getprop_3lib_5Frame_is_pong(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_3lib_5Frame_7is_pong_1__get__(o);
}

static PyMethodDef __pyx_methods_3lib_Frame[] = {
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_3lib_Frame[] = {
  {(char *)"final", __pyx_getprop_3lib_5Frame_final, 0, (char *)0, 0},
  {(char *)"opcode", __pyx_getprop_3lib_5Frame_opcode, 0, (char *)0, 0},
  {(char *)"body", __pyx_getprop_3lib_5Frame_body, 0, (char *)0, 0},
  {(char *)"masking_key", __pyx_getprop_3lib_5Frame_masking_key, 0, (char *)0, 0},
  {(char *)"is_message", __pyx_getprop_3lib_5Frame_is_message, 0, (char *)0, 0},
  {(char *)"is_bytes", __pyx_getprop_3lib_5Frame_is_bytes, 0, (char *)0, 0},
  {(char *)"is_close", __pyx_getprop_3lib_5Frame_is_close, 0, (char *)0, 0},
  {(char *)"is_ping", __pyx_getprop_3lib_5Frame_is_ping, 0, (char *)0, 0},
  {(char *)"is_pong", __pyx_getprop_3lib_5Frame_is_pong, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_3lib_Frame = {
  PyVarObject_HEAD_INIT(0, 0)
  "lib.Frame", /*tp_name*/
  sizeof(struct __pyx_obj_3lib_Frame), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_3lib_Frame, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_3lib_Frame, /*tp_traverse*/
  __pyx_tp_clear_3lib_Frame, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_3lib_Frame, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_3lib_Frame, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_3lib_Frame, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_3lib_FrameParser __pyx_vtable_3lib_FrameParser;

static PyObject *__pyx_tp_new_3lib_FrameParser(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_3lib_FrameParser *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_3lib_FrameParser *)o);
  p->__pyx_vtab = __pyx_vtabptr_3lib_FrameParser;
  p->frame = ((struct __pyx_obj_3lib_Frame *)Py_None); Py_INCREF(Py_None);
  p->buffer = Py_None; Py_INCREF(Py_None);
  p->ProtocolError = Py_None; Py_INCREF(Py_None);
  p->_opcodes = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_extensions = Py_None; Py_INCREF(Py_None);
  p->_protocols = Py_None; Py_INCREF(Py_None);
  p->_close_codes = Py_None; Py_INCREF(Py_None);
  if (unlikely(__pyx_pw_3lib_11FrameParser_1__cinit__(o, a, k) < 0)) {
    Py_DECREF(o); o = 0;
  }
  return o;
}

static void __pyx_tp_dealloc_3lib_FrameParser(PyObject *o) {
  struct __pyx_obj_3lib_FrameParser *p = (struct __pyx_obj_3lib_FrameParser *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->frame);
  Py_CLEAR(p->buffer);
  Py_CLEAR(p->ProtocolError);
  Py_CLEAR(p->_opcodes);
  Py_CLEAR(p->_extensions);
  Py_CLEAR(p->_protocols);
  Py_CLEAR(p->_close_codes);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_3lib_FrameParser(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_3lib_FrameParser *p = (struct __pyx_obj_3lib_FrameParser *)o;
  if (p->frame) {
    e = (*v)(((PyObject*)p->frame), a); if (e) return e;
  }
  if (p->buffer) {
    e = (*v)(p->buffer, a); if (e) return e;
  }
  if (p->ProtocolError) {
    e = (*v)(p->ProtocolError, a); if (e) return e;
  }
  if (p->_opcodes) {
    e = (*v)(p->_opcodes, a); if (e) return e;
  }
  if (p->_extensions) {
    e = (*v)(p->_extensions, a); if (e) return e;
  }
  if (p->_protocols) {
    e = (*v)(p->_protocols, a); if (e) return e;
  }
  if (p->_close_codes) {
    e = (*v)(p->_close_codes, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_3lib_FrameParser(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_3lib_FrameParser *p = (struct __pyx_obj_3lib_FrameParser *)o;
  tmp = ((PyObject*)p->frame);
  p->frame = ((struct __pyx_obj_3lib_Frame *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->buffer);
  p->buffer = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->ProtocolError);
  p->ProtocolError = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_opcodes);
  p->_opcodes = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_extensions);
  p->_extensions = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_protocols);
  p->_protocols = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_close_codes);
  p->_close_codes = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_3lib_11FrameParser_version(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_3lib_11FrameParser_7version_1__get__(o);
}

static PyObject *__pyx_getprop_3lib_11FrameParser_max_payload(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_3lib_11FrameParser_11max_payload_1__get__(o);
}

static PyObject *__pyx_getprop_3lib_11FrameParser_decode_mask_length(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_3lib_11FrameParser_18decode_mask_length_1__get__(o);
}

static PyObject *__pyx_getprop_3lib_11FrameParser_encode_mask_length(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_3lib_11FrameParser_18encode_mask_length_1__get__(o);
}

static PyObject *__pyx_getprop_3lib_11FrameParser_extensions(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_3lib_11FrameParser_10extensions_1__get__(o);
}

static PyObject *__pyx_getprop_3lib_11FrameParser_protocols(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_3lib_11FrameParser_9protocols_1__get__(o);
}

static PyMethodDef __pyx_methods_3lib_FrameParser[] = {
  {"ping", (PyCFunction)__pyx_pw_3lib_11FrameParser_3ping, METH_VARARGS|METH_KEYWORDS, __pyx_doc_3lib_11FrameParser_2ping},
  {"pong", (PyCFunction)__pyx_pw_3lib_11FrameParser_5pong, METH_VARARGS|METH_KEYWORDS, __pyx_doc_3lib_11FrameParser_4pong},
  {"close", (PyCFunction)__pyx_pw_3lib_11FrameParser_7close, METH_VARARGS|METH_KEYWORDS, __pyx_doc_3lib_11FrameParser_6close},
  {"continuation", (PyCFunction)__pyx_pw_3lib_11FrameParser_9continuation, METH_VARARGS|METH_KEYWORDS, __pyx_doc_3lib_11FrameParser_8continuation},
  {"encode", (PyCFunction)__pyx_pw_3lib_11FrameParser_11encode, METH_VARARGS|METH_KEYWORDS, __pyx_doc_3lib_11FrameParser_10encode},
  {"multi_encode", (PyCFunction)__pyx_pw_3lib_11FrameParser_13multi_encode, METH_VARARGS|METH_KEYWORDS, __pyx_doc_3lib_11FrameParser_12multi_encode},
  {"decode", (PyCFunction)__pyx_pw_3lib_11FrameParser_16decode, METH_VARARGS|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_3lib_FrameParser[] = {
  {(char *)"version", __pyx_getprop_3lib_11FrameParser_version, 0, (char *)0, 0},
  {(char *)"max_payload", __pyx_getprop_3lib_11FrameParser_max_payload, 0, (char *)0, 0},
  {(char *)"decode_mask_length", __pyx_getprop_3lib_11FrameParser_decode_mask_length, 0, (char *)0, 0},
  {(char *)"encode_mask_length", __pyx_getprop_3lib_11FrameParser_encode_mask_length, 0, (char *)0, 0},
  {(char *)"extensions", __pyx_getprop_3lib_11FrameParser_extensions, 0, (char *)0, 0},
  {(char *)"protocols", __pyx_getprop_3lib_11FrameParser_protocols, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_3lib_FrameParser = {
  PyVarObject_HEAD_INIT(0, 0)
  "lib.FrameParser", /*tp_name*/
  sizeof(struct __pyx_obj_3lib_FrameParser), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_3lib_FrameParser, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Encoder and decoder of WebSocket frames.\n\n    * kind = 0, server (decode masked messages, encode unmasked messages)\n    * kind = 1, client (decode unmasked messages, encode masked messages)\n    * kind = 2, unmasked both encoding and decoding\n    * kind = 3, masked both encoding and decoding\n    ", /*tp_doc*/
  __pyx_tp_traverse_3lib_FrameParser, /*tp_traverse*/
  __pyx_tp_clear_3lib_FrameParser, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_3lib_FrameParser, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_3lib_FrameParser, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_3lib_FrameParser, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_3lib___pyx_scope_struct__pack_pipeline *__pyx_freelist_3lib___pyx_scope_struct__pack_pipeline[8];
static int __pyx_freecount_3lib___pyx_scope_struct__pack_pipeline = 0;

static PyObject *__pyx_tp_new_3lib___pyx_scope_struct__pack_pipeline(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_3lib___pyx_scope_struct__pack_pipeline > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_3lib___pyx_scope_struct__pack_pipeline)))) {
    o = (PyObject*)__pyx_freelist_3lib___pyx_scope_struct__pack_pipeline[--__pyx_freecount_3lib___pyx_scope_struct__pack_pipeline];
    memset(o, 0, sizeof(struct __pyx_obj_3lib___pyx_scope_struct__pack_pipeline));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_3lib___pyx_scope_struct__pack_pipeline(PyObject *o) {
  struct __pyx_obj_3lib___pyx_scope_struct__pack_pipeline *p = (struct __pyx_obj_3lib___pyx_scope_struct__pack_pipeline *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_commands);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_3lib___pyx_scope_struct__pack_pipeline < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_3lib___pyx_scope_struct__pack_pipeline)))) {
    __pyx_freelist_3lib___pyx_scope_struct__pack_pipeline[__pyx_freecount_3lib___pyx_scope_struct__pack_pipeline++] = ((struct __pyx_obj_3lib___pyx_scope_struct__pack_pipeline *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_3lib___pyx_scope_struct__pack_pipeline(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_3lib___pyx_scope_struct__pack_pipeline *p = (struct __pyx_obj_3lib___pyx_scope_struct__pack_pipeline *)o;
  if (p->__pyx_v_commands) {
    e = (*v)(p->__pyx_v_commands, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject*)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_3lib___pyx_scope_struct__pack_pipeline(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_3lib___pyx_scope_struct__pack_pipeline *p = (struct __pyx_obj_3lib___pyx_scope_struct__pack_pipeline *)o;
  tmp = ((PyObject*)p->__pyx_v_commands);
  p->__pyx_v_commands = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_3lib_RedisParser *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_3lib___pyx_scope_struct__pack_pipeline = {
  PyVarObject_HEAD_INIT(0, 0)
  "lib.__pyx_scope_struct__pack_pipeline", /*tp_name*/
  sizeof(struct __pyx_obj_3lib___pyx_scope_struct__pack_pipeline), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_3lib___pyx_scope_struct__pack_pipeline, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_3lib___pyx_scope_struct__pack_pipeline, /*tp_traverse*/
  __pyx_tp_clear_3lib___pyx_scope_struct__pack_pipeline, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_3lib___pyx_scope_struct__pack_pipeline, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_3lib___pyx_scope_struct_1_genexpr *__pyx_freelist_3lib___pyx_scope_struct_1_genexpr[8];
static int __pyx_freecount_3lib___pyx_scope_struct_1_genexpr = 0;

static PyObject *__pyx_tp_new_3lib___pyx_scope_struct_1_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_3lib___pyx_scope_struct_1_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_3lib___pyx_scope_struct_1_genexpr)))) {
    o = (PyObject*)__pyx_freelist_3lib___pyx_scope_struct_1_genexpr[--__pyx_freecount_3lib___pyx_scope_struct_1_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_3lib___pyx_scope_struct_1_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_3lib___pyx_scope_struct_1_genexpr(PyObject *o) {
  struct __pyx_obj_3lib___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_3lib___pyx_scope_struct_1_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v__);
  Py_CLEAR(p->__pyx_v_args);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_3lib___pyx_scope_struct_1_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_3lib___pyx_scope_struct_1_genexpr)))) {
    __pyx_freelist_3lib___pyx_scope_struct_1_genexpr[__pyx_freecount_3lib___pyx_scope_struct_1_genexpr++] = ((struct __pyx_obj_3lib___pyx_scope_struct_1_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_3lib___pyx_scope_struct_1_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_3lib___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_3lib___pyx_scope_struct_1_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject*)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v__) {
    e = (*v)(p->__pyx_v__, a); if (e) return e;
  }
  if (p->__pyx_v_args) {
    e = (*v)(p->__pyx_v_args, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_3lib___pyx_scope_struct_1_genexpr(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_3lib___pyx_scope_struct_1_genexpr *p = (struct __pyx_obj_3lib___pyx_scope_struct_1_genexpr *)o;
  tmp = ((PyObject*)p->__pyx_outer_scope);
  p->__pyx_outer_scope = ((struct __pyx_obj_3lib___pyx_scope_struct__pack_pipeline *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v__);
  p->__pyx_v__ = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_args);
  p->__pyx_v_args = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_t_0);
  p->__pyx_t_0 = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_3lib___pyx_scope_struct_1_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "lib.__pyx_scope_struct_1_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_3lib___pyx_scope_struct_1_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_3lib___pyx_scope_struct_1_genexpr, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_3lib___pyx_scope_struct_1_genexpr, /*tp_traverse*/
  __pyx_tp_clear_3lib___pyx_scope_struct_1_genexpr, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_3lib___pyx_scope_struct_1_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_3lib___pyx_scope_struct_2__pack_command *__pyx_freelist_3lib___pyx_scope_struct_2__pack_command[8];
static int __pyx_freecount_3lib___pyx_scope_struct_2__pack_command = 0;

static PyObject *__pyx_tp_new_3lib___pyx_scope_struct_2__pack_command(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_3lib___pyx_scope_struct_2__pack_command > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_3lib___pyx_scope_struct_2__pack_command)))) {
    o = (PyObject*)__pyx_freelist_3lib___pyx_scope_struct_2__pack_command[--__pyx_freecount_3lib___pyx_scope_struct_2__pack_command];
    memset(o, 0, sizeof(struct __pyx_obj_3lib___pyx_scope_struct_2__pack_command));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_3lib___pyx_scope_struct_2__pack_command(PyObject *o) {
  struct __pyx_obj_3lib___pyx_scope_struct_2__pack_command *p = (struct __pyx_obj_3lib___pyx_scope_struct_2__pack_command *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_args);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_value);
  Py_CLEAR(p->__pyx_t_1);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_3lib___pyx_scope_struct_2__pack_command < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_3lib___pyx_scope_struct_2__pack_command)))) {
    __pyx_freelist_3lib___pyx_scope_struct_2__pack_command[__pyx_freecount_3lib___pyx_scope_struct_2__pack_command++] = ((struct __pyx_obj_3lib___pyx_scope_struct_2__pack_command *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_3lib___pyx_scope_struct_2__pack_command(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_3lib___pyx_scope_struct_2__pack_command *p = (struct __pyx_obj_3lib___pyx_scope_struct_2__pack_command *)o;
  if (p->__pyx_v_args) {
    e = (*v)(p->__pyx_v_args, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject*)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_value) {
    e = (*v)(p->__pyx_v_value, a); if (e) return e;
  }
  if (p->__pyx_t_1) {
    e = (*v)(p->__pyx_t_1, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_3lib___pyx_scope_struct_2__pack_command(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_3lib___pyx_scope_struct_2__pack_command *p = (struct __pyx_obj_3lib___pyx_scope_struct_2__pack_command *)o;
  tmp = ((PyObject*)p->__pyx_v_args);
  p->__pyx_v_args = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_3lib_RedisParser *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_value);
  p->__pyx_v_value = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_t_1);
  p->__pyx_t_1 = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_3lib___pyx_scope_struct_2__pack_command = {
  PyVarObject_HEAD_INIT(0, 0)
  "lib.__pyx_scope_struct_2__pack_command", /*tp_name*/
  sizeof(struct __pyx_obj_3lib___pyx_scope_struct_2__pack_command), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_3lib___pyx_scope_struct_2__pack_command, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_3lib___pyx_scope_struct_2__pack_command, /*tp_traverse*/
  __pyx_tp_clear_3lib___pyx_scope_struct_2__pack_command, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_3lib___pyx_scope_struct_2__pack_command, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_3lib___pyx_scope_struct_3__pack *__pyx_freelist_3lib___pyx_scope_struct_3__pack[8];
static int __pyx_freecount_3lib___pyx_scope_struct_3__pack = 0;

static PyObject *__pyx_tp_new_3lib___pyx_scope_struct_3__pack(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_3lib___pyx_scope_struct_3__pack > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_3lib___pyx_scope_struct_3__pack)))) {
    o = (PyObject*)__pyx_freelist_3lib___pyx_scope_struct_3__pack[--__pyx_freecount_3lib___pyx_scope_struct_3__pack];
    memset(o, 0, sizeof(struct __pyx_obj_3lib___pyx_scope_struct_3__pack));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_3lib___pyx_scope_struct_3__pack(PyObject *o) {
  struct __pyx_obj_3lib___pyx_scope_struct_3__pack *p = (struct __pyx_obj_3lib___pyx_scope_struct_3__pack *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_args);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_value);
  Py_CLEAR(p->__pyx_t_1);
  Py_CLEAR(p->__pyx_t_3);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_3lib___pyx_scope_struct_3__pack < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_3lib___pyx_scope_struct_3__pack)))) {
    __pyx_freelist_3lib___pyx_scope_struct_3__pack[__pyx_freecount_3lib___pyx_scope_struct_3__pack++] = ((struct __pyx_obj_3lib___pyx_scope_struct_3__pack *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_3lib___pyx_scope_struct_3__pack(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_3lib___pyx_scope_struct_3__pack *p = (struct __pyx_obj_3lib___pyx_scope_struct_3__pack *)o;
  if (p->__pyx_v_args) {
    e = (*v)(p->__pyx_v_args, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject*)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_value) {
    e = (*v)(p->__pyx_v_value, a); if (e) return e;
  }
  if (p->__pyx_t_1) {
    e = (*v)(p->__pyx_t_1, a); if (e) return e;
  }
  if (p->__pyx_t_3) {
    e = (*v)(p->__pyx_t_3, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_3lib___pyx_scope_struct_3__pack(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_3lib___pyx_scope_struct_3__pack *p = (struct __pyx_obj_3lib___pyx_scope_struct_3__pack *)o;
  tmp = ((PyObject*)p->__pyx_v_args);
  p->__pyx_v_args = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_3lib_RedisParser *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_value);
  p->__pyx_v_value = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_t_1);
  p->__pyx_t_1 = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_t_3);
  p->__pyx_t_3 = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_3lib___pyx_scope_struct_3__pack = {
  PyVarObject_HEAD_INIT(0, 0)
  "lib.__pyx_scope_struct_3__pack", /*tp_name*/
  sizeof(struct __pyx_obj_3lib___pyx_scope_struct_3__pack), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_3lib___pyx_scope_struct_3__pack, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_3lib___pyx_scope_struct_3__pack, /*tp_traverse*/
  __pyx_tp_clear_3lib___pyx_scope_struct_3__pack, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_3lib___pyx_scope_struct_3__pack, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_3lib___pyx_scope_struct_4__lua_dict *__pyx_freelist_3lib___pyx_scope_struct_4__lua_dict[8];
static int __pyx_freecount_3lib___pyx_scope_struct_4__lua_dict = 0;

static PyObject *__pyx_tp_new_3lib___pyx_scope_struct_4__lua_dict(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_3lib___pyx_scope_struct_4__lua_dict > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_3lib___pyx_scope_struct_4__lua_dict)))) {
    o = (PyObject*)__pyx_freelist_3lib___pyx_scope_struct_4__lua_dict[--__pyx_freecount_3lib___pyx_scope_struct_4__lua_dict];
    memset(o, 0, sizeof(struct __pyx_obj_3lib___pyx_scope_struct_4__lua_dict));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_3lib___pyx_scope_struct_4__lua_dict(PyObject *o) {
  struct __pyx_obj_3lib___pyx_scope_struct_4__lua_dict *p = (struct __pyx_obj_3lib___pyx_scope_struct_4__lua_dict *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_d);
  Py_CLEAR(p->__pyx_v_index);
  Py_CLEAR(p->__pyx_v_self);
  Py_CLEAR(p->__pyx_v_v);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_3lib___pyx_scope_struct_4__lua_dict < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_3lib___pyx_scope_struct_4__lua_dict)))) {
    __pyx_freelist_3lib___pyx_scope_struct_4__lua_dict[__pyx_freecount_3lib___pyx_scope_struct_4__lua_dict++] = ((struct __pyx_obj_3lib___pyx_scope_struct_4__lua_dict *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_3lib___pyx_scope_struct_4__lua_dict(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_3lib___pyx_scope_struct_4__lua_dict *p = (struct __pyx_obj_3lib___pyx_scope_struct_4__lua_dict *)o;
  if (p->__pyx_v_d) {
    e = (*v)(p->__pyx_v_d, a); if (e) return e;
  }
  if (p->__pyx_v_index) {
    e = (*v)(p->__pyx_v_index, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject*)p->__pyx_v_self), a); if (e) return e;
  }
  if (p->__pyx_v_v) {
    e = (*v)(p->__pyx_v_v, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_3lib___pyx_scope_struct_4__lua_dict(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_3lib___pyx_scope_struct_4__lua_dict *p = (struct __pyx_obj_3lib___pyx_scope_struct_4__lua_dict *)o;
  tmp = ((PyObject*)p->__pyx_v_d);
  p->__pyx_v_d = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_index);
  p->__pyx_v_index = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_3lib_RedisParser *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_v);
  p->__pyx_v_v = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_3lib___pyx_scope_struct_4__lua_dict = {
  PyVarObject_HEAD_INIT(0, 0)
  "lib.__pyx_scope_struct_4__lua_dict", /*tp_name*/
  sizeof(struct __pyx_obj_3lib___pyx_scope_struct_4__lua_dict), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_3lib___pyx_scope_struct_4__lua_dict, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_3lib___pyx_scope_struct_4__lua_dict, /*tp_traverse*/
  __pyx_tp_clear_3lib___pyx_scope_struct_4__lua_dict, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_3lib___pyx_scope_struct_4__lua_dict, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_3lib___pyx_scope_struct_5_multi_encode *__pyx_freelist_3lib___pyx_scope_struct_5_multi_encode[8];
static int __pyx_freecount_3lib___pyx_scope_struct_5_multi_encode = 0;

static PyObject *__pyx_tp_new_3lib___pyx_scope_struct_5_multi_encode(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_3lib___pyx_scope_struct_5_multi_encode > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_3lib___pyx_scope_struct_5_multi_encode)))) {
    o = (PyObject*)__pyx_freelist_3lib___pyx_scope_struct_5_multi_encode[--__pyx_freecount_3lib___pyx_scope_struct_5_multi_encode];
    memset(o, 0, sizeof(struct __pyx_obj_3lib___pyx_scope_struct_5_multi_encode));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_3lib___pyx_scope_struct_5_multi_encode(PyObject *o) {
  struct __pyx_obj_3lib___pyx_scope_struct_5_multi_encode *p = (struct __pyx_obj_3lib___pyx_scope_struct_5_multi_encode *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_chunk);
  Py_CLEAR(p->__pyx_v_data);
  Py_CLEAR(p->__pyx_v_masking_key);
  Py_CLEAR(p->__pyx_v_message);
  Py_CLEAR(p->__pyx_v_opcode);
  Py_CLEAR(p->__pyx_v_self);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_3lib___pyx_scope_struct_5_multi_encode < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_3lib___pyx_scope_struct_5_multi_encode)))) {
    __pyx_freelist_3lib___pyx_scope_struct_5_multi_encode[__pyx_freecount_3lib___pyx_scope_struct_5_multi_encode++] = ((struct __pyx_obj_3lib___pyx_scope_struct_5_multi_encode *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_3lib___pyx_scope_struct_5_multi_encode(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_3lib___pyx_scope_struct_5_multi_encode *p = (struct __pyx_obj_3lib___pyx_scope_struct_5_multi_encode *)o;
  if (p->__pyx_v_message) {
    e = (*v)(p->__pyx_v_message, a); if (e) return e;
  }
  if (p->__pyx_v_opcode) {
    e = (*v)(p->__pyx_v_opcode, a); if (e) return e;
  }
  if (p->__pyx_v_self) {
    e = (*v)(((PyObject*)p->__pyx_v_self), a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_3lib___pyx_scope_struct_5_multi_encode(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_3lib___pyx_scope_struct_5_multi_encode *p = (struct __pyx_obj_3lib___pyx_scope_struct_5_multi_encode *)o;
  tmp = ((PyObject*)p->__pyx_v_message);
  p->__pyx_v_message = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_opcode);
  p->__pyx_v_opcode = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_self);
  p->__pyx_v_self = ((struct __pyx_obj_3lib_FrameParser *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_3lib___pyx_scope_struct_5_multi_encode = {
  PyVarObject_HEAD_INIT(0, 0)
  "lib.__pyx_scope_struct_5_multi_encode", /*tp_name*/
  sizeof(struct __pyx_obj_3lib___pyx_scope_struct_5_multi_encode), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_3lib___pyx_scope_struct_5_multi_encode, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_3lib___pyx_scope_struct_5_multi_encode, /*tp_traverse*/
  __pyx_tp_clear_3lib___pyx_scope_struct_5_multi_encode, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_3lib___pyx_scope_struct_5_multi_encode, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
static struct PyModuleDef __pyx_moduledef = {
  #if PY_VERSION_HEX < 0x03020000
    { PyObject_HEAD_INIT(NULL) NULL, 0, NULL },
  #else
    PyModuleDef_HEAD_INIT,
  #endif
    "lib",
    0, /* m_doc */
    -1, /* m_size */
    __pyx_methods /* m_methods */,
    NULL, /* m_reload */
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_b_, __pyx_k_, sizeof(__pyx_k_), 0, 0, 0, 0},
  {&__pyx_kp_s_, __pyx_k_, sizeof(__pyx_k_), 0, 0, 1, 0},
  {&__pyx_kp_b_1, __pyx_k_1, sizeof(__pyx_k_1), 0, 0, 0, 0},
  {&__pyx_kp_b_1_2, __pyx_k_1_2, sizeof(__pyx_k_1_2), 0, 0, 0, 0},
  {&__pyx_n_s_BB, __pyx_k_BB, sizeof(__pyx_k_BB), 0, 0, 1, 1},
  {&__pyx_kp_s_FIN_must_be_0_or_1, __pyx_k_FIN_must_be_0_or_1, sizeof(__pyx_k_FIN_must_be_0_or_1), 0, 0, 1, 0},
  {&__pyx_n_s_FrameParser_multi_encode, __pyx_k_FrameParser_multi_encode, sizeof(__pyx_k_FrameParser_multi_encode), 0, 0, 1, 1},
  {&__pyx_kp_s_H, __pyx_k_H, sizeof(__pyx_k_H), 0, 0, 1, 0},
  {&__pyx_n_s_ProtocolError, __pyx_k_ProtocolError, sizeof(__pyx_k_ProtocolError), 0, 0, 1, 1},
  {&__pyx_kp_s_Protocol_Error, __pyx_k_Protocol_Error, sizeof(__pyx_k_Protocol_Error), 0, 0, 1, 0},
  {&__pyx_kp_s_Q, __pyx_k_Q, sizeof(__pyx_k_Q), 0, 0, 1, 0},
  {&__pyx_n_s_RedisParser__lua_dict, __pyx_k_RedisParser__lua_dict, sizeof(__pyx_k_RedisParser__lua_dict), 0, 0, 1, 1},
  {&__pyx_n_s_RedisParser__pack, __pyx_k_RedisParser__pack, sizeof(__pyx_k_RedisParser__pack), 0, 0, 1, 1},
  {&__pyx_n_s_RedisParser__pack_command, __pyx_k_RedisParser__pack_command, sizeof(__pyx_k_RedisParser__pack_command), 0, 0, 1, 1},
  {&__pyx_kp_s_Requires_bytes_or_string, __pyx_k_Requires_bytes_or_string, sizeof(__pyx_k_Requires_bytes_or_string), 0, 0, 1, 0},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_kp_s_WEBSOCKET_control_frame_fragment, __pyx_k_WEBSOCKET_control_frame_fragment, sizeof(__pyx_k_WEBSOCKET_control_frame_fragment), 0, 0, 1, 0},
  {&__pyx_kp_s_WEBSOCKET_control_frame_too_larg, __pyx_k_WEBSOCKET_control_frame_too_larg, sizeof(__pyx_k_WEBSOCKET_control_frame_too_larg), 0, 0, 1, 0},
  {&__pyx_kp_s_WEBSOCKET_frame_too_large, __pyx_k_WEBSOCKET_frame_too_large, sizeof(__pyx_k_WEBSOCKET_frame_too_large), 0, 0, 1, 0},
  {&__pyx_kp_s_WEBSOCKET_opcode_a_reserved_valu, __pyx_k_WEBSOCKET_opcode_a_reserved_valu, sizeof(__pyx_k_WEBSOCKET_opcode_a_reserved_valu), 0, 0, 1, 0},
  {&__pyx_kp_b__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 0, 0, 0},
  {&__pyx_kp_b__30, __pyx_k__30, sizeof(__pyx_k__30), 0, 0, 0, 0},
  {&__pyx_kp_b__31, __pyx_k__31, sizeof(__pyx_k__31), 0, 0, 0, 0},
  {&__pyx_kp_b__32, __pyx_k__32, sizeof(__pyx_k__32), 0, 0, 0, 0},
  {&__pyx_kp_b__33, __pyx_k__33, sizeof(__pyx_k__33), 0, 0, 0, 0},
  {&__pyx_kp_b__34, __pyx_k__34, sizeof(__pyx_k__34), 0, 0, 0, 0},
  {&__pyx_n_s_append, __pyx_k_append, sizeof(__pyx_k_append), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_kp_s_bad_masking_key, __pyx_k_bad_masking_key, sizeof(__pyx_k_bad_masking_key), 0, 0, 1, 0},
  {&__pyx_n_s_body, __pyx_k_body, sizeof(__pyx_k_body), 0, 0, 1, 1},
  {&__pyx_n_s_clog2, __pyx_k_clog2, sizeof(__pyx_k_clog2), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_close_codes, __pyx_k_close_codes, sizeof(__pyx_k_close_codes), 0, 0, 1, 1},
  {&__pyx_n_s_code, __pyx_k_code, sizeof(__pyx_k_code), 0, 0, 1, 1},
  {&__pyx_kp_s_d, __pyx_k_d, sizeof(__pyx_k_d), 0, 0, 1, 0},
  {&__pyx_kp_s_d_2, __pyx_k_d_2, sizeof(__pyx_k_d_2), 0, 0, 1, 0},
  {&__pyx_n_s_data, __pyx_k_data, sizeof(__pyx_k_data), 0, 0, 1, 1},
  {&__pyx_n_s_decode, __pyx_k_decode, sizeof(__pyx_k_decode), 0, 0, 1, 1},
  {&__pyx_n_s_decode_responses, __pyx_k_decode_responses, sizeof(__pyx_k_decode_responses), 0, 0, 1, 1},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_encoding, __pyx_k_encoding, sizeof(__pyx_k_encoding), 0, 0, 1, 1},
  {&__pyx_n_s_extend, __pyx_k_extend, sizeof(__pyx_k_extend), 0, 0, 1, 1},
  {&__pyx_n_s_extensions, __pyx_k_extensions, sizeof(__pyx_k_extensions), 0, 0, 1, 1},
  {&__pyx_n_s_final, __pyx_k_final, sizeof(__pyx_k_final), 0, 0, 1, 1},
  {&__pyx_n_s_find, __pyx_k_find, sizeof(__pyx_k_find), 0, 0, 1, 1},
  {&__pyx_n_s_genexpr, __pyx_k_genexpr, sizeof(__pyx_k_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
  {&__pyx_n_s_itertools, __pyx_k_itertools, sizeof(__pyx_k_itertools), 0, 0, 1, 1},
  {&__pyx_n_s_join, __pyx_k_join, sizeof(__pyx_k_join), 0, 0, 1, 1},
  {&__pyx_n_s_kind, __pyx_k_kind, sizeof(__pyx_k_kind), 0, 0, 1, 1},
  {&__pyx_n_s_len, __pyx_k_len, sizeof(__pyx_k_len), 0, 0, 1, 1},
  {&__pyx_n_s_length, __pyx_k_length, sizeof(__pyx_k_length), 0, 0, 1, 1},
  {&__pyx_n_s_lib, __pyx_k_lib, sizeof(__pyx_k_lib), 0, 0, 1, 1},
  {&__pyx_n_s_lua_dict, __pyx_k_lua_dict, sizeof(__pyx_k_lua_dict), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_kp_s_masked_server_frame, __pyx_k_masked_server_frame, sizeof(__pyx_k_masked_server_frame), 0, 0, 1, 0},
  {&__pyx_n_s_masking_key, __pyx_k_masking_key, sizeof(__pyx_k_masking_key), 0, 0, 1, 1},
  {&__pyx_n_s_max_payload, __pyx_k_max_payload, sizeof(__pyx_k_max_payload), 0, 0, 1, 1},
  {&__pyx_n_s_message, __pyx_k_message, sizeof(__pyx_k_message), 0, 0, 1, 1},
  {&__pyx_n_s_multi_encode, __pyx_k_multi_encode, sizeof(__pyx_k_multi_encode), 0, 0, 1, 1},
  {&__pyx_n_s_next, __pyx_k_next, sizeof(__pyx_k_next), 0, 0, 1, 1},
  {&__pyx_n_s_opcode, __pyx_k_opcode, sizeof(__pyx_k_opcode), 0, 0, 1, 1},
  {&__pyx_n_s_os, __pyx_k_os, sizeof(__pyx_k_os), 0, 0, 1, 1},
  {&__pyx_n_s_pack, __pyx_k_pack, sizeof(__pyx_k_pack), 0, 0, 1, 1},
  {&__pyx_n_s_pack_2, __pyx_k_pack_2, sizeof(__pyx_k_pack_2), 0, 0, 1, 1},
  {&__pyx_n_s_pack_command, __pyx_k_pack_command, sizeof(__pyx_k_pack_command), 0, 0, 1, 1},
  {&__pyx_n_s_pack_pipeline_locals_genexpr, __pyx_k_pack_pipeline_locals_genexpr, sizeof(__pyx_k_pack_pipeline_locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_pack_pipeline_locals_lambda, __pyx_k_pack_pipeline_locals_lambda, sizeof(__pyx_k_pack_pipeline_locals_lambda), 0, 0, 1, 1},
  {&__pyx_n_s_payload_length, __pyx_k_payload_length, sizeof(__pyx_k_payload_length), 0, 0, 1, 1},
  {&__pyx_n_s_perr, __pyx_k_perr, sizeof(__pyx_k_perr), 0, 0, 1, 1},
  {&__pyx_n_s_protocols, __pyx_k_protocols, sizeof(__pyx_k_protocols), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_capi, __pyx_k_pyx_capi, sizeof(__pyx_k_pyx_capi), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_receive, __pyx_k_receive, sizeof(__pyx_k_receive), 0, 0, 1, 1},
  {&__pyx_n_s_rerr, __pyx_k_rerr, sizeof(__pyx_k_rerr), 0, 0, 1, 1},
  {&__pyx_n_s_rsv1, __pyx_k_rsv1, sizeof(__pyx_k_rsv1), 0, 0, 1, 1},
  {&__pyx_n_s_rsv2, __pyx_k_rsv2, sizeof(__pyx_k_rsv2), 0, 0, 1, 1},
  {&__pyx_n_s_rsv3, __pyx_k_rsv3, sizeof(__pyx_k_rsv3), 0, 0, 1, 1},
  {&__pyx_kp_s_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 0, 1, 0},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_starmap, __pyx_k_starmap, sizeof(__pyx_k_starmap), 0, 0, 1, 1},
  {&__pyx_n_s_struct, __pyx_k_struct, sizeof(__pyx_k_struct), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_kp_s_unmasked_client_frame, __pyx_k_unmasked_client_frame, sizeof(__pyx_k_unmasked_client_frame), 0, 0, 1, 0},
  {&__pyx_n_s_unpack, __pyx_k_unpack, sizeof(__pyx_k_unpack), 0, 0, 1, 1},
  {&__pyx_n_s_urandom, __pyx_k_urandom, sizeof(__pyx_k_urandom), 0, 0, 1, 1},
  {&__pyx_kp_s_utf_8, __pyx_k_utf_8, sizeof(__pyx_k_utf_8), 0, 0, 1, 0},
  {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
  {&__pyx_n_s_version, __pyx_k_version, sizeof(__pyx_k_version), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(2, 13, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "extensions/lib/rparser.pyx":60
 *             return nil
 *         else:
 *             return ('$%d\r\n' % len(value)).encode('utf-8') + value + b'\r\n'             # <<<<<<<<<<<<<<
 * 
 *     def multi_bulk_len(self, len):
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_utf_8); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "extensions/lib/rparser.pyx":63
 * 
 *     def multi_bulk_len(self, len):
 *         return ('*%s\r\n' % len).encode('utf-8')             # <<<<<<<<<<<<<<
 * 
 *     def multi_bulk(self, args):
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_utf_8); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "extensions/lib/rparser.pyx":70
 *     # INTERNALS
 *     def _pack_command(self, args):
 *         yield ('*%d\r\n' % len(args)).encode('utf-8')             # <<<<<<<<<<<<<<
 *         for value in args:
 *             if isinstance(value, str):
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_utf_8); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "extensions/lib/rparser.pyx":73
 *         for value in args:
 *             if isinstance(value, str):
 *                 value = value.encode('utf-8')             # <<<<<<<<<<<<<<
 *             elif not isinstance(value, bytes):
 *                 value = str(value).encode('utf-8')
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_s_utf_8); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "extensions/lib/rparser.pyx":75
 *                 value = value.encode('utf-8')
 *             elif not isinstance(value, bytes):
 *                 value = str(value).encode('utf-8')             # <<<<<<<<<<<<<<
 *             yield ('$%d\r\n' % len(value)).encode('utf-8')
 *             yield value
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_kp_s_utf_8); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "extensions/lib/rparser.pyx":76
 *             elif not isinstance(value, bytes):
 *                 value = str(value).encode('utf-8')
 *             yield ('$%d\r\n' % len(value)).encode('utf-8')             # <<<<<<<<<<<<<<
 *             yield value
 *             yield CRLF
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_s_utf_8); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 76, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "extensions/lib/rparser.pyx":81
 * 
 *     def _pack(self, args):
 *         yield ('*%d\r\n' % len(args)).encode('utf-8')             # <<<<<<<<<<<<<<
 *         for value in args:
 *             if value is None:
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_s_utf_8); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "extensions/lib/rparser.pyx":86
 *                 yield nil
 *             elif isinstance(value, bytes):
 *                 yield ('$%d\r\n' % len(value)).encode('utf-8')             # <<<<<<<<<<<<<<
 *                 yield value
 *                 yield CRLF
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_s_utf_8); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "extensions/lib/rparser.pyx":90
 *                 yield CRLF
 *             elif isinstance(value, str):
 *                 value = value.encode('utf-8')             # <<<<<<<<<<<<<<
 *                 yield ('$%d\r\n' % len(value)).encode('utf-8')
 *                 yield value
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_s_utf_8); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "extensions/lib/rparser.pyx":91
 *             elif isinstance(value, str):
 *                 value = value.encode('utf-8')
 *                 yield ('$%d\r\n' % len(value)).encode('utf-8')             # <<<<<<<<<<<<<<
 *                 yield value
 *                 yield CRLF
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_s_utf_8); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "extensions/lib/rparser.pyx":101
 *                     yield value
 *             else:
 *                 value = str(value).encode('utf-8')             # <<<<<<<<<<<<<<
 *                 yield ('$%d\r\n' % len(value)).encode('utf-8')
 *                 yield value
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_s_utf_8); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "extensions/lib/rparser.pyx":102
 *             else:
 *                 value = str(value).encode('utf-8')
 *                 yield ('$%d\r\n' % len(value)).encode('utf-8')             # <<<<<<<<<<<<<<
 *                 yield value
 *                 yield CRLF
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_s_utf_8); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "extensions/lib/rparser.pyx":117
 *     cdef object _get(self, Task next):
 *         b = self._inbuffer
 *         cdef int length = b.find(b'\r\n')             # <<<<<<<<<<<<<<
 *         if length >= 0:
 *             self._inbuffer, response = b[length+2:], bytes(b[:length])
 */
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_kp_b__3); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "extensions/lib/rparser.pyx":136
 *                 # Clear the buffer and raise
 *                 self._inbuffer = bytearray()
 *                 raise self._protocolError('Protocol Error')             # <<<<<<<<<<<<<<
 *         else:
 *             return False
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_s_Protocol_Error); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "extensions/lib/websocket.pyx":84
 *         self.buffer = bytearray()
 *         self.ProtocolError = ProtocolError
 *         self._opcodes = (0, 1, 2, 8, 9, 10)             # <<<<<<<<<<<<<<
 *         self._encode_mask_length = 0
 *         self._decode_mask_length = 0
 */
  __pyx_tuple__17 = PyTuple_Pack(6, __pyx_int_0, __pyx_int_1, __pyx_int_2, __pyx_int_8, __pyx_int_9, __pyx_int_10); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "extensions/lib/websocket.pyx":135
 *         body = pack('!H', code)
 *         if self._close_codes is not None:
 *             body += self._close_codes.get(code, '').encode('utf-8')             # <<<<<<<<<<<<<<
 *         return self.encode(body, opcode=0x8)
 * 
 */
  __pyx_tuple__18 = PyTuple_Pack(1, __pyx_kp_s_utf_8); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(1, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "extensions/lib/websocket.pyx":189
 *             if len(self.buffer) < 2:
 *                 return
 *             first_byte, second_byte = unpack("BB", self.buffer[:2])             # <<<<<<<<<<<<<<
 *             fin = (first_byte >> 7) & 1
 *             rsv1 = (first_byte >> 6) & 1
 */
  __pyx_slice__19 = PySlice_New(Py_None, __pyx_int_2, Py_None); if (unlikely(!__pyx_slice__19)) __PYX_ERR(1, 189, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__19);
  __Pyx_GIVEREF(__pyx_slice__19);

  /* "extensions/lib/websocket.pyx":196
 *             opcode = first_byte & 0xf
 *             if fin not in (0, 1):
 *                 raise self.ProtocolError('FIN must be 0 or 1')             # <<<<<<<<<<<<<<
 *             if bool(mask_length) != bool(second_byte & 0x80):
 *                 if mask_length:
 */
  __pyx_tuple__20 = PyTuple_Pack(1, __pyx_kp_s_FIN_must_be_0_or_1); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(1, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "extensions/lib/websocket.pyx":199
 *             if bool(mask_length) != bool(second_byte & 0x80):
 *                 if mask_length:
 *                     raise self.ProtocolError('unmasked client frame.')             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise self.ProtocolError('masked server frame.')
 */
  __pyx_tuple__21 = PyTuple_Pack(1, __pyx_kp_s_unmasked_client_frame); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(1, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);

  /* "extensions/lib/websocket.pyx":201
 *                     raise self.ProtocolError('unmasked client frame.')
 *                 else:
 *                     raise self.ProtocolError('masked server frame.')             # <<<<<<<<<<<<<<
 *             payload_length = second_byte & 0x7f
 *             # All control frames MUST have a payload length of 125 bytes
 */
  __pyx_tuple__22 = PyTuple_Pack(1, __pyx_kp_s_masked_server_frame); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(1, 201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);

  /* "extensions/lib/websocket.pyx":207
 *             if opcode > 7:
 *                 if payload_length > 125:
 *                     raise self.ProtocolError(             # <<<<<<<<<<<<<<
 *                         'WEBSOCKET control frame too large')
 *                 elif not fin:
 */
  __pyx_tuple__23 = PyTuple_Pack(1, __pyx_kp_s_WEBSOCKET_control_frame_too_larg); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(1, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);

  /* "extensions/lib/websocket.pyx":210
 *                         'WEBSOCKET control frame too large')
 *                 elif not fin:
 *                     raise self.ProtocolError(             # <<<<<<<<<<<<<<
 *                         'WEBSOCKET control frame fragmented')
 *             chunk = self._chunk(2)
 */
  __pyx_tuple__24 = PyTuple_Pack(1, __pyx_kp_s_WEBSOCKET_control_frame_fragment); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(1, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);

  /* "extensions/lib/websocket.pyx":266
 *             buffer.extend(pack('!Q', length))
 *         else:
 *             raise self.ProtocolError('WEBSOCKET frame too large')             # <<<<<<<<<<<<<<
 *         if masking_key:
 *             buffer.extend(masking_key)
 */
  __pyx_tuple__25 = PyTuple_Pack(1, __pyx_kp_s_WEBSOCKET_frame_too_large); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(1, 266, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);

  /* "extensions/lib/websocket.pyx":279
 *         #
 *         if mask_length:
 *             masking_key = masking_key or os.urandom(4)             # <<<<<<<<<<<<<<
 *             assert len(masking_key) == mask_length, "bad masking key"
 *         else:
 */
  __pyx_tuple__26 = PyTuple_Pack(1, __pyx_int_4); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(1, 279, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);

  /* "extensions/lib/websocket.pyx":287
 *         data = to_bytes(message or b'', 'utf-8')
 *         if opcode not in self._opcodes:
 *             raise self.ProtocolError('WEBSOCKET opcode a reserved value')             # <<<<<<<<<<<<<<
 *         elif opcode > 7:
 *             if len(data) > 125:
 */
  __pyx_tuple__27 = PyTuple_Pack(1, __pyx_kp_s_WEBSOCKET_opcode_a_reserved_valu); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(1, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);

  /* "extensions/lib/websocket.pyx":290
 *         elif opcode > 7:
 *             if len(data) > 125:
 *                 raise self.ProtocolError('WEBSOCKET control frame too large')             # <<<<<<<<<<<<<<
 *             if opcode == 8:
 *                 #TODO CHECK CLOSE FRAME STATUS CODE
 */
  __pyx_tuple__28 = PyTuple_Pack(1, __pyx_kp_s_WEBSOCKET_control_frame_too_larg); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(1, 290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);

  /* "websocket.pxd":13
 *         return value.encode(encoding)
 *     else:
 *         raise TypeError('Requires bytes or string')             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__29 = PyTuple_Pack(1, __pyx_kp_s_Requires_bytes_or_string); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(2, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(4, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(4, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(4, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(4, 1, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(4, 1, __pyx_L1_error)
  __pyx_int_5 = PyInt_FromLong(5); if (unlikely(!__pyx_int_5)) __PYX_ERR(4, 1, __pyx_L1_error)
  __pyx_int_6 = PyInt_FromLong(6); if (unlikely(!__pyx_int_6)) __PYX_ERR(4, 1, __pyx_L1_error)
  __pyx_int_7 = PyInt_FromLong(7); if (unlikely(!__pyx_int_7)) __PYX_ERR(4, 1, __pyx_L1_error)
  __pyx_int_8 = PyInt_FromLong(8); if (unlikely(!__pyx_int_8)) __PYX_ERR(4, 1, __pyx_L1_error)
  __pyx_int_9 = PyInt_FromLong(9); if (unlikely(!__pyx_int_9)) __PYX_ERR(4, 1, __pyx_L1_error)
  __pyx_int_10 = PyInt_FromLong(10); if (unlikely(!__pyx_int_10)) __PYX_ERR(4, 1, __pyx_L1_error)
  __pyx_int_15 = PyInt_FromLong(15); if (unlikely(!__pyx_int_15)) __PYX_ERR(4, 1, __pyx_L1_error)
  __pyx_int_127 = PyInt_FromLong(127); if (unlikely(!__pyx_int_127)) __PYX_ERR(4, 1, __pyx_L1_error)
  __pyx_int_128 = PyInt_FromLong(128); if (unlikely(!__pyx_int_128)) __PYX_ERR(4, 1, __pyx_L1_error)
  __pyx_int_9223372036854775808 = PyInt_FromString((char *)"9223372036854775808", 0, 0); if (unlikely(!__pyx_int_9223372036854775808)) __PYX_ERR(4, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC initlib(void); /*proto*/
PyMODINIT_FUNC initlib(void)
#else
PyMODINIT_FUNC PyInit_lib(void); /*proto*/
PyMODINIT_FUNC PyInit_lib(void)
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
      PyErr_Clear();
      __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
      if (!__Pyx_RefNanny)
          Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  __Pyx_RefNannySetupContext("PyMODINIT_FUNC PyInit_lib(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(4, 1, __pyx_L1_error)
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(4, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(4, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(4, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(4, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(4, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(4, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(4, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(4, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("lib", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(4, 1, __pyx_L1_error)
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(4, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(4, 1, __pyx_L1_error)
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(4, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(4, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(4, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_lib) {
    if (PyObject_SetAttrString(__pyx_m, "__name__", __pyx_n_s_main) < 0) __PYX_ERR(4, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(4, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "lib")) {
      if (unlikely(PyDict_SetItemString(modules, "lib", __pyx_m) < 0)) __PYX_ERR(4, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(4, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(4, 1, __pyx_L1_error)
  /*--- Global init code ---*/
  __pyx_v_3lib_CRLF = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_3lib_RESPONSE_INTEGER = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_3lib_RESPONSE_STRING = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_3lib_RESPONSE_ARRAY = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_3lib_RESPONSE_STATUS = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_3lib_RESPONSE_ERROR = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_3lib_nil = ((PyObject*)Py_None); Py_INCREF(Py_None);
  __pyx_v_3lib_null_array = ((PyObject*)Py_None); Py_INCREF(Py_None);
  /*--- Variable export code ---*/
  if (__Pyx_ExportVoidPtr(__pyx_n_s_clog2, (void *)&__pyx_v_3lib_clog2, "double") < 0) __PYX_ERR(4, 1, __pyx_L1_error)
  /*--- Function export code ---*/
  /*--- Type init code ---*/
  __pyx_vtabptr_3lib_Task = &__pyx_vtable_3lib_Task;
  __pyx_vtable_3lib_Task.decode = (PyObject *(*)(struct __pyx_obj_3lib_Task *, struct __pyx_obj_3lib_RedisParser *, PyObject *))__pyx_f_3lib_4Task_decode;
  if (PyType_Ready(&__pyx_type_3lib_Task) < 0) __PYX_ERR(0, 148, __pyx_L1_error)
  __pyx_type_3lib_Task.tp_print = 0;
  if (__Pyx_SetVtable(__pyx_type_3lib_Task.tp_dict, __pyx_vtabptr_3lib_Task) < 0) __PYX_ERR(0, 148, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "Task", (PyObject *)&__pyx_type_3lib_Task) < 0) __PYX_ERR(0, 148, __pyx_L1_error)
  __pyx_ptype_3lib_Task = &__pyx_type_3lib_Task;
  __pyx_vtabptr_3lib_RedisParser = &__pyx_vtable_3lib_RedisParser;
  __pyx_vtable_3lib_RedisParser._get = (PyObject *(*)(struct __pyx_obj_3lib_RedisParser *, struct __pyx_obj_3lib_Task *))__pyx_f_3lib_11RedisParser__get;
  __pyx_vtable_3lib_RedisParser._resume = (PyObject *(*)(struct __pyx_obj_3lib_RedisParser *, struct __pyx_obj_3lib_Task *, PyObject *))__pyx_f_3lib_11RedisParser__resume;
  if (PyType_Ready(&__pyx_type_3lib_RedisParser) < 0) __PYX_ERR(0, 15, __pyx_L1_error)
  __pyx_type_3lib_RedisParser.tp_print = 0;
  if (__Pyx_SetVtable(__pyx_type_3lib_RedisParser.tp_dict, __pyx_vtabptr_3lib_RedisParser) < 0) __PYX_ERR(0, 15, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "RedisParser", (PyObject *)&__pyx_type_3lib_RedisParser) < 0) __PYX_ERR(0, 15, __pyx_L1_error)
  __pyx_ptype_3lib_RedisParser = &__pyx_type_3lib_RedisParser;
  __pyx_vtabptr_3lib_ArrayTask = &__pyx_vtable_3lib_ArrayTask;
  __pyx_vtable_3lib_ArrayTask.__pyx_base = *__pyx_vtabptr_3lib_Task;
  __pyx_vtable_3lib_ArrayTask.__pyx_base.decode = (PyObject *(*)(struct __pyx_obj_3lib_Task *, struct __pyx_obj_3lib_RedisParser *, PyObject *))__pyx_f_3lib_9ArrayTask_decode;
  __pyx_type_3lib_ArrayTask.tp_base = __pyx_ptype_3lib_Task;
  if (PyType_Ready(&__pyx_type_3lib_ArrayTask) < 0) __PYX_ERR(0, 173, __pyx_L1_error)
  __pyx_type_3lib_ArrayTask.tp_print = 0;
  if (__Pyx_SetVtable(__pyx_type_3lib_ArrayTask.tp_dict, __pyx_vtabptr_3lib_ArrayTask) < 0) __PYX_ERR(0, 173, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "ArrayTask", (PyObject *)&__pyx_type_3lib_ArrayTask) < 0) __PYX_ERR(0, 173, __pyx_L1_error)
  __pyx_ptype_3lib_ArrayTask = &__pyx_type_3lib_ArrayTask;
  if (PyType_Ready(&__pyx_type_3lib_Frame) < 0) __PYX_ERR(1, 8, __pyx_L1_error)
  __pyx_type_3lib_Frame.tp_print = 0;
  if (PyObject_SetAttrString(__pyx_m, "Frame", (PyObject *)&__pyx_type_3lib_Frame) < 0) __PYX_ERR(1, 8, __pyx_L1_error)
  __pyx_ptype_3lib_Frame = &__pyx_type_3lib_Frame;
  __pyx_vtabptr_3lib_FrameParser = &__pyx_vtable_3lib_FrameParser;
  __pyx_vtable_3lib_FrameParser._encode = (PyObject *(*)(struct __pyx_obj_3lib_FrameParser *, PyObject *, int, PyObject *, int, int, int, int))__pyx_f_3lib_11FrameParser__encode;
  __pyx_vtable_3lib_FrameParser._info = (PyObject *(*)(struct __pyx_obj_3lib_FrameParser *, PyObject *, PyObject *, PyObject *))__pyx_f_3lib_11FrameParser__info;
  __pyx_vtable_3lib_FrameParser._chunk = (PyObject *(*)(struct __pyx_obj_3lib_FrameParser *, int))__pyx_f_3lib_11FrameParser__chunk;
  if (PyType_Ready(&__pyx_type_3lib_FrameParser) < 0) __PYX_ERR(1, 57, __pyx_L1_error)
  __pyx_type_3lib_FrameParser.tp_print = 0;
  if (__Pyx_SetVtable(__pyx_type_3lib_FrameParser.tp_dict, __pyx_vtabptr_3lib_FrameParser) < 0) __PYX_ERR(1, 57, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "FrameParser", (PyObject *)&__pyx_type_3lib_FrameParser) < 0) __PYX_ERR(1, 57, __pyx_L1_error)
  __pyx_ptype_3lib_FrameParser = &__pyx_type_3lib_FrameParser;
  if (PyType_Ready(&__pyx_type_3lib___pyx_scope_struct__pack_pipeline) < 0) __PYX_ERR(0, 51, __pyx_L1_error)
  __pyx_type_3lib___pyx_scope_struct__pack_pipeline.tp_print = 0;
  __pyx_ptype_3lib___pyx_scope_struct__pack_pipeline = &__pyx_type_3lib___pyx_scope_struct__pack_pipeline;
  if (PyType_Ready(&__pyx_type_3lib___pyx_scope_struct_1_genexpr) < 0) __PYX_ERR(0, 53, __pyx_L1_error)
  __pyx_type_3lib___pyx_scope_struct_1_genexpr.tp_print = 0;
  __pyx_ptype_3lib___pyx_scope_struct_1_genexpr = &__pyx_type_3lib___pyx_scope_struct_1_genexpr;
  if (PyType_Ready(&__pyx_type_3lib___pyx_scope_struct_2__pack_command) < 0) __PYX_ERR(0, 69, __pyx_L1_error)
  __pyx_type_3lib___pyx_scope_struct_2__pack_command.tp_print = 0;
  __pyx_ptype_3lib___pyx_scope_struct_2__pack_command = &__pyx_type_3lib___pyx_scope_struct_2__pack_command;
  if (PyType_Ready(&__pyx_type_3lib___pyx_scope_struct_3__pack) < 0) __PYX_ERR(0, 80, __pyx_L1_error)
  __pyx_type_3lib___pyx_scope_struct_3__pack.tp_print = 0;
  __pyx_ptype_3lib___pyx_scope_struct_3__pack = &__pyx_type_3lib___pyx_scope_struct_3__pack;
  if (PyType_Ready(&__pyx_type_3lib___pyx_scope_struct_4__lua_dict) < 0) __PYX_ERR(0, 106, __pyx_L1_error)
  __pyx_type_3lib___pyx_scope_struct_4__lua_dict.tp_print = 0;
  __pyx_ptype_3lib___pyx_scope_struct_4__lua_dict = &__pyx_type_3lib___pyx_scope_struct_4__lua_dict;
  if (PyType_Ready(&__pyx_type_3lib___pyx_scope_struct_5_multi_encode) < 0) __PYX_ERR(1, 156, __pyx_L1_error)
  __pyx_type_3lib___pyx_scope_struct_5_multi_encode.tp_print = 0;
  __pyx_ptype_3lib___pyx_scope_struct_5_multi_encode = &__pyx_type_3lib___pyx_scope_struct_5_multi_encode;
  /*--- Type import code ---*/
  /*--- Variable import code ---*/
  /*--- Function import code ---*/
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(4, 1, __pyx_L1_error)
  #endif

  /* "extensions/lib/rparser.pyx":1
 * from itertools import starmap             # <<<<<<<<<<<<<<
 * 
 * cdef class Task
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_starmap);
  __Pyx_GIVEREF(__pyx_n_s_starmap);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_starmap);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_itertools, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_starmap); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_starmap, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "extensions/lib/rparser.pyx":5
 * cdef class Task
 * 
 * cdef bytes CRLF = b"\r\n"             # <<<<<<<<<<<<<<
 * cdef bytes RESPONSE_INTEGER  = b':'
 * cdef bytes RESPONSE_STRING  = b'$'
 */
  __Pyx_INCREF(__pyx_kp_b__3);
  __Pyx_XGOTREF(__pyx_v_3lib_CRLF);
  __Pyx_DECREF_SET(__pyx_v_3lib_CRLF, __pyx_kp_b__3);
  __Pyx_GIVEREF(__pyx_kp_b__3);

  /* "extensions/lib/rparser.pyx":6
 * 
 * cdef bytes CRLF = b"\r\n"
 * cdef bytes RESPONSE_INTEGER  = b':'             # <<<<<<<<<<<<<<
 * cdef bytes RESPONSE_STRING  = b'$'
 * cdef bytes RESPONSE_ARRAY = b'*'
 */
  __Pyx_INCREF(__pyx_kp_b__30);
  __Pyx_XGOTREF(__pyx_v_3lib_RESPONSE_INTEGER);
  __Pyx_DECREF_SET(__pyx_v_3lib_RESPONSE_INTEGER, __pyx_kp_b__30);
  __Pyx_GIVEREF(__pyx_kp_b__30);

  /* "extensions/lib/rparser.pyx":7
 * cdef bytes CRLF = b"\r\n"
 * cdef bytes RESPONSE_INTEGER  = b':'
 * cdef bytes RESPONSE_STRING  = b'$'             # <<<<<<<<<<<<<<
 * cdef bytes RESPONSE_ARRAY = b'*'
 * cdef bytes RESPONSE_STATUS = b'+'
 */
  __Pyx_INCREF(__pyx_kp_b__31);
  __Pyx_XGOTREF(__pyx_v_3lib_RESPONSE_STRING);
  __Pyx_DECREF_SET(__pyx_v_3lib_RESPONSE_STRING, __pyx_kp_b__31);
  __Pyx_GIVEREF(__pyx_kp_b__31);

  /* "extensions/lib/rparser.pyx":8
 * cdef bytes RESPONSE_INTEGER  = b':'
 * cdef bytes RESPONSE_STRING  = b'$'
 * cdef bytes RESPONSE_ARRAY = b'*'             # <<<<<<<<<<<<<<
 * cdef bytes RESPONSE_STATUS = b'+'
 * cdef bytes RESPONSE_ERROR = b'-'
 */
  __Pyx_INCREF(__pyx_kp_b__32);
  __Pyx_XGOTREF(__pyx_v_3lib_RESPONSE_ARRAY);
  __Pyx_DECREF_SET(__pyx_v_3lib_RESPONSE_ARRAY, __pyx_kp_b__32);
  __Pyx_GIVEREF(__pyx_kp_b__32);

  /* "extensions/lib/rparser.pyx":9
 * cdef bytes RESPONSE_STRING  = b'$'
 * cdef bytes RESPONSE_ARRAY = b'*'
 * cdef bytes RESPONSE_STATUS = b'+'             # <<<<<<<<<<<<<<
 * cdef bytes RESPONSE_ERROR = b'-'
 * cdef bytes nil = b'$-1\r\n'
 */
  __Pyx_INCREF(__pyx_kp_b__33);
  __Pyx_XGOTREF(__pyx_v_3lib_RESPONSE_STATUS);
  __Pyx_DECREF_SET(__pyx_v_3lib_RESPONSE_STATUS, __pyx_kp_b__33);
  __Pyx_GIVEREF(__pyx_kp_b__33);

  /* "extensions/lib/rparser.pyx":10
 * cdef bytes RESPONSE_ARRAY = b'*'
 * cdef bytes RESPONSE_STATUS = b'+'
 * cdef bytes RESPONSE_ERROR = b'-'             # <<<<<<<<<<<<<<
 * cdef bytes nil = b'$-1\r\n'
 * cdef bytes null_array = b'*-1\r\n'
 */
  __Pyx_INCREF(__pyx_kp_b__34);
  __Pyx_XGOTREF(__pyx_v_3lib_RESPONSE_ERROR);
  __Pyx_DECREF_SET(__pyx_v_3lib_RESPONSE_ERROR, __pyx_kp_b__34);
  __Pyx_GIVEREF(__pyx_kp_b__34);

  /* "extensions/lib/rparser.pyx":11
 * cdef bytes RESPONSE_STATUS = b'+'
 * cdef bytes RESPONSE_ERROR = b'-'
 * cdef bytes nil = b'$-1\r\n'             # <<<<<<<<<<<<<<
 * cdef bytes null_array = b'*-1\r\n'
 * 
 */
  __Pyx_INCREF(__pyx_kp_b_1);
  __Pyx_XGOTREF(__pyx_v_3lib_nil);
  __Pyx_DECREF_SET(__pyx_v_3lib_nil, __pyx_kp_b_1);
  __Pyx_GIVEREF(__pyx_kp_b_1);

  /* "extensions/lib/rparser.pyx":12
 * cdef bytes RESPONSE_ERROR = b'-'
 * cdef bytes nil = b'$-1\r\n'
 * cdef bytes null_array = b'*-1\r\n'             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_INCREF(__pyx_kp_b_1_2);
  __Pyx_XGOTREF(__pyx_v_3lib_null_array);
  __Pyx_DECREF_SET(__pyx_v_3lib_null_array, __pyx_kp_b_1_2);
  __Pyx_GIVEREF(__pyx_kp_b_1_2);

  /* "extensions/lib/websocket.pyx":1
 * import os             # <<<<<<<<<<<<<<
 * from struct import pack, unpack
 * 
 */
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_os, 0, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_os, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "extensions/lib/websocket.pyx":2
 * import os
 * from struct import pack, unpack             # <<<<<<<<<<<<<<
 * 
 * cimport cython
 */
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_pack_2);
  __Pyx_GIVEREF(__pyx_n_s_pack_2);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_pack_2);
  __Pyx_INCREF(__pyx_n_s_unpack);
  __Pyx_GIVEREF(__pyx_n_s_unpack);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_unpack);
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_struct, __pyx_t_2, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_pack_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pack_2, __pyx_t_2) < 0) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_unpack); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_unpack, __pyx_t_2) < 0) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "lib.pyx":1
 * include "rparser.pyx"             # <<<<<<<<<<<<<<
 * include "websocket.pyx"
 */
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(4, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "lib.pxd":14
 * 
 * # MSVC does not have log2!
 * cdef inline double Log2(double x):             # <<<<<<<<<<<<<<
 *     return log(x) / clog2
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init lib", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init lib");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if PY_MAJOR_VERSION < 3
  return;
  #else
  return __pyx_m;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || PyObject_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* StringJoin */
  #if !CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values) {
    return PyObject_CallMethodObjArgs(sep, __pyx_n_s_join, values, NULL);
}
#endif

/* KeywordStringCheck */
  static CYTHON_INLINE int __Pyx_CheckKeywordStrings(
    PyObject *kwdict,
    const char* function_name,
    int kw_allowed)
{
    PyObject* key = 0;
    Py_ssize_t pos = 0;
#if CYTHON_COMPILING_IN_PYPY
    if (!kw_allowed && PyDict_Next(kwdict, &pos, &key, 0))
        goto invalid_keyword;
    return 1;
#else
    while (PyDict_Next(kwdict, &pos, &key, 0)) {
        #if PY_MAJOR_VERSION < 3
        if (unlikely(!PyString_CheckExact(key)) && unlikely(!PyString_Check(key)))
        #endif
            if (unlikely(!PyUnicode_Check(key)))
                goto invalid_keyword_type;
    }
    if ((!kw_allowed) && unlikely(key))
        goto invalid_keyword;
    return 1;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    return 0;
#endif
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
    return 0;
}

/* None */
  static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

/* RaiseTooManyValuesToUnpack */
  static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
  static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
  static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyThreadState *tstate = PyThreadState_GET();
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(exc_type == PyExc_StopIteration) || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
  static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* FetchCommonType */
  static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunction */
  static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_COMPILING_IN_CPYTHON
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(__pyx_CyFunctionObject, func.m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromString(m->func.m_ml->ml_name);
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_New(PyTypeObject *type, PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    __pyx_CyFunctionObject *op = PyObject_GC_New(__pyx_CyFunctionObject, type);
    if (op == NULL)
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    PyObject_GC_Track(op);
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
#if CYTHON_COMPILING_IN_PYPY
static PyObject * __Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    PyObject *self = f->m_self;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0 = PySequence_ITEM(arg, 0);
                if (unlikely(!arg0)) return NULL;
                result = (*meth)(self, arg0);
                Py_DECREF(arg0);
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
#else
static PyObject * __Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
	return PyCFunction_Call(func, arg, kw);
}
#endif
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_Call,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
#if !CYTHON_COMPILING_IN_PYPY
    __pyx_CyFunctionType_type.tp_call = PyCFunction_Call;
#endif
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (__pyx_CyFunctionType == NULL) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (!m->defaults)
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* GetBuiltinName */
      static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* GetModuleGlobalName */
      static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON
    result = PyDict_GetItem(__pyx_d, name);
    if (likely(result)) {
        Py_INCREF(result);
    } else {
#else
    result = PyObject_GetItem(__pyx_d, name);
    if (!result) {
        PyErr_Clear();
#endif
        result = __Pyx_GetBuiltinName(name);
    }
    return result;
}

/* ArgTypeTest */
        static void __Pyx_RaiseArgumentTypeInvalid(const char* name, PyObject *obj, PyTypeObject *type) {
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
}
static CYTHON_INLINE int __Pyx_ArgTypeTest(PyObject *obj, PyTypeObject *type, int none_allowed,
    const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (none_allowed && obj == Py_None) return 1;
    else if (exact) {
        if (likely(Py_TYPE(obj) == type)) return 1;
        #if PY_MAJOR_VERSION == 2
        else if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(PyObject_TypeCheck(obj, type))) return 1;
    }
    __Pyx_RaiseArgumentTypeInvalid(name, obj, type);
    return 0;
}

/* PyIntBinop */
        #if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED int inplace) {
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS && PY_MAJOR_VERSION >= 3
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    }
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    }
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    }
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    }
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    }
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    }
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* SliceObject */
        static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_COMPILING_IN_CPYTHON
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* BytesEquals */
        static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* decode_c_bytes */
        static CYTHON_INLINE PyObject* __Pyx_decode_c_bytes(
         const char* cstring, Py_ssize_t length, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    if (unlikely((start < 0) | (stop < 0))) {
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    if (stop > length)
        stop = length;
    length = stop - start;
    if (unlikely(length <= 0))
        return PyUnicode_FromUnicode(NULL, 0);
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* PyErrFetchRestore */
        #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
        #if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
#if PY_VERSION_HEX >= 0x03030000
    if (cause) {
#else
    if (cause && cause != Py_None) {
#endif
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = PyThreadState_GET();
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* RaiseNoneIterError */
          static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* PyIntBinop */
          #if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx_PyInt_AndObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED int inplace) {
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
            return PyInt_FromLong(a & b);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS && PY_MAJOR_VERSION >= 3
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    }
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    }
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    }
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    }
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    }
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    }
                default: return PyLong_Type.tp_as_number->nb_and(op1, op2);
            }
        }
                x = a & b;
            return PyLong_FromLong(x);
        long_long:
                llx = lla & llb;
            return PyLong_FromLongLong(llx);
    }
    #endif
    return (inplace ? PyNumber_InPlaceAnd : PyNumber_And)(op1, op2);
}
#endif

/* PyIntBinop */
          #if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx_PyInt_RshiftObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED int inplace) {
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
            return PyInt_FromLong(a >> b);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS && PY_MAJOR_VERSION >= 3
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    }
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    }
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    }
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    }
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    }
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
                    }
                default: return PyLong_Type.tp_as_number->nb_rshift(op1, op2);
            }
        }
                x = a >> b;
            return PyLong_FromLong(x);
        long_long:
                llx = lla >> llb;
            return PyLong_FromLongLong(llx);
    }
    #endif
    return (inplace ? PyNumber_InPlaceRshift : PyNumber_Rshift)(op1, op2);
}
#endif

/* GetItemInt */
          static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (wraparound & unlikely(i < 0)) i += PyList_GET_SIZE(o);
    if ((!boundscheck) || likely((0 <= i) & (i < PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (wraparound & unlikely(i < 0)) i += PyTuple_GET_SIZE(o);
    if ((!boundscheck) || likely((0 <= i) & (i < PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_COMPILING_IN_CPYTHON
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely((n >= 0) & (n < PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely((n >= 0) & (n < PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* PyIntBinop */
          #if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED int inplace) {
    if (op1 == op2) {
        Py_RETURN_TRUE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a == b) {
            Py_RETURN_TRUE;
        } else {
            Py_RETURN_FALSE;
        }
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS && PY_MAJOR_VERSION >= 3
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a;
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                #if PyLong_SHIFT < 30 && PyLong_SHIFT != 15
                default: return PyLong_Type.tp_richcompare(op1, op2, Py_EQ);
                #else
                default: Py_RETURN_FALSE;
                #endif
            }
        }
            if (a == b) {
                Py_RETURN_TRUE;
            } else {
                Py_RETURN_FALSE;
            }
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            if ((double)a == (double)b) {
                Py_RETURN_TRUE;
            } else {
                Py_RETURN_FALSE;
            }
    }
    return PyObject_RichCompare(op1, op2, Py_EQ);
}
#endif

/* PyObjectCallMethod1 */
          static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method, *result = NULL;
    method = __Pyx_PyObject_GetAttrStr(obj, method_name);
    if (unlikely(!method)) goto bad;
#if CYTHON_COMPILING_IN_CPYTHON
    if (likely(PyMethod_Check(method))) {
        PyObject *self = PyMethod_GET_SELF(method);
        if (likely(self)) {
            PyObject *args;
            PyObject *function = PyMethod_GET_FUNCTION(method);
            args = PyTuple_New(2);
            if (unlikely(!args)) goto bad;
            Py_INCREF(self);
            PyTuple_SET_ITEM(args, 0, self);
            Py_INCREF(arg);
            PyTuple_SET_ITEM(args, 1, arg);
            Py_INCREF(function);
            Py_DECREF(method); method = NULL;
            result = __Pyx_PyObject_Call(function, args, NULL);
            Py_DECREF(args);
            Py_DECREF(function);
            return result;
        }
    }
#endif
    result = __Pyx_PyObject_CallOneArg(method, arg);
bad:
    Py_XDECREF(method);
    return result;
}

/* append */
          static CYTHON_INLINE int __Pyx_PyObject_Append(PyObject* L, PyObject* x) {
    if (likely(PyList_CheckExact(L))) {
        if (unlikely(__Pyx_PyList_Append(L, x) < 0)) return -1;
    } else {
        PyObject* retval = __Pyx_PyObject_CallMethod1(L, __pyx_n_s_append, x);
        if (unlikely(!retval))
            return -1;
        Py_DECREF(retval);
    }
    return 0;
}

/* WriteUnraisableException */
          static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* SetVTable */
          static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* Import */
          static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_VERSION_HEX < 0x03030000
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                #if PY_VERSION_HEX < 0x03030000
                PyObject *py_level = PyInt_FromLong(1);
                if (!py_level)
                    goto bad;
                module = PyObject_CallFunctionObjArgs(py_import,
                    name, global_dict, empty_dict, list, py_level, NULL);
                Py_DECREF(py_level);
                #else
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                #endif
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_VERSION_HEX < 0x03030000
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_VERSION_HEX < 0x03030000
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
          static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* CodeObjectCache */
          static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
          #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    py_code = __pyx_find_code_object(c_line ? c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        PyThreadState_GET(), /*PyThreadState *tstate,*/
        py_code,             /*PyCodeObject *code,*/
        __pyx_d,      /*PyObject *globals,*/
        0                    /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    py_frame->f_lineno = py_line;
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
          #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
          static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
          static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
          static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_PY_LONG_LONG(unsigned PY_LONG_LONG value) {
    const unsigned PY_LONG_LONG neg_one = (unsigned PY_LONG_LONG) -1, const_zero = (unsigned PY_LONG_LONG) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned PY_LONG_LONG) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned PY_LONG_LONG) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(unsigned PY_LONG_LONG) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(unsigned PY_LONG_LONG) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned PY_LONG_LONG) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned PY_LONG_LONG),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
          static CYTHON_INLINE PyObject* __Pyx_PyInt_From_PY_LONG_LONG(PY_LONG_LONG value) {
    const PY_LONG_LONG neg_one = (PY_LONG_LONG) -1, const_zero = (PY_LONG_LONG) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(PY_LONG_LONG) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(PY_LONG_LONG) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
        } else if (sizeof(PY_LONG_LONG) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
        }
    } else {
        if (sizeof(PY_LONG_LONG) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(PY_LONG_LONG) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(PY_LONG_LONG),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
          static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
          static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
          static CYTHON_INLINE unsigned PY_LONG_LONG __Pyx_PyInt_As_unsigned_PY_LONG_LONG(PyObject *x) {
    const unsigned PY_LONG_LONG neg_one = (unsigned PY_LONG_LONG) -1, const_zero = (unsigned PY_LONG_LONG) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(unsigned PY_LONG_LONG) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (unsigned PY_LONG_LONG) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned PY_LONG_LONG) 0;
                case  1: __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, digit, digits[0])
                case 2:
                    if (8 * sizeof(unsigned PY_LONG_LONG) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned PY_LONG_LONG) >= 2 * PyLong_SHIFT) {
                            return (unsigned PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned PY_LONG_LONG) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned PY_LONG_LONG) >= 3 * PyLong_SHIFT) {
                            return (unsigned PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned PY_LONG_LONG) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned PY_LONG_LONG) >= 4 * PyLong_SHIFT) {
                            return (unsigned PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (unsigned PY_LONG_LONG) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(unsigned PY_LONG_LONG) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned PY_LONG_LONG, unsigned long, PyLong_AsUnsignedLong(x))
            } else if (sizeof(unsigned PY_LONG_LONG) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned PY_LONG_LONG, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned PY_LONG_LONG) 0;
                case -1: __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(unsigned PY_LONG_LONG) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned PY_LONG_LONG) (((unsigned PY_LONG_LONG)-1)*(((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(unsigned PY_LONG_LONG) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned PY_LONG_LONG) ((((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(unsigned PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned PY_LONG_LONG) (((unsigned PY_LONG_LONG)-1)*(((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned PY_LONG_LONG) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned PY_LONG_LONG) ((((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(unsigned PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned PY_LONG_LONG) (((unsigned PY_LONG_LONG)-1)*(((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned PY_LONG_LONG) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned PY_LONG_LONG, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned PY_LONG_LONG) ((((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(unsigned PY_LONG_LONG) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned PY_LONG_LONG, long, PyLong_AsLong(x))
            } else if (sizeof(unsigned PY_LONG_LONG) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned PY_LONG_LONG, PY_LONG_LONG, PyLong_AsLongLong(x))
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            unsigned PY_LONG_LONG val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (unsigned PY_LONG_LONG) -1;
        }
    } else {
        unsigned PY_LONG_LONG val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (unsigned PY_LONG_LONG) -1;
        val = __Pyx_PyInt_As_unsigned_PY_LONG_LONG(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned PY_LONG_LONG");
    return (unsigned PY_LONG_LONG) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned PY_LONG_LONG");
    return (unsigned PY_LONG_LONG) -1;
}

/* SwapException */
          #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* CoroutineBase */
          #include <structmember.h>
#include <frameobject.h>
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
#if 1 || PY_VERSION_HEX < 0x030300B0
static int __Pyx_PyGen_FetchStopIterationValue(PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
#if PY_VERSION_HEX >= 0x030300A0
        if (ev && Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
#endif
        if (!ev || !PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            if (!ev) {
                Py_INCREF(Py_None);
                ev = Py_None;
            } else if (PyTuple_Check(ev)) {
                if (PyTuple_GET_SIZE(ev) >= 1) {
                    PyObject *value;
#if CYTHON_COMPILING_IN_CPYTHON
                    value = PySequence_ITEM(ev, 0);
#else
                    value = PyTuple_GET_ITEM(ev, 0);
                    Py_INCREF(value);
#endif
                    Py_DECREF(ev);
                    ev = value;
                } else {
                    Py_INCREF(Py_None);
                    Py_DECREF(ev);
                    ev = Py_None;
                }
            }
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = ev;
            return 0;
        }
    } else if (!PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
#endif
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__pyx_CoroutineObject *self) {
    PyObject *exc_type = self->exc_type;
    PyObject *exc_value = self->exc_value;
    PyObject *exc_traceback = self->exc_traceback;
    self->exc_type = NULL;
    self->exc_value = NULL;
    self->exc_traceback = NULL;
    Py_XDECREF(exc_type);
    Py_XDECREF(exc_value);
    Py_XDECREF(exc_traceback);
}
static CYTHON_INLINE
int __Pyx_Coroutine_CheckRunning(__pyx_CoroutineObject *gen) {
    if (unlikely(gen->is_running)) {
        PyErr_SetString(PyExc_ValueError,
                        "generator already executing");
        return 1;
    }
    return 0;
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value) {
    PyObject *retval;
    __Pyx_PyThreadState_declare
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            PyErr_SetString(PyExc_TypeError,
                            "can't send non-None value to a "
                            "just-started generator");
            return NULL;
        }
    }
    if (unlikely(self->resume_label == -1)) {
        PyErr_SetNone(PyExc_StopIteration);
        return NULL;
    }
    __Pyx_PyThreadState_assign
    if (value) {
#if CYTHON_COMPILING_IN_PYPY
#else
        if (self->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) self->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            Py_XINCREF(__pyx_tstate->frame);
            assert(f->f_back == NULL);
            f->f_back = __pyx_tstate->frame;
        }
#endif
        __Pyx_ExceptionSwap(&self->exc_type, &self->exc_value,
                            &self->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(self);
    }
    self->is_running = 1;
    retval = self->body((PyObject *) self, value);
    self->is_running = 0;
    if (retval) {
        __Pyx_ExceptionSwap(&self->exc_type, &self->exc_value,
                            &self->exc_traceback);
#if CYTHON_COMPILING_IN_PYPY
#else
        if (self->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) self->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            Py_CLEAR(f->f_back);
        }
#endif
    } else {
        __Pyx_Coroutine_ExceptionClear(self);
    }
    return retval;
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(PyObject *retval) {
    if (unlikely(!retval && !PyErr_Occurred())) {
        PyErr_SetNone(PyExc_StopIteration);
    }
    return retval;
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen_FetchStopIterationValue(&val);
    ret = __Pyx_Coroutine_SendEx(gen, val);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(__Pyx_Coroutine_CheckRunning(gen)))
        return NULL;
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value);
    }
    return __Pyx_Coroutine_MethodReturn(retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(__Pyx_Coroutine_CheckRunning(gen)))
        return NULL;
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(__Pyx_Coroutine_CheckRunning(gen)))
        return NULL;
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL);
    if (retval) {
        Py_DECREF(retval);
        PyErr_SetString(PyExc_RuntimeError,
                        "generator ignored GeneratorExit");
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (!raised_exception
        || raised_exception == PyExc_StopIteration
        || raised_exception == PyExc_GeneratorExit
        || PyErr_GivenExceptionMatches(raised_exception, PyExc_GeneratorExit)
        || PyErr_GivenExceptionMatches(raised_exception, PyExc_StopIteration))
    {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *typ;
    PyObject *tb = NULL;
    PyObject *val = NULL;
    PyObject *yf = gen->yieldfrom;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    if (unlikely(__Pyx_Coroutine_CheckRunning(gen)))
        return NULL;
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit)) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(__Pyx_Coroutine_SendEx(gen, NULL));
            goto throw_here;
        }
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Throw(yf, args);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Throw(yf, args);
        } else
        #endif
        {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            ret = PyObject_CallObject(meth, args);
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(__Pyx_Coroutine_SendEx(gen, NULL));
}
static int __Pyx_Coroutine_traverse(PyObject *self, visitproc visit, void *arg) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    Py_VISIT(gen->exc_type);
    Py_VISIT(gen->exc_value);
    Py_VISIT(gen->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    Py_CLEAR(gen->exc_type);
    Py_CLEAR(gen->exc_value);
    Py_CLEAR(gen->exc_traceback);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label > 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (self->ob_refcnt > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *res;
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label <= 0)
        return ;
#if PY_VERSION_HEX < 0x030400a1
    assert(self->ob_refcnt == 0);
    self->ob_refcnt = 1;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
    res = __Pyx_Coroutine_Close(self);
    if (res == NULL)
        PyErr_WriteUnraisable(self);
    else
        Py_DECREF(res);
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if PY_VERSION_HEX < 0x030400a1
    assert(self->ob_refcnt > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = self->ob_refcnt;
        _Py_NewReference(self);
        self->ob_refcnt = refcnt;
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(self->ob_type) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self)
{
    Py_INCREF(self->gi_name);
    return self->gi_name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self)
{
    Py_INCREF(self->gi_qualname);
    return self->gi_qualname;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(PyTypeObject* type, __pyx_coroutine_body_t body,
                                                   PyObject *closure, PyObject *name, PyObject *qualname) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (gen == NULL)
        return NULL;
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->exc_type = NULL;
    gen->exc_value = NULL;
    gen->exc_traceback = NULL;
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
              static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
              #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (!abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_VERSION_HEX >= 0x03030000) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_VERSION_HEX >= 0x03030000) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if (0) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
              static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if PY_VERSION_HEX >= 0x030400a1
    __Pyx_Coroutine_del,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = PyObject_GenericGetAttr;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
              static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* VoidPtrExport */
              static int __Pyx_ExportVoidPtr(PyObject *name, void *p, const char *sig) {
    PyObject *d;
    PyObject *cobj = 0;
    d = PyDict_GetItem(__pyx_d, __pyx_n_s_pyx_capi);
    Py_XINCREF(d);
    if (!d) {
        d = PyDict_New();
        if (!d)
            goto bad;
        if (__Pyx_PyObject_SetAttrStr(__pyx_m, __pyx_n_s_pyx_capi, d) < 0)
            goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(p, sig, 0);
#else
    cobj = PyCObject_FromVoidPtrAndDesc(p, (void *)sig, 0);
#endif
    if (!cobj)
        goto bad;
    if (PyDict_SetItem(d, name, cobj) < 0)
        goto bad;
    Py_DECREF(cobj);
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(cobj);
    Py_XDECREF(d);
    return -1;
}

/* InitStrings */
              static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if CYTHON_COMPILING_IN_CPYTHON && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
#if PY_VERSION_HEX < 0x03030000
        char* defenc_c;
        PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
        if (!defenc) return NULL;
        defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        {
            char* end = defenc_c + PyBytes_GET_SIZE(defenc);
            char* c;
            for (c = defenc_c; c < end; c++) {
                if ((unsigned char) (*c) >= 128) {
                    PyUnicode_AsASCIIString(o);
                    return NULL;
                }
            }
        }
#endif
        *length = PyBytes_GET_SIZE(defenc);
        return defenc_c;
#else
        if (__Pyx_PyUnicode_READY(o) == -1) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        if (PyUnicode_IS_ASCII(o)) {
            *length = PyUnicode_GET_LENGTH(o);
            return PyUnicode_AsUTF8(o);
        } else {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
#else
        return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
  PyNumberMethods *m;
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (PyInt_Check(x) || PyLong_Check(x))
#else
  if (PyLong_Check(x))
#endif
    return __Pyx_NewRef(x);
  m = Py_TYPE(x)->tp_as_number;
#if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = PyNumber_Long(x);
  }
#else
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Long(x);
  }
#endif
  if (res) {
#if PY_MAJOR_VERSION < 3
    if (!PyInt_Check(res) && !PyLong_Check(res)) {
#else
    if (!PyLong_Check(res)) {
#endif
      PyErr_Format(PyExc_TypeError,
                   "__%.4s__ returned non-%.4s (type %.200s)",
                   name, name, Py_TYPE(res)->tp_name);
      Py_DECREF(res);
      return NULL;
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(x);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
